package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnectionPool;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.BasicSshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.stitching.VlanRange;
import be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.RSpecContents;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.testing.base.ApiTest;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import be.iminds.ilabt.jfed.util.TextUtil;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.StringReader;
import java.util.*;

/**
 * CommonAMTest:
 * <p/>
 * methods for common tests in AM's
 */
public class CommonAMTest {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(CommonAMTest.class);
    private static Random random = new Random(System.currentTimeMillis());
    private static int MAX_CREATE_SLICE_TRIES = 3;
    private ApiTest test;
    private final SfaAuthority userAuthority;
    private final List<SfaAuthority> testedAuthoritiesForRspec;
    private List<AnyCredential> userCredentials;
    //    private ProtogeniSliceAuthority sa;  //not used directly, use userAndSliceApiWrapperUser1
    private UserAndSliceApiWrapper userAndSliceApiWrapperUser1;
    private UserAndSliceApiWrapper userAndSliceApiWrapperUser2;
    private List<AnyCredential> user2Credentials;
    private List<String> speaksForOptionNames = new ArrayList<String>();
    private TestCredentialType testCredentialType = TestCredentialType.REGULAR;

    private SfaCredential delegatedUserCredential;
    private SfaCredential speaksForUserCredential;
    private List<SliceInfo> allSliceInfo = new ArrayList<SliceInfo>();


    protected GeniUser secondUser;


    public CommonAMTest(@Nonnull ApiTest test) {
        this(test, test.getUser().getUserAuthority(), test.getTestedAuthority().getSfaAuthorityForRspec());
    }

    public CommonAMTest(@Nonnull ApiTest test, SfaAuthority userAuthority) {
        this(test, userAuthority, Arrays.asList(test.getTestedAuthority().getSfaAuthorityForRspec()));
    }
    public CommonAMTest(@Nonnull ApiTest test, SfaAuthority userAuthority, SfaAuthority testedAuthorityForRspec) {
        this(test, userAuthority, Arrays.asList(testedAuthorityForRspec));
    }
    public CommonAMTest(@Nonnull ApiTest test, SfaAuthority userAuthority, List<SfaAuthority> testedAuthoritiesForRspec) {
        this.test = test;
        this.userAuthority = userAuthority;
        this.testedAuthoritiesForRspec = testedAuthoritiesForRspec;

        assert this.test != null;
        assert this.userAuthority != null;

//        sa = new ProtogeniSliceAuthority(test.getLogger());

        userAndSliceApiWrapperUser1 =
                test.getAutomaticUserAndSliceApiWrapperFactory().create();

        //set both known speaksFor option names
        speaksForOptionNames.add("geni_experimenter_urn");
        speaksForOptionNames.add("geni_speaking_for");
    }

    public static String getComponentManagerIdForSite(SfaAuthority auth) {
        if (auth.getType() != null && auth.getType().equals("emulab"))
            return auth.getUrnString();
        if (auth.getType() != null && auth.getType().equals("instageni"))
            return auth.getUrnString();
        if (auth.getType() != null && auth.getType().equals("planetlab"))
            return auth.getUrnString();
        if (auth.getType() != null && (auth.getType().equals("orca") || auth.getType().equals("exogeni"))) {
            //see https://geni-orca.renci.org/trac/wiki/orca-and-rspec
            String sitename = null;
            if (auth.getUrnString().equals("urn:publicid:IDN+geni.renci.org+authority+cm")) //"ties resources from all"
                sitename = "exogeni.net"; //just a guess, perhaps this is ANY?
            // Note: There is a way to select any, but I do not know the urn: see https://wiki.exogeni.net/doku.php?id=public:experimenters:start#selecting_which_orca_actor_to_request_resources_from
            if (auth.getUrnString().equals("urn:publicid:IDN+bbn-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:bbnvmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+dbc1-16.nicl.cs.duke.edu+authority+cm"))
                sitename = "exogeni.net:dukevmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+fiu-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:fiuvmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+nicta-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:nictavmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+rci-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:rcivmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+uh-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:uhvmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+uva-nl-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:uvanlvmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+ufl-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:uflvmsite";
            if (auth.getUrnString().equals("urn:publicid:IDN+osf-hn.exogeni.net+authority+cm"))
                sitename = "exogeni.net:osfvmsite";
            //note, there is also uncvmsite: University NC (but URN and URL not known)


            if (sitename != null)
                return "urn:publicid:IDN+" + sitename + "+authority+cm";
        }
        return auth.getUrnString();
    }

    public static boolean getExclusiveForSite(SfaAuthority auth) {
        String sliverType = getSliverTypeForSite(auth);
        if (sliverType.equals("raw-pc"))
            return true;
        else
            return false;
    }

    public static String getSliverTypeForSite(SfaAuthority auth) {
        if (auth.getType() != null && auth.getType().equals("emulab"))
            return "raw-pc";
        if (auth.getType() != null && auth.getType().equals("instageni"))
            return "emulab-xen";
        if (auth.getType() != null && auth.getType().equals("planetlab"))
            return "plab-vserver";
        if (auth.getType() != null && (auth.getType().equals("orca") || auth.getType().equals("exogeni")))
            return "xo.small";
        if (auth.getNameForUrn().contains("exogeni") || auth.getNameForUrn().contains("vmsite") || auth.getNameForUrn().contains("orca"))
            return "xo.small";

        return null;
    }

    /**
     * returns an RSpec for one node, for the given authority.
     */
    public static String getOneNodeRequestRSpec(SfaAuthority auth, GeniUrn sliceUrn) {
        return getOneNodeRequestRSpec(auth, sliceUrn, null/*fixedNodeUrn*/);
    }

    public static String getOneNodeRequestRSpec(SfaAuthority auth, GeniUrn sliceUrn, @Nullable GeniUrn fixedNodeUrn) {
        String componentId = "";
        if (fixedNodeUrn != null)
            componentId = " component_id=\"" + fixedNodeUrn + "\" ";

        if (auth.getUrn().equals(GeniUrn.parse("urn:publicid:IDN+fiteagle+authority+am"))) {
            Random rand = new Random(System.currentTimeMillis());
            return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/ad.xsd http://fiteagle.org/rspec/ext/1 http://fiteagle.org/rspec/ext/1\" xmlns:ns2=\"http://fiteagle.org/rspec/ext/1\" xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:ns3=\"http://fiteagle.org/rspec/ext/1/openstackVMResource\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n" +
                    "<ns3:openstackResource resourceId=\"42ad1e74-c9cd-4fe0-9d18-522b7f8bede1\">\n" +
                    "<ns3:vmToInstantiate imageId=\"d6521c1b-4407-454a-981f-b5d4d4d5fcae\" flavorId=\"3\" vmName=\"randomName" + rand.nextInt() + "\" keyPairName=\"randomKeyPair" + rand.nextInt() + "\"/>\n" +
                    "</ns3:openstackResource>\n" +
                    "</rspec>";
        }
        if (auth.getUrn().equals(GeniUrn.parse("urn:publicid:IDN+clab+authority+cm"))) {
            return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \" " +
                    "xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" " +
                    "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                    "xmlns=\"http://www.geni.net/resources/rspec/3\">\n" +
                    "     <node client_id=\"myNodeId\" component_manager_id=\"urn:publicid:IDN+clab+authority+cm\" " + componentId + " exclusive=\"false\">\n" +
                    "         <sliver_type name=\"RD_sliver\"/>\n" +
                    "     </node>\n" +
                    "</rspec>";
        }
        if (auth.getUrn().getEncodedTopLevelAuthority().equals("omf:nitos")) {
            Random rand = new Random(System.currentTimeMillis());
            long leaseId = Math.abs(rand.nextInt());
            Date lease_start_date = new Date();
            Date lease_end_date = new Date(lease_start_date.getTime() + (2 * 60 * 60 * 1000L)); //2 hour lease

            String usedNodeCompman = /*auth.getUrn().getTopLevelAuthority()*/"omf:nitos.outdoor";
            String usedNodeName = "node001";
            if (fixedNodeUrn != null) {
                usedNodeCompman = fixedNodeUrn.getEncodedTopLevelAuthority();
                usedNodeName = fixedNodeUrn.getEncodedResourceName();
            }

            componentId = " component_id=\"urn:publicid:IDN+" + usedNodeCompman + "+node+" + usedNodeName + "\" ";

            return "<?xml version=\"1.0\"?>\n" +
                    "<rspec type=\"request\" " +
                    "       xmlns=\"http://www.geni.net/resources/rspec/3\" " +
                    "       xmlns:ol=\"http://nitlab.inf.uth.gr/schema/sfa/rspec/1\" " +
                    "       xmlns:omf=\"http://schema.mytestbed.net/sfa/rspec/1\" " +
                    "       xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                    "       xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd http://nitlab.inf.uth.gr/schema/sfa/rspec/1 http://nitlab.inf.uth.gr/schema/sfa/rspec/1/request-reservation.xsd\">\n" +
                    "  <ol:lease client_id=\"l" + leaseId + "\" " +
                    "            valid_from=\"" + RFC3339Util.dateToRFC3339String(lease_start_date) + "\" " +
                    "            valid_until=\"" + RFC3339Util.dateToRFC3339String(lease_end_date) + "\"/>\n" +
                    "  <node " + componentId + " " +
                    "        component_manager_id=\"" + auth.getUrn() + "\" " +
                    "        component_name=\"" + usedNodeName + "\" " +
                    "        exclusive=\"true\" " +
                    "        client_id=\"my_node\">\n" +
                    "    <interface component_id=\"urn:publicid:IDN+" + usedNodeCompman + "+interface+control\" " +
                    "               component_name=\"control\">\n" +
                    "      <ip address=\"10.0.0.1\"/>\n" +
                    "    </interface>\n" +
                    "    <ol:lease_ref id_ref=\"l" + leaseId + "\"/>\n" +
                    "  </node>\n" +
                    "</rspec>";
        }
        if (auth.getUrn().getEncodedTopLevelAuthority().equals("omf:netmode")) {
            Random rand = new Random(System.currentTimeMillis());
            long leaseId = Math.abs(rand.nextInt());
            Date lease_start_date = new Date();
            Date lease_end_date = new Date(lease_start_date.getTime() + (2 * 60 * 60 * 1000L)); //2 hour lease

            String componentName;
            if (fixedNodeUrn == null) {
                componentId =
                        " component_id=\"urn:publicid:IDN+" + auth.getUrn().getEncodedTopLevelAuthority() + "+node+omf.netmode.node1\" ";
                componentName = "omf.netmode.node1";
            } else {
                componentName = fixedNodeUrn.getEncodedResourceName();
            }

            return "<?xml version=\"1.0\"?>\n" +
                    "<rspec type=\"request\" " +
                    "       xmlns=\"http://www.geni.net/resources/rspec/3\" " +
                    "       xmlns:ol=\"http://nitlab.inf.uth.gr/schema/sfa/rspec/1\" " +
                    "       xmlns:omf=\"http://schema.mytestbed.net/sfa/rspec/1\" " +
                    "       xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                    "       xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd http://nitlab.inf.uth.gr/schema/sfa/rspec/1 http://nitlab.inf.uth.gr/schema/sfa/rspec/1/request-reservation.xsd\">\n" +
                    "  <ol:lease client_id=\"l" + leaseId + "\" " +
                    "            valid_from=\"" + RFC3339Util.dateToRFC3339String(lease_start_date) + "\" " +
                    "            valid_until=\"" + RFC3339Util.dateToRFC3339String(lease_end_date) + "\"/>\n" +
                    "  <node " + componentId + " " +
                    "        component_manager_id=\"" + auth.getUrn() + "\" " +
                    "        component_name=\"" + componentName + "\" " +
                    "        exclusive=\"true\" " +
                    "        client_id=\"my_node\">\n" +
                    "    <interface component_id=\"urn:publicid:IDN+" + auth.getUrn().getEncodedTopLevelAuthority() + "+interface+control\" " +
                    "               component_name=\"control\">\n" +
                    "      <ip address=\"10.0.0.1\"/>\n" +
                    "    </interface>\n" +
                    "    <ol:lease_ref id_ref=\"l" + leaseId + "\"/>\n" +
                    "  </node>\n" +
                    "</rspec>";
        }
        if (auth.getUrn().getEncodedTopLevelAuthority().equals("omf")) { //TODO fix this urn and rspec
            Random rand = new Random(System.currentTimeMillis());
            Date lease_start_date = new Date();
            long duration_unknownunit = 6;
            Date lease_end_date = new Date(lease_start_date.getTime() + (duration_unknownunit * 60 * 60 * 1000L)); //?
            return "<?xml version=\"1.0\"?>\n" +
                    "<RSpec type=\"NITOS\" expires=\"2013-10-31T19:30:08Z\" generated=\"2013-10-31T18:30:08Z\">\n" +
                    "  <network name=\"omf\">\n" +
                    "    <node component_manager_id=\"urn:publicid:IDN+omf+authority+cm\" " +
                    "          component_id=\"urn:publicid:IDN+" + auth.getUrn().getEncodedTopLevelAuthority() + "+node+node001\"" +
                    "          component_name=\"node001\" " +
                    "          site_id=\"urn:publicid:IDN+" + auth.getUrn().getEncodedTopLevelAuthority() + "+authority+sa\">\n" +
                    "      <hostname>node001</hostname>\n" +
                    "      <location country=\"unknown\" longitude=\"39.360839\" latitude=\"22.949989\"/>\n" +
                    "      <position_3d x=\"1\" y=\"1\" z=\"5\"/>\n" +
                    "      <exclusive>TRUE</exclusive>\n" +
                    "      <gateway>nitlab.inf.uth.gr</gateway>\n" +
                    "      <granularity>1800</granularity>\n" +
                    "      <hardware_type>orbit</hardware_type>\n" +
                    "      <sliver/>\n" +
                    "    </node>\n" +
                    "    <lease slice_id=\"" + sliceUrn + "\" start_time=\"" + lease_start_date.getTime() + "\" duration=\"" + duration_unknownunit + "\">\n" +
                    "      <node component_id=\"urn:publicid:IDN+" + auth.getUrn().getEncodedTopLevelAuthority() + "+node+node001\"/>\n" +
                    "      </lease>\n" +
                    "  </network>\n" +
                    "</RSpec>";
        }


        if (auth.getType() != null && auth.getType().equals("emulab")) {
            return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \" xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.geni.net/resources/rspec/3\">\n" +
                    "  <node client_id=\"PC\" component_manager_id=\"" + auth.getUrnString() + "\" " + componentId + " exclusive=\"true\">\n" +
                    "    <sliver_type name=\"raw-pc\"/>\n" +
                    "  </node>\n" +
                    "</rspec>\n";
        }
        if (auth.getType() != null && auth.getType().equals("instageni")) {
            return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \" xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.geni.net/resources/rspec/3\">\n" +
                    "  <node client_id=\"PC\" component_manager_id=\"" + auth.getUrnString() + "\" " + componentId + " exclusive=\"false\">\n" +
                    "    <sliver_type name=\"emulab-xen\"/>\n" +
//                    "    <sliver_type name=\"emulab-xen\"/>\n" +
                    "  </node>\n" +
                    "</rspec>\n";
        }
        if (auth.getType() != null && auth.getType().equals("planetlab")) {
            if (fixedNodeUrn == null)
                componentId = " component_id=\"urn:publicid:IDN+ple:unistra+node+planetlab1.u-strasbg.fr\" ";
            return "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n" +
                    "   <rspec type=\"request\" " +
                    " xmlns=\"http://www.geni.net/resources/rspec/3\" >\n" +
                    "       <node client_id=\"node0\" " +
                    componentId +
                    " component_manager_id=\"urn:publicid:IDN+ple:ibbtple+authority+cm\" " +
//                    " component_name=\"planetlab1.u-strasbg.fr\" " +
                    " exclusive=\"false\">\n" +
                    "           <sliver/>\n" +
                    "           <sliver_type name=\"plab-vserver\"/>\n" +
//                    "           <location country=\"unknown\" longitude=\"7.73833\" latitude=\"48.5237\"/>\n" +
                    "           <hardware_type name=\"plab-pc\"/>\n" +
                    "           <hardware_type name=\"pc\"/>\n" +
                    "       </node>\n" +
                    "   </rspec>";
            //OLD
//            return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \" xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.geni.net/resources/rspec/3\">\n" +
//                    "    <node component_id=\"urn:publicid:IDN+ple:unistra+node+planetlab1.u-strasbg.fr\" component_manager_id=\"urn:publicid:IDN+ple+authority+cm\" component_name=\"planetlab1.u-strasbg.fr\" exclusive=\"false\">\n" +
//                    "           <hardware_type name=\"plab-pc\"/>\n" +
//                    "           <hardware_type name=\"pc\"/>\n" +
//                    "           <location country=\"unknown\" latitude=\"48.5237\" longitude=\"7.73833\"/>\n" +
//                    "           <sliver_type name=\"plab-vserver\">\n" +
//                    "           </sliver_type>\n" +
//                    "       </node>\n" +
//                    "   </rspec>";
        }
        if (auth.getType() != null && (auth.getType().equals("orca") || auth.getType().equals("exogeni"))) {
            //see https://geni-orca.renci.org/trac/wiki/orca-and-rspec
            String componentManagerId = getComponentManagerIdForSite(auth);

            return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \" xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.geni.net/resources/rspec/3\">\n" +
                    //                    "  <node client_id=\"PC\" component_manager_id=\""+auth.getUrnString()+"\" exclusive=\"false\">\n" +
                    "  <node client_id=\"PC\" component_manager_id=\"" + componentManagerId + "\" " + componentId + " exclusive=\"false\">\n" +
                    //                    "    <sliver_type name=\"ExoGENI-M4\"/>\n" + //bare metal node
                    "    <sliver_type name=\"xo.small\"/>\n" +   //1 core node (smallest available)
                    "  </node>\n" +
                    "</rspec>\n";
        }

        //otherwise, request a non exclusive node without specifying sliver type
        return "<rspec type=\"request\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \" xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.geni.net/resources/rspec/3\">\n" +
                "  <node client_id=\"PC\" component_manager_id=\"" + auth.getUrnString() + "\" " + componentId + " exclusive=\"false\">\n" +
                "  </node>\n" +
                "</rspec>\n";
    }

    public static boolean isEmptyRspec(ApiTest test, String rspec) {
        if (rspec == null || rspec.trim().isEmpty())
            return true;

        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(rspec));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("rspec")) {
                throw new RuntimeException("Not an rspec (root element tag name=" + rootElName + "): " + rspec);
            }

            NodeList childNodeList = rootEl.getChildNodes();
            for (int i = 0; i < childNodeList.getLength(); i++) {
                Node n = childNodeList.item(i);

                if (n instanceof Element) {
                    Element e = (Element) n;
                    String childName = e.getTagName();
                    if (childName.equalsIgnoreCase("node"))
                        return false;
                    if (childName.equalsIgnoreCase("link"))
                        return false;

                    //hack for PLE
                    if (childName.equalsIgnoreCase("network"))
                        return false;
                }
            }
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException while testing if rspec is empty: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException while testing if rspec is empty: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException while testing if rspec is empty: " + e.getMessage(), e);
        }
        return true;
    }

    /**
     * returns nodecount
     */
    public static int testValidGeni3ManifestRspec(ApiTest test, String rspec) {
        if (rspec == null) return -1;
        test.assertNotNull(rspec, "Rspec is null");

        int nodeCount = 0;

        //quick HACK to allow the planetlab Europe rspec manifest
        // FEDIBBTDEV-113, bvermeul, copied this nice stuff from suite/rspec-edit/src/main/java/be/iminds/ilabt/jfed/ui/rspeceditor/model/Rspec.java
        if (rspec.contains("<RSpec")) {
            test.note("We received an SFA RSpec instead of a GENI RSpec v3, but we'll try to work around.");
            //logger.debug("HACK: rewriting received planetlab manifest XML to make it parsable xml");
            rspec =
                    rspec.replace("<RSpec type=\"SFA\"", "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" type=\"manifest\"");
            rspec =
                    rspec.replace("<RSpec ", "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" type=\"manifest\"");
            rspec = rspec.replace("</RSpec>", "</rspec>");
            rspec = rspec.replace("<network name=\"ple\">", "");
            rspec = rspec.replace("</network>", "");
            //logger.debug("HACK: rewriten xml: "+rspec);
        }


        //parse the RSpec XML
        try {
            Class docClass = be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.RSpecContents.class;
            String packageName = docClass.getPackage().getName();
            JAXBContext jc = JAXBContext.newInstance(packageName);
            Unmarshaller u = jc.createUnmarshaller();
            JAXBElement<RSpecContents> doc =
                    (JAXBElement<be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.RSpecContents>) u.unmarshal(new StringReader(rspec));
            be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.RSpecContents c = doc.getValue();

            be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.RspecTypeContents typ = c.getType();
            test.assertNotNull(typ, "Received manifest RSpec does not specify a type: " + rspec);
            test.assertEquals(typ.value(), "manifest", "Received manifest RSpec is not a manifest: " + rspec);

            for (Object o : c.getAnyOrNodeOrLink()) {
                if (o instanceof JAXBElement) {
                    JAXBElement el = (JAXBElement) o;
                    if (el.getValue() instanceof be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.NodeContents) {
                        be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.NodeContents node =
                                (be.iminds.ilabt.jfed.rspec.manifest.geni_rspec_3.NodeContents) el.getValue();

                        String nodeName = node.getClientId();
                        nodeCount++;

                        for (Object nodeElO : node.getAnyOrRelationOrLocation()) {
                            if (nodeElO instanceof JAXBElement) {
                                JAXBElement nodeEl = (JAXBElement) nodeElO;
                                //                                if (nodeEl.getValue() instanceof InterfaceContents) {
                                //                                    InterfaceContents ic = (InterfaceContents) nodeEl.getValue();
                                //                                    String interfaceName = ic.getClientId();
                                //                                }
                            }
                        }
                    }
                    //                    if (el.getValue() instanceof LinkContents) {
                    //                        LinkContents link = (LinkContents) el.getValue();
                    //
                    //                        String linkname = link.getClientId();
                    //
                    //                        for (Object linkElO : link.getAnyOrPropertyOrLinkType()) {
                    //                            if (linkElO instanceof JAXBElement) {
                    //                                JAXBElement linkEl = (JAXBElement) linkElO;
                    //                                if (linkEl.getValue() instanceof InterfaceRefContents) {
                    //                                    InterfaceRefContents ic = (InterfaceRefContents) linkEl.getValue();
                    //                                    String interfaceName = ic.getClientId();
                    //                                }
                    //                            }
                    //                        }
                    //                    }
                }
            }
        } catch (JAXBException e) {
            throw new RuntimeException("Exception parsing manifest RSpec Xml: " + e.getMessage(), e);
        }

        //        assertTrue(nodeCount > 0, "No nodes found in manifest rspec: "+rspec);

        return nodeCount;
    }

    /**
     * returns nodecount
     */
    public static int testValidGeni3AdvertisementRspec(ApiTest test, String rspec) {
        if (rspec == null) return -1;
        test.assertNotNull(rspec, "Rspec is null");

        int nodeCount = 0;

        //parse the RSpec XML
        try {
            Class docClass = be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.RSpecContents.class;
            String packageName = docClass.getPackage().getName();
            JAXBContext jc = JAXBContext.newInstance(packageName);
            Unmarshaller u = jc.createUnmarshaller();
            JAXBElement<be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.RSpecContents> doc =
                    (JAXBElement<be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.RSpecContents>) u.unmarshal(new StringReader(rspec));
            be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.RSpecContents c = doc.getValue();

            String typ = c.getType();
            test.assertNotNull(typ, "Received advertisement RSpec does not specify a type: " + rspec);
            test.assertEquals(typ, "advertisement", "Received advertisement RSpec is not a advertisement, it is a: \"" + typ + "\"");

            for (Object o : c.getAnyOrNodeOrLink()) {
                if (o instanceof JAXBElement) {
                    JAXBElement el = (JAXBElement) o;
                    if (el.getValue() instanceof be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.NodeContents) {
                        be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.NodeContents node =
                                (be.iminds.ilabt.jfed.rspec.advertisement.geni_rspec_3.NodeContents) el.getValue();

                        nodeCount++;

                        for (Object nodeElO : node.getAnyOrRelationOrLocation()) {
                            if (nodeElO instanceof JAXBElement) {
                                JAXBElement nodeEl = (JAXBElement) nodeElO;
                                //                                if (nodeEl.getValue() instanceof InterfaceContents) {
                                //                                    InterfaceContents ic = (InterfaceContents) nodeEl.getValue();
                                //                                    String interfaceName = ic.getClientId();
                                //                                }
                            }
                        }
                    }
                    //                    if (el.getValue() instanceof LinkContents) {
                    //                        LinkContents link = (LinkContents) el.getValue();
                    //
                    //                        String linkname = link.getClientId();
                    //
                    //                        for (Object linkElO : link.getAnyOrPropertyOrLinkType()) {
                    //                            if (linkElO instanceof JAXBElement) {
                    //                                JAXBElement linkEl = (JAXBElement) linkElO;
                    //                                if (linkEl.getValue() instanceof InterfaceRefContents) {
                    //                                    InterfaceRefContents ic = (InterfaceRefContents) linkEl.getValue();
                    //                                    String interfaceName = ic.getClientId();
                    //                                }
                    //                            }
                    //                        }
                    //                    }
                }
            }
        } catch (JAXBException e) {
            throw new RuntimeException("Exception parsing advertisement RSpec Xml: " + e.getMessage(), e);
        }

        //        assertTrue(nodeCount > 0, "No nodes found in advertisement rspec: "+rspec);

        return nodeCount;
    }

    public static String createSliceName(ApiTest test, SfaAuthority userAuthority, List<GeniUrn> testedAuthorityCms, String prefix) {
        int maxSliceNameLength = 19; //Geni max length

        int maxTestedAuthorityCmLen = testedAuthorityCms.get(0).getEncodedTopLevelAuthority().length();
        String longestCmUrn = null;
        for (GeniUrn u : testedAuthorityCms)
                if (u.getEncodedTopLevelAuthority().length() >= maxTestedAuthorityCmLen) {
                    maxTestedAuthorityCmLen = u.getEncodedTopLevelAuthority().length();
                    longestCmUrn = u.getEncodedTopLevelAuthority();
                }
        test.note("The longest CM urn top level authority is: \""+longestCmUrn+"\" ("+maxTestedAuthorityCmLen+" chars)");

        //planetlab sfa max length
        int authLen = userAuthority.getNameForUrn().replaceAll("\\.", "").length();
        int planetlab_l = 32 - authLen;
        //protogeni error example:            Hostname > 63 characters: PC.sfbD481228.wall2-ilabt-iminds-be.instageni.metrodatacenter.com
        //protogeni max length of FQDN = 64 characters (unix limit):  PC.st1383985554.wall2-ilabt-iminds-be.instageni.rnet.missouri.edu
        final int dotcount = 3;
        final int nodeNameLen = 3;  // we use 3 for hostnames (pc1,pc2 in stitching) (and 1 for the .)
        int protogeni_l = 63 - maxTestedAuthorityCmLen - userAuthority.getNameForUrn().length() - nodeNameLen - dotcount;
        int l = 0;
        if (protogeni_l < planetlab_l) {
            l = protogeni_l;
        } else {
            l = planetlab_l;
        }
        if (l < maxSliceNameLength) {
            maxSliceNameLength = l;
            test.note("For " + userAuthority.getNameForUrn() + " (" + authLen + " non-dot chars), the maximum slice name length (taking into account planetlab and protogeni limits) is " + maxSliceNameLength);
        } else
            test.note("Slice name length is " + maxSliceNameLength);

//        long uniqueNumber = System.currentTimeMillis() / 1000; //unique each second
//        // uniqueNumber /= 10; //unique each 10 seconds
//        String uniqueNumberString = uniqueNumber + "";
//        int uniqueNumberMaxLen = maxSliceNameLength - prefix.length();
//        uniqueNumberMaxLen -= 3; //for uniqueText
//        if (uniqueNumberString.length() > uniqueNumberMaxLen) {
//            //drop from front of string, not from back
//            uniqueNumberString = uniqueNumberString.substring(uniqueNumberString.length() - uniqueNumberMaxLen);
//            test.assertEquals(uniqueNumberString.length(), uniqueNumberMaxLen, "bug in test: uniqueNumberString.length()=" + uniqueNumberString.length() + " !=  uniqueNumberMaxLen=" + uniqueNumberMaxLen);
//
//            test.note("Slice unique number is " + uniqueNumber + " after making it shorter it, it is " + uniqueNumberString);
//        }

        //just use random text.
        int uniqueTextLen = maxSliceNameLength - prefix.length();

        if (uniqueTextLen < 2) {
            uniqueTextLen = 2;
            test.note("The calculated slice name length is " + maxSliceNameLength+" which is too short. It will be increased to "+(prefix.length()+uniqueTextLen)+" create a usable slice name for this test.\n" +
                    "This slicename consists of a prefix (\""+prefix+"\" = "+prefix.length()+" chars) and a unique suffix ("+uniqueTextLen+" random chars)");
        }

        String uniqueTextHelper = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
        String uniqueText = "";
        for (int i = 0; i < uniqueTextLen; i++) {
            int pos = random.nextInt(uniqueTextHelper.length());
            uniqueText += uniqueTextHelper.charAt(pos);
        }

//        String longSliceName = prefix + uniqueText + uniqueNumberString;
        String longSliceName = prefix + uniqueText;
        return longSliceName;
    }

    public static void addOptionsForNodeLoginSshProxy(List<String> res) {
        //proxy for ssh connections
        res.add("loginSshProxyMode"); //valid option: off (default), manual(use preferences below), user(automatically use proxy of user authority)

        res.add("loginSshProxyHostName");
        res.add("loginSshProxyHostKey");
        res.add("loginSshProxyHostPort");
        res.add("loginSshProxyUsername"); //default: same username as logged in user (same as "$USER"). Some limited variable substitution is done:
        // - $NODE: username for target node
        // - $USER: logged in user name (last part of user URN)
        // - $SLICE: slice name (last part of slice URN)
        res.add("loginSshProxyKeySource"); //valid choices:
        // - user (take SSH keys from user login cert and key)
        // - node (use the same SSH keys as the one to login to the actual target node, these are typically generated by NodeLoginTester, unless fixed_ssh_public_key_file says otherwise)
        // - file: take from file name specified in sshProxyPrivateKeyFile and sshProxyPublicKeyFile option (not yet implemented)
    }

    public static void addOptionsForConnectionSshProxy(List<String> res) {
        res.add("sfaSshProxyMode"); //valid option: off (default), manual(use preferences below), user(automatically use proxy of user authority)

        //proxy for sfa connections
        res.add("sfaSshProxyHostName");
        res.add("sfaSshProxyHostKey");
        res.add("sfaSshProxyHostPort");
        res.add("sfaSshProxyUsername"); //default: same username as logged in user (same as "$USER"). Some limited variable substitution is done:
        // - $USER: logged in user name (last part of user URN)
        res.add("sfaSshProxyKeySource"); //valid choices:
        // - user (take SSH keys from user login cert and key)
        // - file: take from file name specified in sshProxyPrivateKeyFile and sshProxyPublicKeyFile option (not yet implemented)

    }
    public static void addOptionsForRspecTypeVersion(List<String> res) {
        res.add("rspecTypeName"); //default: "geni"
        res.add("rspecTypeVersion"); //default: "3"
    }

    public String getRspecTypeName(Properties testConfig) {
        String res = TextUtil.objectToStringOrNull(testConfig.get("rspecTypeName"), true);
        if (res == null)
            return "geni";
        return res;
    }
    public String getRspecTypeVersion(Properties testConfig) {
        String res = TextUtil.objectToStringOrNull(testConfig.get("rspecTypeVersion"), true);
        if (res == null)
            return "3";
        return res;
    }
    public boolean isRspecTypeGeni3(Properties testConfig) {
        String name = TextUtil.objectToStringOrNull(testConfig.get("rspecTypeName"), true);
        String ver = TextUtil.objectToStringOrNull(testConfig.get("rspecTypeVersion"), true);
        return (name == null || name.equalsIgnoreCase("geni")) && (ver == null || ver.equals("3"));
    }

    /**
     * returns true if A is preferred over B
     */
    public static boolean preferSliverType(String sliverTypeNameA, String sliverTypeNameB) {
        if (sliverTypeNameB == null) return true;
        int scoreA = getSliverTypeNameScore(sliverTypeNameA);
        int scoreB = getSliverTypeNameScore(sliverTypeNameB);
        return scoreA > scoreB;
    }

    /**
     * returns a score for the sliver type name. Higher scores are preferred.
     */
    public static int getSliverTypeNameScore(String sliverTypeName) {
        if (sliverTypeName == null) return 0;
        if (sliverTypeName.equalsIgnoreCase("emulab-openvz")) return 100;
        if (sliverTypeName.equalsIgnoreCase("emulab-xen")) return 98;
        if (sliverTypeName.equalsIgnoreCase("plab-vserver")) return 99;

        if (sliverTypeName.equalsIgnoreCase("raw-pc")) return 90;

        //exogeni
        if (sliverTypeName.equalsIgnoreCase("xo.small")) return 80;
        if (sliverTypeName.equalsIgnoreCase("xo.tiny")) return 81;
        if (sliverTypeName.startsWith("xo.")) return 79;

        //various
        if (sliverTypeName.contains("xen")) return 63;
        if (sliverTypeName.contains("openvz")) return 62;
        if (sliverTypeName.contains("virt")) return 61;
        if (sliverTypeName.contains("vz")) return 60;

        if (sliverTypeName.contains("raw")) return 11;
        if (sliverTypeName.contains("pc")) return 10;

        return 1;
    }

    public List<AnyCredential> getUser2Credentials() {
        assert test != null;
        assert secondUser != null;

        if (user2Credentials != null)
            return user2Credentials;

        try {
            user2Credentials =
                    userAndSliceApiWrapperUser2.getUserCredentials(test.getLogger(), GeniUrn.parse(secondUser.getUserUrnString()));
        } catch (JFedException e) {
            throw new RuntimeException(e);
        }
        assert user2Credentials != null;
        assert !user2Credentials.isEmpty();
        return user2Credentials;
    }

    public List<AnyCredential> getUserCredentials() {
        assert test != null;
        assert userAuthority != null;

        if (userCredentials != null && !userCredentials.isEmpty())
            return userCredentials;

        try {
            userCredentials =
                    userAndSliceApiWrapperUser1.getUserCredentials(test.getLogger(), GeniUrn.parse(test.getUser().getUserUrnString()));
        } catch (JFedException e) {
            throw new RuntimeException(e);
        }
        assert userCredentials != null;
        assert !userCredentials.isEmpty();
        return userCredentials;
    }

    public List<AnyCredential> getUserCredentialList() {
        if (testCredentialType == TestCredentialType.REGULAR)
            return getUserCredentials();

        if (testCredentialType == TestCredentialType.DELEGATION)
            return AnyCredential.toCredentialList(getDelegatedUserCredential());

        assert testCredentialType == TestCredentialType.SPEAKSFOR;
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        res.addAll(getUser2Credentials());
        res.add(getSpeaksForUserCredential());
        return res;
    }

    public List<AnyCredential> getSliceCredentialList(SliceInfo sliceInfo) {
        if (testCredentialType == TestCredentialType.REGULAR)
            return sliceInfo.credentials;

        if (testCredentialType == TestCredentialType.DELEGATION)
            return AnyCredential.toCredentialList(sliceInfo.getDelegatedCredential(test.getUser(), secondUser));

        assert testCredentialType == TestCredentialType.SPEAKSFOR;
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        res.addAll(sliceInfo.credentials);
        res.add(getSpeaksForUserCredential());
        return res;
    }

    public void setSpeaksForOptionNames(List<String> newSpeaksForOptionNames) {
        speaksForOptionNames.clear();
        speaksForOptionNames.addAll(newSpeaksForOptionNames);
    }

    /**
     * add speaksfor option if needed
     *
     * @param extraOptions does not change original extraOptions, makes a copy. allowed to be null.
     */
    public Hashtable addCredentialExtraOptions(Hashtable extraOptions) {
        if (testCredentialType != TestCredentialType.SPEAKSFOR)
            return extraOptions;

        assert secondUser != null;

        Hashtable res = new Hashtable();
        if (extraOptions != null) res.putAll(extraOptions);

        for (String speaksForOptionName : speaksForOptionNames)
            res.put(speaksForOptionName, secondUser.getUserUrnString());

        return res;
    }

    public void setupProxyForConnectionPool(Properties testConfig) {
        String mode;
        if (testConfig.get("sfaSshProxyMode") != null) {
            mode = testConfig.getProperty("sfaSshProxyMode");
            if (!mode.equals("manual") && !mode.equals("user"))
                return;
        } else
            return;

        if (mode.equals("manual") && testConfig.get("sfaSshProxyHostName") != null) {
            String sfaSshProxyHostName = (String) testConfig.get("sfaSshProxyHostName");
            int sfaSshProxyHostPort =
                    Integer.parseInt(testConfig.get("sfaSshProxyHostPort") + ""); //TODO: does Properties store an Integer or a String? (this code works in both cases)
            String sfaSshProxyHostKey = (String) testConfig.get("sfaSshProxyHostKey");

            String username;
            boolean limitName;
            if (testConfig.get("sfaSshProxyUsername") != null) {
                String sfaSshProxyUsername = (String) testConfig.get("sfaSshProxyUsername");
                String loggedInUserName = test.getUser().getUserUrn().getEncodedResourceName();

                username = sfaSshProxyUsername;
                limitName = username.contains("$USER");
                username = username.replace("$USER", loggedInUserName);
            } else {
                limitName = true;
                username = test.getUser().getUserUrn().getEncodedResourceName();
            }

            String sfaSshProxyKeySource = (String) testConfig.get("sfaSshProxyKeySource");
            SshKeyInfo sshKeyInfo = null;
            if (sfaSshProxyKeySource.equalsIgnoreCase("user")) {
                test.note("Using user private key as proxy private key.");
                sshKeyInfo = new BasicSshKeyInfo(
                        test.getUser().getPublicKey(),
                        test.getUser().getPrivateKey());
            }
            if (sfaSshProxyKeySource.equalsIgnoreCase("file")) {
                test.warn("Configuration warning: Test option sfaSshProxyKeySource=\"" + sfaSshProxyKeySource + "\" is not implemented yet. Cannot continue, will do test WITHOUT proxy.");
            }

            if (sshKeyInfo == null) {
                test.warn("Configuration error: Test option sfaSshProxyKeySource=\"" + sfaSshProxyKeySource + "\" is not known. Cannot continue, will do test WITHOUT proxy.");
            } else {
                JFedConnection.SshProxyInfo sfaSshProxyInfo = new JFedConnection.SshProxyInfo(
                        sfaSshProxyHostName,
                        sfaSshProxyHostPort,
                        username,
                        sshKeyInfo,
                        sfaSshProxyHostKey,
                        limitName);
                test.note("Using SSH proxy for SSH login test: " + sfaSshProxyInfo);
                ((SfaConnectionPool) test.getConnectionProvider()).setDefaultProxy(sfaSshProxyInfo);
            }
            return;
        }

        if (mode.equals("user")) {
            GeniUser user = test.getUser();
            SfaAuthority userAuth = user.getUserAuthority();

            if (userAuth.getProxies().isEmpty()) {
                test.warn("This logged in user's authority does not have a proxy that jFed can use.");
                return;
            }

            SfaAuthority.ProxyInfo authProxyInfo = userAuth.getProxies().get(0);

            VlanRange portRange = new VlanRange(authProxyInfo.getPortRange());
            Integer port = portRange.getFirst();

//            while (port != null && unreachableProxies.contains(new ProxyInfo(authProxyInfo.getHostname(), port))) {
//                port = portRange.getNext();
//            }
//            if (port == null) {
//                LOG.debug("Proxy at " + authProxyInfo.getHostname() + " has no reachable ports, skipping.");
//                continue;
//            }
//            LOG.info("SSH Proxy is activated, and there is a reachable SSH Proxy for user authority " + userAuth.getName() + ": " + authProxyInfo.getHostname() + ":" + port);

            SshKeyInfo userSshKeyInfo = new BasicSshKeyInfo(
                    test.getUser().getPublicKey(),
                    test.getUser().getPrivateKey());

            JFedConnection.ProxyInfo proxyInfo = new JFedConnection.SshProxyInfo(
                    authProxyInfo.getHostname(),
                    port,
                    user.getUserUrn().getEncodedResourceName(),
                    userSshKeyInfo,
                    authProxyInfo.getHostKey(),
                    true/*limitName*/);

            ((SfaConnectionPool) test.getConnectionProvider()).setDefaultProxy(proxyInfo);
            return;
        }
    }

    public void setupProxyForNodeLogin(Properties testConfig, NodeLoginTester nodeLoginTester, GeniUrn sliceUrn) {
        //ugly cut and paste code for backward compatibility. May be removed later!
        //    (this ignores loginSshProxyMode when sshProxyHostName property is found. This is the old property, and it is not advertised anymore, so only old configs still use this)
        if (testConfig.get("sshProxyHostName") != null) {
            boolean limitName;
            String sshProxyHostName = (String) testConfig.get("sshProxyHostName");
            int sshProxyHostPort = Integer.parseInt(testConfig.get("sshProxyHostPort") + "");
            String sshProxyHostKey = (String) testConfig.get("sshProxyHostKey");

            String username;
            if (testConfig.get("sshProxyUsername") != null) {
                String sshProxyUsername = (String) testConfig.get("sshProxyUsername");
                String loggedInUserName = test.getUser().getUserUrn().getEncodedResourceName();
                String sliceName = sliceUrn.getEncodedResourceName();
                String nodeUsername = nodeLoginTester.getSshUsername();

                username = sshProxyUsername;
                limitName = username.contains("$USER") || username.contains("$NODE") || username.contains("$SLICE");
                username = username.replace("$USER", loggedInUserName);
                username = username.replace("$NODE", nodeUsername);
                username = username.replace("$SLICE", sliceName);
            } else {
                limitName = true;
                username = test.getUser().getUserUrn().getEncodedResourceName();
            }

            String sshProxyKeySource = (String) testConfig.get("sshProxyKeySource");
            SshKeyInfo sshKeyInfo = null;
            if (sshProxyKeySource.equalsIgnoreCase("user")) {
                test.note("Using user private key as proxy private key.");
                sshKeyInfo = new BasicSshKeyInfo(
                        test.getUser().getPublicKey(),
                        test.getUser().getPrivateKey());
            }
            if (sshProxyKeySource.equalsIgnoreCase("node")) {
                test.note("Using node private key as proxy private key.");
                sshKeyInfo = new BasicSshKeyInfo(
                        nodeLoginTester.getSshKeyHelper().getSshPublicKey(),
                        nodeLoginTester.getSshKeyHelper().getSshPrivateKey());
            }
            if (sshProxyKeySource.equalsIgnoreCase("file")) {
                test.warn("Configuration warning: Test option sshProxyKeySource=\"" + sshProxyKeySource + "\" is not implemented yet. Cannot continue, will do test WITHOUT proxy.");
            }

            if (sshKeyInfo == null) {
                test.warn("Configuration error: Test option sshProxyKeySource=\"" + sshProxyKeySource + "\" is not unknown. Cannot continue, will do test WITHOUT proxy.");
            } else {
                JFedConnection.SshProxyInfo sshProxyInfo = new JFedConnection.SshProxyInfo(
                        sshProxyHostName,
                        sshProxyHostPort,
                        username,
                        sshKeyInfo,
                        sshProxyHostKey,
                        limitName);
                test.note("Using SSH proxy for SSH login test: " + sshProxyInfo);

                nodeLoginTester.setSshProxy(sshProxyInfo);
            }
        }


        //normal (new version) code:

        String mode;
        if (testConfig.get("loginSshProxyMode") != null) {
            mode = testConfig.getProperty("loginSshProxyMode");
            if (!mode.equals("manual") && !mode.equals("user"))
                return;
        } else
            return;

        if (mode.equals("manual") && testConfig.get("loginSshProxyHostName") != null) {
            String loginSshProxyHostName = (String) testConfig.get("loginSshProxyHostName");
            int loginSshProxyHostPort =
                    Integer.parseInt(testConfig.get("loginSshProxyHostPort") + ""); //TODO: does Properties store an Integer or a String? (this code works in both cases)
            String loginSshProxyHostKey = (String) testConfig.get("loginSshProxyHostKey");

            boolean limitName;
            String username;
            if (testConfig.get("loginSshProxyUsername") != null) {
                String loginSshProxyUsername = (String) testConfig.get("loginSshProxyUsername");
                String loggedInUserName = test.getUser().getUserUrn().getEncodedResourceName();
                String sliceName = sliceUrn.getEncodedResourceName();
                String nodeUsername = nodeLoginTester.getSshUsername();

                username = loginSshProxyUsername;
                limitName = username.contains("$USER") || username.contains("$NODE") || username.contains("$SLICE");
                username = username.replace("$USER", loggedInUserName);
                username = username.replace("$NODE", nodeUsername);
                username = username.replace("$SLICE", sliceName);
            } else {
                limitName = true;
                username = test.getUser().getUserUrn().getEncodedResourceName();
            }

            String loginSshProxyKeySource = (String) testConfig.get("loginSshProxyKeySource");
            SshKeyInfo sshKeyInfo = null;
            if (loginSshProxyKeySource.equalsIgnoreCase("user")) {
                test.note("Using user private key as proxy private key.");
                sshKeyInfo = new BasicSshKeyInfo(
                        test.getUser().getPublicKey(),
                        test.getUser().getPrivateKey());
            }
            if (loginSshProxyKeySource.equalsIgnoreCase("node")) {
                test.note("Using node private key as proxy private key.");
                sshKeyInfo = new BasicSshKeyInfo(
                        nodeLoginTester.getSshKeyHelper().getSshPublicKey(),
                        nodeLoginTester.getSshKeyHelper().getSshPrivateKey());
            }
            if (loginSshProxyKeySource.equalsIgnoreCase("file")) {
                test.warn("Configuration warning: Test option loginSshProxyKeySource=\"" + loginSshProxyKeySource + "\" is not implemented yet. Cannot continue, will do test WITHOUT proxy.");
            }

            if (sshKeyInfo == null) {
                test.warn("Configuration error: Test option loginSshProxyKeySource=\"" + loginSshProxyKeySource + "\" is not unknown. Cannot continue, will do test WITHOUT proxy.");
            } else {
                JFedConnection.SshProxyInfo loginSshProxyInfo = new JFedConnection.SshProxyInfo(
                        loginSshProxyHostName,
                        loginSshProxyHostPort,
                        username,
                        sshKeyInfo,
                        loginSshProxyHostKey,
                        limitName);
                test.note("Using SSH proxy for SSH login test: " + loginSshProxyInfo);

                nodeLoginTester.setSshProxy(loginSshProxyInfo);
            }
        }

        if (mode.equals("user")) {
            GeniUser user = test.getUser();
            SfaAuthority userAuth = user.getUserAuthority();

            if (userAuth.getProxies().isEmpty()) {
                test.warn("This logged in user's authority does not have a proxy that jFed can use.");
                return;
            }

            SfaAuthority.ProxyInfo authProxyInfo = userAuth.getProxies().get(0);

            VlanRange portRange = new VlanRange(authProxyInfo.getPortRange());
            Integer port = portRange.getFirst();

//            while (port != null && unreachableProxies.contains(new ProxyInfo(authProxyInfo.getHostname(), port))) {
//                port = portRange.getNext();
//            }
//            if (port == null) {
//                LOG.debug("Proxy at " + authProxyInfo.getHostname() + " has no reachable ports, skipping.");
//                continue;
//            }
//            LOG.info("SSH Proxy is activated, and there is a reachable SSH Proxy for user authority " + userAuth.getName() + ": " + authProxyInfo.getHostname() + ":" + port);

            SshKeyInfo userSshKeyInfo = new BasicSshKeyInfo(
                    test.getUser().getPublicKey(),
                    test.getUser().getPrivateKey());

            JFedConnection.SshProxyInfo proxyInfo = new JFedConnection.SshProxyInfo(
                    authProxyInfo.getHostname(),
                    port,
                    user.getUserUrn().getEncodedResourceName(),
                    userSshKeyInfo,
                    authProxyInfo.getHostKey(),
                    true);

            nodeLoginTester.setSshProxy(proxyInfo);
            return;
        }
    }

    public void selectRegularTestCredentialType(ApiTest test) {
        testCredentialType = TestCredentialType.REGULAR;
        this.secondUser = null;
        delegatedUserCredential = null;
        speaksForUserCredential = null;
        user2Credentials = null;
        for (SliceInfo sliceInfo : allSliceInfo)
            sliceInfo.delegatedCredential = null;

        userAndSliceApiWrapperUser2 = null;
    }

    /**
     * user 1 is speaking for user 2
     */
    public void selectSpeaksForTestCredentialType(ApiTest test, GeniUser secondUser) {
        testCredentialType = TestCredentialType.SPEAKSFOR;
        this.secondUser = secondUser;
        delegatedUserCredential = null;
        speaksForUserCredential = null;
        user2Credentials = null;
        for (SliceInfo sliceInfo : allSliceInfo)
            sliceInfo.delegatedCredential = null;

        userAndSliceApiWrapperUser2 =
                test.getAutomaticUserAndSliceApiWrapperFactory().create();

        test.note("User 2=" + secondUser.getUserUrnString() + " delegates rights to user 1=" + test.getUser().getUserUrnString() + "");
    }

    /**
     * user 1 uses delegated credential from user 2 (so user 2's rights)
     */
    public void selectDelegationTestCredentialType(ApiTest test, GeniUser secondUser) {
        testCredentialType = TestCredentialType.DELEGATION;
        this.secondUser = secondUser;
        delegatedUserCredential = null;
        speaksForUserCredential = null;
        user2Credentials = null;
        for (SliceInfo sliceInfo : allSliceInfo)
            sliceInfo.delegatedCredential = null;

        userAndSliceApiWrapperUser2 =
                test.getAutomaticUserAndSliceApiWrapperFactory().create(secondUser);

        test.note("User 1=" + test.getUser().getUserUrnString() + " speaks for user 2=" + secondUser.getUserUrnString() + "");
    }

    public SfaCredential getDelegatedUserCredential() {
        if (delegatedUserCredential != null) return delegatedUserCredential;
//        if (userCredential == null) getUserCredential();
//        if (userCredential == null) throw new RuntimeException("User credential is null");
        if (user2Credentials == null || user2Credentials.isEmpty()) getUser2Credentials();
        if (user2Credentials == null || user2Credentials.isEmpty())
            throw new RuntimeException("User credential is null");

        GeniUser user1 = test.getUser();
        GeniUser user2 = secondUser;

        SfaCredential sfaUser2Credential = null;
        for (AnyCredential user2Credential : user2Credentials)
            if (user2Credential instanceof SfaCredential)
                sfaUser2Credential = (SfaCredential) user2Credential;
        if (sfaUser2Credential == null) {
            throw new RuntimeException("No SFA user 2 credential available: " + user2Credentials);
        }
        try {
            delegatedUserCredential = sfaUser2Credential.delegate(
                    user1.getUserUrnString()/*newOwnerUrn*/,
                    user1.getClientCertificateChain().get(0)/*newOwnerCert*/,
                    user2.getPrivateKey()/*originalOwnerPrivateKey*/,
                    new Date(System.currentTimeMillis() + 24 * 60 * 60 * 1000)/*expireDate*/,
                    "*"/*delegatedRights*/,
                    false/*canDelegate*/);
        } catch (CredentialException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
        assert delegatedUserCredential != null;
        return delegatedUserCredential;
    }

    public SfaCredential getSpeaksForUserCredential() {
        if (speaksForUserCredential != null) return speaksForUserCredential;
//        if (user2Credential == null) getUser2Credential();
//        if (user2Credential == null) throw new RuntimeException("User credential is null");

        GeniUser user1 = test.getUser();
        GeniUser user2 = secondUser;

        try {
            speaksForUserCredential = SfaCredential.createSpeaksFor(
                    user2.getUserUrnString()/*spokenForUrn*/,
                    user1.getUserUrnString()/*speakerUrn*/,
                    user2.getClientCertificateChain().get(0)/*spokenForCert*/,
                    user1.getClientCertificateChain().get(0)/*speakerCert*/,
                    user2.getPrivateKey()/*spokenForPrivateKey*/,
                    new Date(System.currentTimeMillis() + 24 * 60 * 60 * 1000)/*expireDate*/,
                    "*"/*privilegeName*/,
                    false/*canDelegate*/);
        } catch (CredentialException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
        assert speaksForUserCredential != null;
        return speaksForUserCredential;
    }

    /**
     * @param prefix             a prefix for the generated unique slice name
     * @param project            the project/subauthroity, may be null
     * @param preferUsingProject even if not specified in "project" and not mandatory, try to use a project. But don't use one if none can be found.
     * @return
     * @throws JFedException
     */
    public SliceInfo createSlice(String prefix, String project, boolean preferUsingProject) throws JFedException {
        return createSlice(prefix, true/*createName*/, project, preferUsingProject);
    }

    /**
     * @param prefixOrName
     * @param createName
     * @param project            the project/subauthroity, may be null
     * @param preferUsingProject even if not specified in "project" and not mandatory, try to use a project. But don't use one if none can be found.
     * @return
     * @throws JFedException
     */
    public SliceInfo createSlice(String prefixOrName, boolean createName, String project, boolean preferUsingProject) throws JFedException {
        boolean callAsUser1 = testCredentialType.equals(TestCredentialType.REGULAR);
        assert callAsUser1 != testCredentialType.equals(TestCredentialType.SPEAKSFOR) || testCredentialType.equals(TestCredentialType.DELEGATION);

        Date expirationDate = new Date(System.currentTimeMillis() + (1 * 60 * 60 * 100L)); //1h in the future

        SliceInfo res = new SliceInfo();
        int tries = 0;
        boolean success = false;
        while (!success) {
            if (createName)
                res.name = CommonAMTest.createSliceName(test, userAuthority, SfaAuthority.authoritiesToUrns(testedAuthoritiesForRspec), prefixOrName);
            else
                res.name = prefixOrName;
            res.urn =
                    GeniUrn.createGeniUrnFromEncodedParts(userAuthority.getNameForUrn(), "slice", res.name);  //not always correct, but overwritten below
            res.urnString = res.urn.toString();


            try {
                if (callAsUser1) {
                    GeniUser user1 = test.getUser();

                    UserAndSliceApiWrapper.SubAuthoritySupport user1SubAuthoritySupport =
                            userAndSliceApiWrapperUser1.getSubAuthoritySupport(test.getLogger());

                    if (!userAndSliceApiWrapperUser1.hasUserCredentials())
                        userAndSliceApiWrapperUser1.getUserCredentials(test.getLogger(), user1.getUserUrn());

                    UserAndSliceApiWrapper.SliceInfo sliceInfo;
                    if (user1SubAuthoritySupport.equals(UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY) || preferUsingProject) {
                        String subAuthorityName;
                        if (project == null) {
                            List<String> subAuthorityNames;
                            try {
                                subAuthorityNames =
                                        userAndSliceApiWrapperUser1.getSubAuthorityNames(test.getLogger(), test.getUser().getUserUrn());
                            } catch (Throwable t) {
                                LOG.warn("Exception while getting sub authority names. will try to ignore.", t);
                                test.note("Exception while getting sub authority names. will try to ignore. Exception: " + t.getMessage());
                                subAuthorityNames = new ArrayList<String>();
                            }

                            if (subAuthorityNames.isEmpty()) {
                                if (user1SubAuthoritySupport.equals(UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY))
                                    throw new RuntimeException("There are no sub authorities for user " + user1.getUserUrn());
                                else
                                    subAuthorityName = project;
                            } else {
                                subAuthorityName = subAuthorityNames.get(0);
                            }
                        } else
                            subAuthorityName = project;

                        sliceInfo =
                                userAndSliceApiWrapperUser1.createSlice(test.getLogger(), res.name, expirationDate, subAuthorityName);
                    } else {
                        String subAuthorityName = project; //may be null
                        sliceInfo =
                                userAndSliceApiWrapperUser1.createSlice(test.getLogger(), res.name, expirationDate, subAuthorityName);
                    }
                    res.credentials = sliceInfo.getCredentials();
                    res.urn = sliceInfo.getUrn();
                    res.urnString = res.urn.toString();
                } else {
                    UserAndSliceApiWrapper.SubAuthoritySupport user2SubAuthoritySupport =
                            userAndSliceApiWrapperUser2.getSubAuthoritySupport(test.getLogger());

                    GeniUser user2 = secondUser;

                    if (!userAndSliceApiWrapperUser2.hasUserCredentials())
                        userAndSliceApiWrapperUser2.getUserCredentials(test.getLogger(), user2.getUserUrn());

                    UserAndSliceApiWrapper.SliceInfo sliceInfo;
                    if (user2SubAuthoritySupport.equals(UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY) || preferUsingProject) {
                        String subAuthorityName;
                        if (project == null) {
                            List<String> subAuthorityNames;
                            try {
                                subAuthorityNames =
                                        userAndSliceApiWrapperUser2.getSubAuthorityNames(test.getLogger(), user2.getUserUrn());
                            } catch (Throwable t) {
                                LOG.warn("Exception while getting sub authority names. will try to ignore.", t);
                                test.note("Exception while getting sub authority names. will try to ignore. Exception: " + t.getMessage());
                                subAuthorityNames = new ArrayList<String>();
                            }
                            if (subAuthorityNames.isEmpty()) {
                                if (user2SubAuthoritySupport.equals(UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY))
                                    throw new RuntimeException("There are no sub authorities for user " + user2.getUserUrn());
                                else
                                    subAuthorityName = project;
                            } else {
                                subAuthorityName = subAuthorityNames.get(0);
                            }
                        } else
                            subAuthorityName = project;
                        sliceInfo =
                                userAndSliceApiWrapperUser2.createSlice(test.getLogger(), res.name, expirationDate, subAuthorityName);
                    } else {
                        String subAuthorityName = project; //may be null
                        sliceInfo =
                                userAndSliceApiWrapperUser2.createSlice(test.getLogger(), res.name, expirationDate, subAuthorityName);
                    }

                    res.credentials = sliceInfo.getCredentials();
                    res.urn = sliceInfo.getUrn();
                    res.urnString = res.urn.toString();
                }
                success = true;
            } catch (Exception e) {
                //ignore
                LOG.warn("Exception creating slice, will ignore for now", e);
                success = false;
            }

            if (!success) {
                //maybe name is taken

                tries++;
                test.assertTrue(tries <= MAX_CREATE_SLICE_TRIES, "tried " + tries + " times. Maximum tries is " + MAX_CREATE_SLICE_TRIES + ". -> Giving up!");
                if (tries > MAX_CREATE_SLICE_TRIES) {
                    res.credentials = null;
                    break;
                }

                //wait before trying again
                try {
                    Thread.sleep(1200 /*ms*/);
                } catch (InterruptedException ignored) {
                }
            }
        }
        test.assertNotNull(res.credentials);
        test.assertNotEmpty(res.credentials);

        //old version without UserAndSliceApiWrapper
//        ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> registerRes = null;
//
//        int tries = 0;
//        boolean success = false;
//        while (!success) {
//            res.name = CommonAMTest.createSliceName(test, userAuthority, prefix);
//            res.urn = new GeniUrn(userAuthority.getNameForUrn(), "slice", res.name);
//            res.urnString = res.urn.toString();
//
//            if (callAsUser1)
//                registerRes = sa.register(getSAConnection(), getUserCredential(), res.urn);
//            else
//                registerRes = sa.register(getSAUser2Connection(), getUser2Credential(), res.urn);
//            TestSliceAuthority.testSACorrectnessXmlRpcResult(registerRes.getRawResult());
//            success = registerRes.getGeniResponseCode().isSuccess();
//
//            if (!success) {
//                //maybe name is taken
//                test.note("Could not create slice \"" + res.urnString + "\": " + registerRes.getGeniResponseCode() + " output: " + registerRes.getOutput());
//
//                tries++;
//                test.assertTrue(tries <= MAX_CREATE_SLICE_TRIES, "tried "+tries+" times. Maximum tries is "+MAX_CREATE_SLICE_TRIES+". -> Giving up!");
//                if (tries > MAX_CREATE_SLICE_TRIES)
//                    break;
//
//                //wait before trying again
//                try { Thread.sleep(1200 /*ms*/); } catch (InterruptedException e) { }
//            }
//        }
//
//        test.assertNotNull(registerRes);
//        test.assertTrue(registerRes.getGeniResponseCode().isSuccess(),"Register call was not successful in creating \""+res.urn.getValue()+"\": "+registerRes.getGeniResponseCode()+" output="+registerRes.getOutput());
//        res.credential = registerRes.getValue();
//        test.assertNotNull(res.credential);
//        test.assertNotNull(res.credential.getCredentialXml());
//        test.assertTrue(res.credential.getCredentialXml().length() > 10, "credential too short "+ res.credential.getCredentialXml());

        test.assertNotNull(res.credentials);
        test.assertNotEmpty(res.credentials);

        allSliceInfo.add(res);

        return res;
    }

    /**
     * @return the SliceInfo, or null if the slice does not exist.
     */
    public SliceInfo reuseSlice(String sliceName) throws JFedException {
        boolean callAsUser1 = testCredentialType.equals(TestCredentialType.REGULAR);
        assert callAsUser1 != testCredentialType.equals(TestCredentialType.SPEAKSFOR) || testCredentialType.equals(TestCredentialType.DELEGATION);

        SliceInfo res = new SliceInfo();
        int tries = 0;
        res.name = sliceName;
        res.urn = null; //could be sub authority slice, so we don't know urn!
        res.urnString = null;

        try {
            if (callAsUser1) {
                List<GeniUrn> sliceUrns =
                        userAndSliceApiWrapperUser1.getSlicesForUser(test.getLogger(), test.getUser().getUserUrn());

                //could be sub authority slice
                for (GeniUrn sliceUrn : sliceUrns) {
                    if (sliceUrn.getEncodedResourceName().equals(sliceName)) {
                        res.urn = sliceUrn;
                        break;
                    }
                }

                //no slice by that name
                if (res.urn == null)
                    return null;

                res.credentials = userAndSliceApiWrapperUser1.getSliceCredentials(test.getLogger(), res.urn);
            } else {
                List<GeniUrn> sliceUrns =
                        userAndSliceApiWrapperUser2.getSlicesForUser(test.getLogger(), secondUser.getUserUrn());

                //could be sub authority slice
                for (GeniUrn sliceUrn : sliceUrns) {
                    if (sliceUrn.getEncodedResourceName().equals(sliceName)) {
                        res.urn = sliceUrn;
                        break;
                    }
                }

                //no slice by that name
                if (res.urn == null)
                    return null;

                res.credentials = userAndSliceApiWrapperUser2.getSliceCredentials(test.getLogger(), res.urn);
            }
        } catch (Exception e) {
            //ignore
            LOG.info("Exception tring to reuse slice, will ignore", e);
            return null;
        }

        test.assertNotNull(res.urn);
        res.urnString = res.urn.toString();

        test.assertNotNull(res.credentials);
        test.assertNotEmpty(res.credentials);

        allSliceInfo.add(res);

        return res;
    }

    public SliceInfo renewSlice(SliceInfo sliceInfo, Date newExpirationtime) throws JFedException {
        //this just changes the actual sliceInfo in place

        boolean callAsUser1 = testCredentialType.equals(TestCredentialType.REGULAR);
        assert callAsUser1 != testCredentialType.equals(TestCredentialType.SPEAKSFOR) || testCredentialType.equals(TestCredentialType.DELEGATION);

        if (callAsUser1) {
            sliceInfo.credentials =
                    userAndSliceApiWrapperUser1.renewSlice(test.getLogger(), sliceInfo.credentials, newExpirationtime);
        } else {
            sliceInfo.credentials =
                    userAndSliceApiWrapperUser2.renewSlice(test.getLogger(), sliceInfo.credentials, newExpirationtime);
        }

        test.assertNotNull(sliceInfo.credentials);
        test.assertNotEmpty(sliceInfo.credentials);

        return sliceInfo;
    }

    public static boolean sameNodesInRequestAndManifest(ApiTest test, String requestRspec, String manifestRspec) {
        List<BasicStringRspec.BasicNodeInfo> basicNodeInfoRequest =
                new BasicStringRspec(requestRspec).getBasicNodeInfo();
        List<BasicStringRspec.BasicNodeInfo> basicNodeInfoManifest =
                new BasicStringRspec(manifestRspec).getBasicNodeInfo();
        if (basicNodeInfoManifest.size() < basicNodeInfoRequest.size())
            return false;
        if (basicNodeInfoManifest.size() > basicNodeInfoRequest.size())
            test.warn("There are more nodes in manifest(" + basicNodeInfoManifest.size() + ") " +
                    "than in request(" + basicNodeInfoRequest.size() + ")!");
        boolean res = true;
        String clientIds = "";
        for (BasicStringRspec.BasicNodeInfo requestNode : basicNodeInfoRequest) {
            String clientId = requestNode.getClientId();
            clientIds += " \"" + clientId + "\"";
            if (clientId != null) {
                boolean found = false;
                for (BasicStringRspec.BasicNodeInfo manifestNode : basicNodeInfoManifest)
                    if (manifestNode.getClientId() != null && manifestNode.getClientId().equals(clientId)) {
                        found = true;
                        break;
                    }
                if (!found) {
                    test.errorNonFatal("Request RSpec had node with client_id \"" + clientId + "\" but the received manifest does not have this node.");
                    res = false;
                }
            }
        }
        if (res)
            test.note("Successfully found the following client_id's in both request and manifest:" + clientIds);
        return res;
    }

    private static enum TestCredentialType {REGULAR, SPEAKSFOR, DELEGATION}

    public static class SliceInfo {
        String name;
        GeniUrn urn;
        List<AnyCredential> credentials;
        SfaCredential delegatedCredential;
        String urnString;

        public SfaCredential getDelegatedCredential(GeniUser receiverUser, GeniUser granterUser) {
            if (delegatedCredential != null) return delegatedCredential;
            if (credentials == null || credentials.isEmpty()) return null;

            SfaCredential sfaCredential = null;
            for (AnyCredential credential : credentials)
                if (credential instanceof SfaCredential)
                    sfaCredential = (SfaCredential) credential;
            if (sfaCredential == null)
                return null;

            try {
                delegatedCredential = sfaCredential.delegate(
                        receiverUser.getUserUrnString()/*newOwnerUrn*/,
                        receiverUser.getClientCertificateChain().get(0)/*newOwnerCert*/,
                        granterUser.getPrivateKey()/*originalOwnerPrivateKey*/,
                        new Date(System.currentTimeMillis() + 24 * 60 * 60 * 1000)/*expireDate*/,
                        "*"/*delegatedRights*/,
                        false/*canDelegate*/);
            } catch (CredentialException e) {
                throw new RuntimeException(e.getMessage(), e);
            }
            return delegatedCredential;
        }
    }
}
