package be.iminds.ilabt.jfed.preferences;

import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.util.KeyUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Singleton;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 16:26
 */
@Singleton
public class JFedPreferencesApplier {
    private static final Logger LOG = LoggerFactory.getLogger(JFedPreferencesApplier.class);

    private final JFedGuiPreferences jFedPreferences;
    private final EasyModel easyModel;
    private final GeniUserProvider geniUserProvider;

    @Inject
    public JFedPreferencesApplier(JFedGuiPreferences jFedPreferences, EasyModel easyModel, GeniUserProvider geniUserProvider) {
        this.jFedPreferences = jFedPreferences;
        this.easyModel = easyModel;
        this.geniUserProvider = geniUserProvider;
    }

    public void applyPreferences() {
        registerPrivateKeys();
    }

    private void registerPrivateKeys() {
        //set user key to key in login cert (always do this: used for jFed internal SSH connections, see https://ibcn-jira.intec.ugent.be/browse/FEDIBBTDEV-406),
        // also add key from configuration if present.
        SshKeyInfo userSshKeyInfo = new UserSshKeyInfo(geniUserProvider.getLoggedInGeniUser());
        SshKeyInfo overrideSshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();
        if (overrideSshKeyInfo != null) {
            String overridePubKeyStr =
                    KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(overrideSshKeyInfo.getPublicKey());
            LOG.info("Will add public key from configuration to created slivers: " + overridePubKeyStr);
            easyModel.addUserKey(overridePubKeyStr);
        }

        if (userSshKeyInfo.getPublicKey() != null) {
            String userPubKeyStr = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(userSshKeyInfo.getPublicKey());
            LOG.info("Will add user specified public key to created slivers: " + userPubKeyStr);
            easyModel.addUserKey(userPubKeyStr);
        } else
            LOG.warn("No public key in userSshKeyInfo (class=" + userSshKeyInfo.getClass().getName() + ")");


    }

}
