package be.iminds.ilabt.jfed.experimenter_gui;

import be.iminds.ilabt.jfed.experimenter_gui.call_gui.TasksWindow;
import be.iminds.ilabt.jfed.experimenter_gui.connectivity_tester.ConnectivityTesterGUI;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ExperimentEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ExperimentEditorTab;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ribbon_tabs.EditorRibbonTab;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewControllerTab;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceRibbonTabs;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ribbon_tabs.SliceRibbonTab;
import be.iminds.ilabt.jfed.experimenter_gui.tabs.RibbonEnabled;
import be.iminds.ilabt.jfed.experimenter_gui.tabs.StatusEnabled;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.OpenRequestRspecTask;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.RecoverSliceTask;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ExperimentEditorListProvider;
import be.iminds.ilabt.jfed.experimenter_gui.ui.StatusUpdateListener;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TasksProgressIndicator;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonBar;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonTab;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.JavaFXTaskThread;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import javafx.application.Platform;
import javafx.collections.ListChangeListener;
import javafx.collections.MapChangeListener;
import javafx.collections.ObservableMap;
import javafx.concurrent.Task;
import javafx.fxml.FXML;
import javafx.scene.control.Label;
import javafx.scene.control.Tab;
import javafx.scene.control.TabPane;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.Glyph;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 10/18/13
 * Time: 4:21 PM
 */
@Singleton
public class ExperimenterGUI implements StatusUpdateListener, TaskService, ExperimentEditorListProvider, ExperimentsTabPaneProvider {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimenterGUI.class);

    private final ExecutorService executorService = Executors.newCachedThreadPool();

    private final Provider<ConnectivityTesterGUI> connectivityTesterGUIProvider;
    private final ObservableMap<GeniUrn, ExperimentViewController> sliceControllers;
    private final JavaFXTaskThread javaFXTaskThread;
    private final HighLevelController hlc;
    private final GeniUserProvider geniUserProvider;
    private final EasyModel easyModel;
    private final TasksFactory tasksFactory;
    private final TasksWindow tasksWindow;
    private final GeneralRibbonTab generalRibbonTab;
    private final JFedGuiPreferences jFedPreferences;
    private final SliceRibbonTabs sliceRibbonTabs;
    private final StartupParameters startupParameters;

    @Inject
    public ExperimenterGUI(Provider<ConnectivityTesterGUI> connectivityTesterGUIProvider, EasyModel easyModel,
                           GeneralRibbonTab generalRibbonTab, HighLevelController hlc, JavaFXTaskThread javaFXTaskThread,
                           TasksFactory tasksFactory, ObservableMap<GeniUrn, ExperimentViewController> sliceControllers,
                           GeniUserProvider geniUserProvider, TasksWindow tasksWindow, JFedGuiPreferences jFedPreferences,
                           SliceRibbonTabs sliceRibbonTabs, StartupParameters startupParameters) {
        this.connectivityTesterGUIProvider = connectivityTesterGUIProvider;
        this.easyModel = easyModel;
        this.generalRibbonTab = generalRibbonTab;
        this.hlc = hlc;
        this.javaFXTaskThread = javaFXTaskThread;
        this.tasksFactory = tasksFactory;
        this.sliceControllers = sliceControllers;
        this.geniUserProvider = geniUserProvider;
        this.tasksWindow = tasksWindow;
        this.jFedPreferences = jFedPreferences;
        this.sliceRibbonTabs = sliceRibbonTabs;
        this.startupParameters = startupParameters;
    }

    @FXML
    private BorderPane root;
    @FXML
    private RibbonBar ribbonBar;
    @FXML
    private VBox noTabsOpenBox;
    @FXML
    private TabPane tabPane;
    @FXML
    private Label leftStatus;
    @FXML
    private Label rightStatus;
    @FXML
    private Label connectivityLabel;
    @FXML
    private Glyph connectivityLabelIcon;
    @FXML
    private Label pageantLabel;
    @FXML
    private TaskStatusIndicator callsStatusIndicator;

    @FXML
    private TasksProgressIndicator tasksProgressIndicator;

    @FXML
    private void initialize() {
        //add and initialize the general ribbon tab
        ribbonBar.getTabs().add(generalRibbonTab);
        generalRibbonTab.registerExperimentsTabPane(tabPane);

        sliceControllers.addListener((MapChangeListener<GeniUrn, ExperimentViewController>) change -> {
            if (change.wasRemoved()) {
                ExperimentViewController removedSliceController = change.getValueRemoved();
                ExperimentViewControllerTab experimentViewControllerTab = null;
                for (Tab tab : getTabPane().getTabs()) {
                    if (tab instanceof ExperimentViewControllerTab &&
                            ((ExperimentViewControllerTab) tab).getExperimentViewController() == removedSliceController) {
                        experimentViewControllerTab = (ExperimentViewControllerTab) tab;
                        break;
                    }
                }
                if (experimentViewControllerTab != null) {
                    tabPane.getTabs().remove(experimentViewControllerTab);
                }
            }
            if (change.wasAdded()) {
                ExperimentViewControllerTab experimentViewControllerTab =
                        new ExperimentViewControllerTab(change.getValueAdded(), sliceControllers, sliceRibbonTabs);
                getTabPane().getTabs().add(experimentViewControllerTab);
                getTabPane().getSelectionModel().select(experimentViewControllerTab);
            }
        });


        // hide/show "no tabs" notification when a tab is opened/all tabas are closed
        tabPane.getTabs().addListener((ListChangeListener<Tab>) change -> noTabsOpenBox.setVisible(tabPane.getTabs().size() == 0));


        tabPane.getSelectionModel().selectedItemProperty().addListener(
                (observableValue, oldValue, newValue) -> {
                    //bind to status-property of tab if it is available

                    if (oldValue != null && oldValue instanceof StatusEnabled)
                        leftStatus.textProperty().unbind();

                    if (newValue != null && newValue instanceof StatusEnabled)
                        leftStatus.textProperty().bind(((StatusEnabled) newValue).statusProperty());
                    else {
                        leftStatus.setText("");
                    }

                    //add and/or remove ribbontabs if necessary
                    if (oldValue != null && oldValue instanceof RibbonEnabled) {
                        ribbonBar.getTabs().removeAll(((RibbonEnabled) oldValue).getRibbonTabs());
                    }

                    if (newValue != null && newValue instanceof RibbonEnabled) {

                        //register ExperimentEditor in EditorRibbonTab's
                        if (newValue instanceof ExperimentEditorTab) {
                            for (RibbonTab rt : ((ExperimentEditorTab) newValue).getRibbonTabs()) {
                                if (rt instanceof EditorRibbonTab) {
                                    ((EditorRibbonTab) rt).setActiveExperimentEditor(
                                            ((ExperimentEditorTab) newValue).getExperimentEditor());
                                }
                            }
                        }

                        //register SliceController in SliceRibbonTab's
                        if (newValue instanceof ExperimentViewControllerTab) {
                            for (RibbonTab rt : ((ExperimentViewControllerTab) newValue).getRibbonTabs()) {
                                if (rt instanceof SliceRibbonTab) {
                                    ((SliceRibbonTab) rt).setActiveExperimentViewController(
                                            ((ExperimentViewControllerTab) newValue).getExperimentViewController());
                                }
                            }
                        }

                        ribbonBar.getSelectionModel().selectFirst();
                        ribbonBar.getTabs().addAll(((RibbonEnabled) newValue).getRibbonTabs());

                        Tab tabToActivate = ((RibbonEnabled) newValue).getInitialTabToActivate();
                        ribbonBar.getSelectionModel().select(tabToActivate);


                    } else {
                        ribbonBar.getSelectionModel().select(0); //select general tab
                    }

                    fixRibbonBar();
                }
        );

        //initialize progress indicator
        javaFXTaskThread.busyProperty().addListener((observableValue, oldValue, newValue) -> callsStatusIndicator.setStatus(
                newValue ? TaskStatusIndicator.Status.BUSY : TaskStatusIndicator.Status.SUCCESS));
        callsStatusIndicator.setOnMouseClicked(mouseEvent -> onOpenCallOverviewAction());

        //request only slices but don't resolve all of them yet
        final HighLevelController.ListSlicesTask listSlicesTask = hlc.getSlicesForUser();

        listSlicesTask.addCallback((task, taskExecution, state) -> {
            if (taskExecution.getState() == TaskExecution.TaskState.FAILED) {
                Platform.runLater(() -> {
                    JFDialogs.create()
                            .masthead("Information call failed")
                            .message("jFed could not successfully execute a basic information call to your slice authority. " +
                                    "Other calls will fail most likely too.\n\n" +
                                    "This probably indicates an authentication problem. Has your certificate been revoked? Certain actions " +
                                    "(like changing your password at your authority) will invalidate all issued certificates.\n\n" +
                                    "Review the call log for more information.")
                            .showError();
                });
            }
        });

        hlc.submitTask(listSlicesTask);


        //If emulab-user: contact emulab-registery
        if (geniUserProvider.getLoggedInGeniUser().getUserAuthority().getUrnString().equals("urn:publicid:IDN+emulab.net+authority+cm")) {
            //HACK to treat emulab different
            be.iminds.ilabt.jfed.highlevel.controller.Task t =
                    hlc.emulabRegistry(geniUserProvider.getLoggedInGeniUser().getUserAuthority());
            hlc.submitTask(t);
        }

        updateConnectivityStatusLabel();

        //start with  the recovery of a slice if a slice_urn was provided as a startup parameter
        if (startupParameters.contains(StartupParameters.PARAM_SLICE_URN)) {
            String sliceUrnParam = startupParameters.getValue(StartupParameters.PARAM_SLICE_URN);
            sliceUrnParam = sliceUrnParam.trim();
            if (!sliceUrnParam.isEmpty()) {
                fetchAndRecoverSlice(sliceUrnParam);
            }

        }
        //open an rspec if a rspec-parameter was provided as a startup parameter
        if (startupParameters.contains(StartupParameters.PARAM_RSPEC)) {
            String rspec = startupParameters.getValue(StartupParameters.PARAM_RSPEC);
            rspec = rspec.trim();
            if (!rspec.isEmpty()) {
                try {
                    LOG.debug("Opening rspec '{}', which was provided as startup parameter", rspec);
                    OpenRequestRspecTask openRequestRspecTask =
                            tasksFactory.createOpenRequestRspecTask(null, rspec);
                    submitTask(openRequestRspecTask);
                } catch (IOException ex) {
                    LOG.warn("Could not open rspec '{}' provided in startup parameters", rspec, ex);
                }
            }
        }
    }

    private void fetchAndRecoverSlice(String sliceUrnParam) {
        final GeniUrn sliceUrn = GeniUrn.parse(sliceUrnParam);
        if (sliceUrn == null) {
            LOG.error("Could not parse '{}' into a valid slice", sliceUrnParam);

            JFDialogs.create()
                    .owner(ExperimenterGUI.this.getRoot().getScene().getWindow())
                    .message(String.format("jFed was unable to recover slice '%s' because the URN cannot be parsed properly. Please verify your input.", sliceUrnParam))
                    .masthead("Slice recovery failed").showError();

        } else if (!"slice".equals(sliceUrn.getEncodedResourceType())) {

            JFDialogs.create()
                    .owner(ExperimenterGUI.this.getRoot().getScene().getWindow())
                    .message(String.format("jFed was unable to recover slice '%s' because the URN has an incorrect resource type: only the resource type 'slice' is supported! Please verify your input.", sliceUrnParam))
                    .masthead("Slice recovery failed").showError();

            LOG.error("The given URN '{}' does not have the resource type 'slice'!", sliceUrn.toString());
        }


        Task<Slice> fetchAndRecoverSliceTask = new Task<Slice>() {
            @Override
            protected Slice call() throws Exception {
                updateProgress(-1, 0);
                updateMessage("Fetching available slices");

                final HighLevelController.ListSlicesTask refreshSlicesTask = hlc.getSlicesForUser();

                TaskExecution taskExecution = hlc.submitTask(refreshSlicesTask);

                while (!taskExecution.isCompleted()) {
                    Thread.sleep(100);
                }

                return easyModel.getSlice(sliceUrn);
            }

            @Override
            protected void succeeded() {
                Slice slice = getValue();

                if (slice == null) {
                    LOG.warn("Could not retrieve '{}' as one of the slices of the current user!", sliceUrn);

                    JFDialogs.create()
                            .owner(ExperimenterGUI.this.getRoot().getScene().getWindow())
                            .message(String.format("jFed was unable to recover slice '%s'. Please check if the slice exists, hasn't expired, and if you have access to this slice.", sliceUrn))
                            .masthead("Slice recovery failed").showError();

                    return;
                }

                RecoverSliceTask recoverSliceTask = tasksFactory.createRecoverSliceTask(slice);

                submitTask(recoverSliceTask);
            }
        };

        submitTask(fetchAndRecoverSliceTask);
    }


    public void setSecondaryStatus(String status) {
        rightStatus.setText(status);
    }

    @Override
    public TabPane getTabPane() {
        return tabPane;
    }


    @FXML
    private void onOpenCallOverviewAction() {
        tasksWindow.showTasks();
    }

    @FXML
    public void onOpenConnectivityTestAction() {
        connectivityTesterGUIProvider.get().showDialog();
    }


    public void updateConnectivityStatusLabel() {
        if (connectivityLabel == null) //the GUI hasn't been injected yet by JavaFX. We can safely ignore this update request
            return;

        assert connectivityLabelIcon != null;
        assert pageantLabel != null;

        String jFedProxy = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED);
        String sshProxy = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH);
        if (jFedProxy == null)
            jFedProxy = ProxyPreferencesManager.NEVER_USE_PROXY;
        if (sshProxy == null)
            sshProxy = ProxyPreferencesManager.NEVER_USE_PROXY;


        String statusText;
        Color color;

        if (jFedProxy.equals(ProxyPreferencesManager.NEVER_USE_PROXY) &&
                sshProxy.equals(ProxyPreferencesManager.NEVER_USE_PROXY)) {
            statusText = "Direct connection";
            color = Color.LAWNGREEN;
        } else if (jFedProxy.equals(ProxyPreferencesManager.ALWAYS_USE_PROXY) &&
                sshProxy.equals(ProxyPreferencesManager.ALWAYS_USE_PROXY)) {
            statusText = "Proxy fully enabled";
            color = Color.AQUA;
        } else if (jFedProxy.equals(ProxyPreferencesManager.ALWAYS_USE_PROXY) &&
                sshProxy.equals(ProxyPreferencesManager.NEVER_USE_PROXY)) {
            statusText = "Proxy enabled for jFed only";
            color = Color.DARKORANGE;
        } else {
            statusText = "Proxy enabled for SSH-sessions only";
            color = Color.YELLOW;
        }

        connectivityLabel.setText(statusText);
        connectivityLabelIcon.setColor(color);

        pageantLabel.setVisible(jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT, false)
                || jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SSHAGENT_USE, false));

    }


    @Override
    public void submitTask(Task<?> task) {
        tasksProgressIndicator.followTaskProgress(task);
        executorService.submit(task);
    }


    /**
     * Works around bug in Java8u20. See   FEDIBBTDEV-1046
     */
    private void fixRibbonBar() {
        List<Tab> tabs = new ArrayList<>(ribbonBar.getTabs());
        Tab activeTab = ribbonBar.getSelectionModel().getSelectedItem();
        ribbonBar.getTabs().clear();

        RibbonBar newRibbonBar = new RibbonBar();
        newRibbonBar.getTabs().addAll(tabs);
        newRibbonBar.getSelectionModel().select(activeTab);

        root.setTop(newRibbonBar);
        this.ribbonBar = newRibbonBar;


    }

    public BorderPane getRoot() {
        return root;
    }

    @Override
    public void setStatus(String status) {
        setSecondaryStatus(status);
    }

    @Override
    public List<ExperimentEditor> getExperimentEditors() {
        return tabPane.getTabs().stream()
                .filter(tab -> tab instanceof ExperimentEditorTab)
                .map(tab -> ((ExperimentEditorTab) tab).getExperimentEditor())
                .collect(Collectors.toList());
    }
}
