package be.iminds.ilabt.jfed.experimenter_gui;

import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.NodeDescription;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.RspecNodeDescription;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.ImageMap;
import be.iminds.ilabt.jfed.jaxb.testbed_nodes_map.BoundsType;
import be.iminds.ilabt.jfed.jaxb.testbed_nodes_map.TestbedNodeType;
import be.iminds.ilabt.jfed.jaxb.testbed_nodes_map.TestbedNodesMap;
import be.iminds.ilabt.jfed.jaxb.testbed_nodes_maps.TestbedNodesMaps;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import javafx.scene.paint.Color;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.HierarchicalConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

/**
 * User: twalcari
 * Date: 1/9/14
 * Time: 5:17 PM
 */
@Singleton
public class JFedConfiguration {
    private static final Logger LOG = LoggerFactory.getLogger(JFedConfiguration.class);

    private static final String DEFAULT_MAPS_LOCATION = "https://flsmonitor.fed4fire.eu/maps/testbed_nodes_maps.xml";
    private final static String[] CONFIGURATION_LOCS = {"/configuration1.xml" , "/configuration2.xml"};
    private final static String CONFIGURATION_EMULAB_LOC = "/configuration_emulab.xml";
    private final static String CONFIGURATION_GENI_LOC = "/configuration_geni.xml";


    private final static String ENABLED_TESTBEDS = "enabled-testbeds.enabled-testbed";
    private static final String ENABLED_TESTBED_URN = "urn";
    private static final String ENABLED_COMPONENT_ID_URN = "enabled-componentIds.urn";
    private final static String NODE_DESCRIPTIONS = "node-descriptions.node-description";
    private final static String TYPE_DEFINITION = "type-definition";

    private static final String TESTBED_DESCRIPTIONS = "testbed-descriptions.testbed";
    private static final String TESTBED_SLIVER_TYPES = "sliver-types.sliver-type";
    private static final String TESTBED_REQUIRE_FIXED_NODE_ASSIGNMENT = "require-fixed-node-assignment";
    private static final String TESTBED_HAS_LINK_SUPPORT = "allow-links";
    private static final String TESTBED_HAS_CHANNEL_SUPPORT = "has-channels";
    private static final String TESTBED_COLOR = "[@color]";
    private static final String TESTBED_LOGO = "[@logo]";
    /**
     * Should this testbed appear in the list of potential testbeds when an unbound node is found in an RSpec?
     */
    private static final String TESTBED_SHOW_AS_BINDABLE = "show-as-bindable";
    private static final String TESTBED_NODES_VISUALIZATION = "nodes-visualization";
    private static final String TESTBED_SUPPORTS_LIST_DISKIMAGES = "supports-list-diskimages";
    //does the testbed support links between nodes?
    private static final String INTERFACE_DEFINITION = "interfaces.interface";
    private static final String KEY_INTERFACES = "interfaces";
    private static final String KEY_OS_IMAGE = "os_image";
    private static final String KEY_SLIVER_TYPE = "sliver_type";
    private static final String KEY_COMPONENT_ID = "component_id";
    private static final String KEY_EXCLUSIVE = "exclusive";
    private static final String KEY_CONFIGURE_INTERFACES = "configure-interfaces";
    private static final String KEY_COMPONENT_MANAGER_ID = "component_manager_id";
    private static final String KEY_IP_ADDRESS = "ip[@address]";
    private static final String KEY_IP_NETMASK = "ip[@netmask]";
    private static final String KEY_IP_TYPE = "ip[@type]";
    private static final String COUNTER_INTERFACE_DEFINITION = "counter-interfaces.interface";
    private static final String SLIVERTYPE_NAME = "sliver";
    private static final String SLIVERTYPE_EXCLUSIVE = "[@exclusive]";
    private static final String SLIVERTYPE_DISKIMAGENAME = "[@disk-image-name]";
    private final List<NodeDescription> nodeDescriptions;
    private final Map<String, NodeDescription> nodeDescriptionsMap;
    private final Map<GeniUrn, TestbedDescription> testbedDescriptions;

    private final GeniUserProvider geniUserProvider;

    @Inject
    JFedConfiguration(GeniUserProvider geniUserProvider) {
        this.geniUserProvider = geniUserProvider;

        Map<GeniUrn, ImageMap<GeniUrn>> imageMaps = new HashMap<>();

        //now try to load imagemaps
        try {
            JAXBContext contextMaps = null;
            JAXBContext contextMap =  null;
            try {

                contextMaps = JAXBContext.newInstance(TestbedNodesMaps.class);
                contextMap = JAXBContext.newInstance("be.iminds.ilabt.jfed.jaxb.testbed_nodes_map");
            } catch (JAXBException e) {
                e.printStackTrace();
            }
            Unmarshaller unmarshaller = contextMaps.createUnmarshaller();

            URL mapsUrl = new URL(DEFAULT_MAPS_LOCATION);

            TestbedNodesMaps testbedNodesMaps =
                    (TestbedNodesMaps) unmarshaller.unmarshal(mapsUrl);

            unmarshaller = contextMap.createUnmarshaller();
            for (TestbedNodesMaps.TestbedNodeMap tnmRef : testbedNodesMaps.getTestbedNodeMap()) {

                final URL mapUrl = new URL(mapsUrl, tnmRef.getUrl());
                TestbedNodesMap testbedNodesMap =
                        (TestbedNodesMap) unmarshaller.unmarshal(mapUrl);

                final URL imageUrl = new URL(mapUrl, testbedNodesMap.getMapUrl());

                ImageMap<GeniUrn> testbedImageMap = new ImageMap<>(imageUrl);
                for (TestbedNodeType testbedNode : testbedNodesMap.getTestbedNode()) {
                    BoundsType b = testbedNode.getBounds();
                    testbedImageMap.addSelectableRegion(new ImageMap.Bounds(b.getMinX(), b.getMaxX(), b.getMinY(), b.getMaxY()),
                            GeniUrn.parse(testbedNode.getComponentId()));
                }

                imageMaps.put(GeniUrn.parse(tnmRef.getComponentManagerId()), testbedImageMap);
            }
        } catch (JAXBException | MalformedURLException e) {
            LOG.debug("Could not initialize testbed nodes maps", e);
        }


        try {
            HierarchicalConfiguration conf;

            if (this.geniUserProvider.getLoggedInGeniUser().getUserAuthority().getUrnString().equals("urn:publicid:IDN+emulab.net+authority+cm")) {
                //HACK to treat emulab different
                conf = new XMLConfiguration(JFedConfiguration.class.getResource(CONFIGURATION_EMULAB_LOC));
            } else if (geniUserProvider.getLoggedInGeniUser().getUserAuthority().getUrnString().equals("urn:publicid:IDN+ch.geni.net+authority+cm")) {
                //HACK to treat GENI different
                conf = new XMLConfiguration(JFedConfiguration.class.getResource(CONFIGURATION_GENI_LOC));
            } else {
                int randomConfigIndex = (int) (System.currentTimeMillis() % CONFIGURATION_LOCS.length);
                String randomConfig = CONFIGURATION_LOCS[randomConfigIndex];
                LOG.trace("Config #{} was chosen: {}", randomConfigIndex, randomConfig);

                conf = new XMLConfiguration(JFedConfiguration.class.getResource(
                        randomConfig));  //Fed4FIRE


            }
            //load the node descriptions
            nodeDescriptions = new ArrayList<>();
            nodeDescriptionsMap = new HashMap<>();
            for (HierarchicalConfiguration nodeConf : conf.configurationsAt(NODE_DESCRIPTIONS)) {
                String id = nodeConf.getString("[@id]");
                String name = nodeConf.getString("[@friendlyName]");
/*                NodeDescription.NodeType type =
                        NodeDescription.NodeType.valueOf(nodeConf.getString("[@type]").toUpperCase());         */


                List<RspecNodeDescription> rspecNodeDescriptions = new ArrayList<>();
                for (HierarchicalConfiguration typeConf : nodeConf.configurationsAt(TYPE_DEFINITION)) {
                    String cmUrn = typeConf.getString(KEY_COMPONENT_MANAGER_ID);
                    boolean exclusive = typeConf.getBoolean(KEY_EXCLUSIVE);
                    String cUrn = typeConf.getString(KEY_COMPONENT_ID);
                    String sliverType = typeConf.getString(KEY_SLIVER_TYPE);

                    String osImage = typeConf.getString(KEY_OS_IMAGE, null);
                    boolean configureInterfaces = typeConf.getBoolean(KEY_CONFIGURE_INTERFACES, true);

                    rspecNodeDescriptions.add(
                            new RspecNodeDescription(cmUrn, cUrn, exclusive, sliverType, osImage, configureInterfaces)
                    );
                }

                Map<GeniUrn, EnabledTestbed> enabledTestbeds = new HashMap<>();
                //load the enabled testbeds
                for (HierarchicalConfiguration tbConf : nodeConf.configurationsAt(ENABLED_TESTBEDS)) {
                    GeniUrn tbUrn = GeniUrn.parse(tbConf.getString(ENABLED_TESTBED_URN));

                    String[] componentIdsArray = tbConf.getStringArray(ENABLED_COMPONENT_ID_URN);

                    ImmutableSet.Builder<GeniUrn> componentIds = ImmutableSet.builder();
                    for (String componentId : componentIdsArray) {
                        componentIds.add(GeniUrn.parse(componentId));
                    }

                    enabledTestbeds.put(tbUrn, new EnabledTestbed(tbUrn, componentIds.build()));


                }
                NodeDescription newNodeDescription =
                        new NodeDescription(id, name,/* type,*/ rspecNodeDescriptions, enabledTestbeds);
                nodeDescriptionsMap.put(id, newNodeDescription);
                nodeDescriptions.add(newNodeDescription);

            }

            testbedDescriptions = new HashMap<>();
            for (HierarchicalConfiguration testbedDesc : conf.configurationsAt(TESTBED_DESCRIPTIONS)) {
                GeniUrn urn = GeniUrn.parse(testbedDesc.getString("[@urn]"));

                boolean requireFixedNodeAssignment =
                        testbedDesc.getBoolean(TESTBED_REQUIRE_FIXED_NODE_ASSIGNMENT, false);

                boolean hasLinkSupport = testbedDesc.getBoolean(TESTBED_HAS_LINK_SUPPORT, true);
                boolean channelSupport = testbedDesc.getBoolean(TESTBED_HAS_CHANNEL_SUPPORT, false);
                boolean showAsBindable = testbedDesc.getBoolean(TESTBED_SHOW_AS_BINDABLE, false);
                boolean supportsListDiskImages = testbedDesc.getBoolean(TESTBED_SUPPORTS_LIST_DISKIMAGES, false);

                ImmutableList.Builder<SliverType> sliverTypes = ImmutableList.builder();
                for (HierarchicalConfiguration sliverTypeConf : testbedDesc.configurationsAt(TESTBED_SLIVER_TYPES)) {
                    String diskImageName = sliverTypeConf.getString(SLIVERTYPE_DISKIMAGENAME, null);
                    sliverTypes.add(new SliverType(sliverTypeConf.getString(""), sliverTypeConf.getBoolean(SLIVERTYPE_EXCLUSIVE, null), diskImageName));
                }

                Color color = Color.TRANSPARENT;
                try {
                    color = Color.valueOf(testbedDesc.getString(TESTBED_COLOR, "transparent"));
                } catch (IllegalArgumentException ex) {
                    LOG.warn("Cannot convert color {}", testbedDesc.getString(TESTBED_COLOR, "transparent"), ex);
                }

                String logo = testbedDesc.getString(TESTBED_LOGO);

                ImageMap<GeniUrn> testbedImageMap = imageMaps.get(urn);

                NodesVisualization nodesVisualization = testbedDesc.containsKey(TESTBED_NODES_VISUALIZATION) ?
                        NodesVisualization.valueOf(testbedDesc.getString(TESTBED_NODES_VISUALIZATION).toUpperCase()) : null;

                if(nodesVisualization == null && testbedImageMap != null){
                    nodesVisualization = NodesVisualization.IMAGE;
                }
                if(nodesVisualization == NodesVisualization.IMAGE && testbedImageMap == null){
                    LOG.warn("NodesVisualization is set to IMAGE for {}, but there is no image map loaded!", urn);
                }


                testbedDescriptions.put(urn, new TestbedDescription(urn,
                        requireFixedNodeAssignment, hasLinkSupport, channelSupport, showAsBindable, supportsListDiskImages,
                        ImmutableList.copyOf(sliverTypes.build()), color, logo,
                        nodesVisualization, testbedImageMap));
            }


        } catch (ConfigurationException e) {
            throw new RuntimeException("Error while loading jFed Experimenter Toolkit configuration", e);
        }




    }

    public Collection<NodeDescription> getNodeDescriptions() {
        return nodeDescriptions;
    }

    public
    @Nullable
    TestbedDescription getTestbedDescription(@Nonnull GeniUrn urn) {
        return testbedDescriptions.get(urn);
    }

    public
    @Nonnull
    Collection<TestbedDescription> getTestbedDescriptions() {
        return testbedDescriptions.values();
    }

    public
    @Nullable
    NodeDescription getNodeDescription(@Nonnull String id) {
        return nodeDescriptionsMap.get(id);
    }


    public static class EnabledTestbed {
        private final GeniUrn urn;

        private final ImmutableSet<GeniUrn> enabledComponentIds;

        public EnabledTestbed(@Nonnull GeniUrn urn, @Nonnull ImmutableSet<GeniUrn> enabledComponentIds) {
            this.urn = urn;
            this.enabledComponentIds = enabledComponentIds;
        }

        public
        @Nonnull
        GeniUrn getUrn() {
            return urn;
        }

        public
        @Nonnull
        ImmutableSet<GeniUrn> getEnabledComponentIds() {
            return enabledComponentIds;
        }
    }

    public static class TestbedDescription {
        private final GeniUrn urn;

        private final boolean requireFixedNodeAssignment;
        private final boolean linkSupport;
        private final boolean channelSupport;
        private final boolean showAsBindable;
        private final boolean supportsListDiskImages;
        private final ImmutableList<SliverType> sliverTypes;
        private final Color color;
        private final String logo;
        private final NodesVisualization nodesVisualization;
        private final ImageMap<GeniUrn> imageMap;


        public TestbedDescription(@Nonnull GeniUrn urn, boolean requireFixedNodeAssignment, boolean linkSupport, boolean channelSupport, boolean showAsBindable,
                                  boolean supportsListDiskImages,
                                  @Nonnull ImmutableList<SliverType> sliverTypes, @Nullable Color color, @Nullable String logo,
                                  NodesVisualization nodesVisualization, ImageMap<GeniUrn> imageMap) {
            this.urn = urn;
            this.requireFixedNodeAssignment = requireFixedNodeAssignment;
            this.linkSupport = linkSupport;
            this.channelSupport = channelSupport;
            this.showAsBindable = showAsBindable;
            this.supportsListDiskImages = supportsListDiskImages;
            this.sliverTypes = sliverTypes;
            this.color = color;
            this.logo = logo;
            this.nodesVisualization = nodesVisualization;
            this.imageMap = imageMap;
        }

        public
        @Nonnull
        GeniUrn getUrn() {
            return urn;
        }

        public boolean requiresFixedNodeAssignment() {
            return requireFixedNodeAssignment;
        }

        public boolean hasLinkSupport() {
            return linkSupport;
        }

        /**
         * Should this testbed appear in the list of potential testbeds when an unbound node is found in an RSpec?
         *
         * @return
         */
        public boolean isShowAsBindable() {
            return showAsBindable;
        }

        /**
         * Does this testbed support fetching diskimages in the AdaptivePropertiesDialog by using ListResources?
         *
         * @return
         */
        public boolean isSupportsListDiskImages() {
            return supportsListDiskImages;
        }

        public
        @Nonnull
        ImmutableList<SliverType> getSliverTypes() {
            return sliverTypes;
        }

        @Nullable
        public Color getColor() {
            return color;
        }

        @Nullable
        public String getLogo() {
            return logo;
        }

        @Nullable
        public NodesVisualization getNodesVisualization() {
            return nodesVisualization;
        }

        public ImageMap<GeniUrn> getImageMap() {
            return imageMap;
        }

        public boolean hasChannelSupport() {
            return channelSupport;
        }
    }

    public static enum NodesVisualization {
        MAP,
        IMAGE
    }


    public static class SliverType {
        private final String name;
        private final Boolean exclusive;
        private final String diskImageName;

        public SliverType(String name, Boolean exclusive) {
            this.name = name;
            this.exclusive = exclusive;
            this.diskImageName = null;
        }

        public SliverType(String name, Boolean exclusive, String diskImageName) {
            this.name = name;
            this.exclusive = exclusive;
            this.diskImageName = diskImageName;
        }

        public String getName() {
            return name;
        }

        public Boolean getExclusive() {
            return exclusive;
        }

        public String getDiskImageName() {
            return diskImageName;
        }
    }

    public boolean getSupportsListDiskImages(GeniUrn componentManagerUrn) {
        TestbedDescription testbedDescription = testbedDescriptions.get(componentManagerUrn);
        if (testbedDescription != null)
            return testbedDescription.isSupportsListDiskImages();
        return false;
    }

    public Boolean getNodeExclusivity(GeniUrn componentManagerUrn, String sliverTypeName) {
        TestbedDescription testbedDescription = testbedDescriptions.get(componentManagerUrn);

        if (testbedDescription != null) {
            for (SliverType sliverType : testbedDescription.getSliverTypes()) {
                if (sliverType.getName().equalsIgnoreCase(sliverTypeName)) {
                    return sliverType.getExclusive();
                }
            }

        }

        return null;
    }


    public NodeDescription detectJFedNodeType(FXRspecNode rspecNode) {

        /* This list is filled when we find a nodeDescription that matches the given rspecNode,
           except for the default componentManager.

           However, the componentManager of the rspecNode must still appear in the EnabledTestbeds-list
         */
        List<NodeDescription> nonDefaultCMNodeDesciptions = new ArrayList<>();

        for (NodeDescription nodeDescription : nodeDescriptionsMap.values()) {
            for (RspecNodeDescription rspecNodeDescription : nodeDescription.getRspecNodeDescriptions()) {
                if (rspecNodeDescription.getSliverType() != null &&
                        !rspecNodeDescription.getSliverType().equalsIgnoreCase(rspecNode.getSliverTypeName()))
                    continue;

                if (rspecNode.getExclusive() != (Boolean) rspecNodeDescription.isExclusive())
                    continue;

                if (rspecNodeDescription.getComponentIdUrn() != null && rspecNode.getComponentId() != null &&
                        !rspecNodeDescription.getComponentIdUrn().equalsIgnoreCase(rspecNode.getComponentId().toString()))
                    continue;

                if (rspecNodeDescription.getOsImage() != null &&
                        !rspecNode.getSliverTypes().isEmpty() &&
                        !rspecNode.getSliverTypes().get(0).getDiskImages().isEmpty() &&
                        !rspecNodeDescription.getOsImage().equalsIgnoreCase(rspecNode.getSliverTypes().get(0).getDiskImages().get(0).getName()))
                    continue;

                //return immediately if default CM matches that of rspecNode, check with other enabled testbeds otherwise
                if (rspecNodeDescription.getComponentManagerUrn() != null && rspecNode.getComponentManagerId() != null &&
                        rspecNodeDescription.getComponentManagerUrn().equalsIgnoreCase(rspecNode.getComponentManagerId().toString())) {
                    return nodeDescription;
                } else if (nodeDescription.getEnabledTestbeds().containsKey(rspecNode.getComponentManagerId())) {
                    nonDefaultCMNodeDesciptions.add(nodeDescription);
                }
            }
        }

        NodeDescription oldDescription = nodeDescriptionsMap.get(rspecNode.getNodeDescriptionId());

        if (oldDescription != null && nonDefaultCMNodeDesciptions.contains(oldDescription))
            return oldDescription;

        return !nonDefaultCMNodeDesciptions.isEmpty() ? nonDefaultCMNodeDesciptions.get(0) : null;
    }
}
