package be.iminds.ilabt.jfed.experimenter_gui.controller.view.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.controller.controller.ControllerController;
import be.iminds.ilabt.jfed.experimenter_gui.controller.view.TimeTextField;
import be.iminds.ilabt.jfed.rspec.model.controller.Timestamp;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.TextField;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import org.controlsfx.control.ButtonBar;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.validation.ValidationMessage;
import org.controlsfx.validation.ValidationSupport;
import org.controlsfx.validation.Validator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.stream.Collectors;

public class AddCommandDialog extends GridPane {
    private static final Logger LOG = LoggerFactory.getLogger(AddCommandDialog.class);

    private static final String ADD_COMMAND_FXML = "AddCommandDialog.fxml";
    private final ControllerController controller;
    private final FXRspecNode node;
    private int barrierSegmentID;
    @FXML
    private TextField tagTextField;
    @FXML
    private TextField commandTextField;
    @FXML
    private TimeTextField startTextField;

    private final Dialog dialog;
    private final DialogAction addCommandAction;

    private final ValidationSupport validationSupport;


    private AddCommandDialog(ControllerController controller, FXRspecNode node, Timestamp startingTime) {
        this.controller = controller;
        this.node = node;

        //initialize the Scene
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(ADD_COMMAND_FXML));

        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }


        startTextField.setTime(startingTime.toString());
        tagTextField.setText("Command " + controller.getModel().nextCommandNumber());

        this.barrierSegmentID = 0;

        //initialize dialog
        dialog = new Dialog(null, String.format("Add a command to node %s", node.getClientId()));
        dialog.setMasthead(dialog.getTitle());
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLUS, Color.GREEN));
        dialog.setContent(this);

        addCommandAction = new DialogAction("Add command", ButtonBar.ButtonType.OK_DONE, false, false, true,
                this::onAddCommandAction);


        dialog.getActions().setAll(addCommandAction, Dialog.ACTION_CANCEL);

        //initialize validation support
        validationSupport = new ValidationSupport();
        validationSupport.registerValidator(tagTextField, Validator.createEmptyValidator("Tag is a required field!"));
        validationSupport.registerValidator(commandTextField, Validator.createEmptyValidator("Command is a required field!"));
    }

    private void onAddCommandAction(ActionEvent actionEvent) {
        if (validationSupport.isInvalid()) {
            JFDialogs.create().message(String.format("Please correct the following issue(s): %s",
                    validationSupport.getValidationResult().getMessages().stream()
                            .map(ValidationMessage::getText)
                            .collect(Collectors.joining(", ")))).showWarning();
            return;
        }

        controller.addCommand(tagTextField.getText(), node.getUniqueId(), barrierSegmentID
                , new Timestamp(startTextField.toMillis())
                , new Timestamp(0)
                , commandTextField.getText());

        dialog.setResult(addCommandAction);
        dialog.hide();
    }

    private AddCommandDialog(ControllerController controller, FXRspecNode node, int barrierSegmentID, Timestamp startingTime) {
        this(controller, node, startingTime);
        this.barrierSegmentID = barrierSegmentID;
    }

    public void showDialog() {
        dialog.show();
    }

    public static void showSimplePropertiesDialog(ControllerController controller, FXRspecNode node, Timestamp startingTime) {
        AddCommandDialog npd = new AddCommandDialog(controller, node, startingTime);
        npd.showDialog();
    }

    public static void showSimplePropertiesDialog(ControllerController controller, FXRspecNode node, int barrierSegmentID, Timestamp startingTime) {
        AddCommandDialog npd = new AddCommandDialog(controller, node, barrierSegmentID, startingTime);

        npd.showDialog();
    }
}
