package be.iminds.ilabt.jfed.experimenter_gui.controller.view.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.controller.controller.ControllerController;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.CheckBoxTreeItem;
import javafx.scene.control.TextField;
import javafx.scene.control.TreeView;
import javafx.scene.control.cell.CheckBoxTreeCell;
import javafx.scene.input.KeyCode;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import org.controlsfx.control.ButtonBar;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class InstantCommandDialog extends GridPane {
    private static final Logger LOG = LoggerFactory.getLogger(AddCommandDialog.class);

    private static final String FXML = "InstantCommandDialog.fxml";
    private final ControllerController controller;
    @FXML
    private TextField commandTextField;
    @FXML
    private TextField tagTextField;
    @FXML
    private TreeView<String> treeView;

    private Map<CheckBoxTreeItem<String>, FXRspecNode> nodesMap = new HashMap<>();

    private final Dialog dialog;
    private final DialogAction okAction;

    private InstantCommandDialog(final ControllerController controller, List<FXRspecNode> nodes) {
        this.controller = controller;

        //initialize the Scene
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(FXML));

        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }


        CheckBoxTreeItem<String> rootItem = new CheckBoxTreeItem<>("Your experiment");
        rootItem.setExpanded(true);
        treeView.setRoot(rootItem);
        treeView.setEditable(false);

        treeView.setCellFactory(CheckBoxTreeCell.<String>forTreeView());

        for (FXRspecNode n : nodes) {
            CheckBoxTreeItem<String> cb = new CheckBoxTreeItem<>(n.getClientId());
            rootItem.getChildren().add(cb);
            nodesMap.put(cb, n);
        }

        tagTextField.setText("Instant command " + controller.getModel().nextInstantCommandNumber());

        setOnKeyPressed(keyEvent -> {
            if (keyEvent.getCode() == KeyCode.ENTER) {
                onOKAction();
                keyEvent.consume();
            }
        });

        Platform.runLater(tagTextField::requestFocus);

        //initialize dialog
        dialog = new Dialog(null, "Instant command");
        dialog.setMasthead("Execute a command instantly and off the record on the selected nodes");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.BOLT, Color.YELLOW));
        dialog.setContent(this);

        okAction =
                new DialogAction("OK", ButtonBar.ButtonType.OK_DONE, false, false, true, event -> onOKAction());
        dialog.getActions().setAll(okAction, Dialog.ACTION_CANCEL);


    }

    public static void showSimpleInstantCommandDialog(ControllerController controller, List<FXRspecNode> nodes) {
        InstantCommandDialog npd = new InstantCommandDialog(controller, nodes);
        npd.showDialog();
    }

    private void showDialog() {
        dialog.show();
    }


    public void onOKAction() {
        List<String> selectedNodes = nodesMap.entrySet().stream().filter(entry -> entry.getKey().isSelected())
                .map(entry -> entry.getValue().getUniqueId())
                .collect(Collectors.toList());

        if(selectedNodes.isEmpty()){
            JFDialogs.create()
                    .message("Please select at least one node on which the command should be executed")
                    .showWarning();
            return;
        }

        controller.instantCommand(tagTextField.getText(), commandTextField.getText(), selectedNodes);

        dialog.setResult(okAction);
        dialog.hide();

    }
}
