package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import com.cathive.fx.guice.FXMLComponent;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Dialog;
import javafx.scene.control.DialogPane;
import javafx.scene.control.ListView;
import javafx.scene.control.cell.CheckBoxListCell;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.util.Callback;
import javafx.util.StringConverter;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/5/13
 * Time: 9:21 PM
 */
@FXMLComponent
public class SelectAuthoritiesDialog extends VBox {
    private static final Logger LOG = LoggerFactory.getLogger(SelectAuthoritiesDialog.class);

    private static final String SELECT_AUTHORITIES_DIALOG_FXML = "SelectAuthoritiesDialog.fxml";

    private final ObservableList<SelectableAuthority> selectableAuthorities = FXCollections.observableArrayList();

    private final IntegerProperty selectedAuthoritiesCount = new SimpleIntegerProperty();
    private final Set<SfaAuthority> selectedUsedAuthorities = new HashSet<>();

    @FXML
    private ListView<SelectableAuthority> authoritiesListView;
    private Dialog<Set<SfaAuthority>> dialog;

    @Inject
    private AuthorityListModel authorityListModel;
    @Inject
    private AuthorityFinder authorityFinder;

    SelectAuthoritiesDialog(){}

    @FXML
    private void initialize(){

        selectableAuthorities.addAll(
                authorityListModel.getAuthorities().stream()
                        .map(SelectableAuthority::new)
                        .sorted(Comparator.comparing(auth -> auth.getExactSfaAuthority().getHrn().toLowerCase()))
                        .collect(Collectors.toList()));

        authoritiesListView.setItems(selectableAuthorities);


        authoritiesListView.setCellFactory(
                CheckBoxListCell.forListView(SelectableAuthority::selectedProperty,
                        new StringConverter<SelectableAuthority>() {
                            @Override
                            public String toString(SelectableAuthority selectableAuthority) {
                                return selectableAuthority.getExactSfaAuthority().getHrn();
                            }

                            @Override
                            public SelectableAuthority fromString(String s) {
                                return null;
                            }
                        }));

        DialogPane dialogPane = new DialogPane();
        dialogPane.setContent(this);
        dialogPane.setHeaderText("Select the involved authorities");
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.AMBULANCE, Color.GREEN));
        dialogPane.getButtonTypes().setAll(ButtonType.OK, ButtonType.CANCEL);

        Node okButton = dialogPane.lookupButton(ButtonType.OK);

        okButton.disableProperty().bind(selectedAuthoritiesCount.isEqualTo(0));


        dialog = new Dialog<>();
       // dialog.initOwner(experimenterGUI.getScene().getWindow());
        dialog.setDialogPane(dialogPane);
        dialog.setResizable(false);

        dialog.setResultConverter(new Callback<ButtonType, Set<SfaAuthority>>() {
            @Override
            public Set<SfaAuthority> call(ButtonType buttonType) {
                if (buttonType == ButtonType.CANCEL)
                    return null;
                else {
                    return selectableAuthorities.stream().filter(SelectableAuthority::getSelected).map(SelectableAuthority::getUsedSfaAuthority).collect(Collectors.toSet());
                }
            }
        });
    }

    public Optional<Set<SfaAuthority>> showDialog() {
        return dialog.showAndWait();
    }

    public Set<SfaAuthority> getSelectedUsedAuthorities() {
        return Collections.unmodifiableSet(selectedUsedAuthorities);
    }


    private class SelectableAuthority {
        private final BooleanProperty selected = new SimpleBooleanProperty(false);
        private final SfaAuthority sfaAuthority;

        private SelectableAuthority(SfaAuthority sfaAuthority) {
            this.sfaAuthority = sfaAuthority;

            selected.addListener(new ChangeListener<Boolean>() {
                @Override
                public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue) {
                    if (newValue != oldValue) {
                        if (newValue) {
                            selectedAuthoritiesCount.set(selectedAuthoritiesCount.get() + 1);
                        } else {
                            selectedAuthoritiesCount.set(selectedAuthoritiesCount.get() - 1);
                        }
                    }
                }
            });

        }

        public boolean getSelected() {
            return selected.get();
        }

        public BooleanProperty selectedProperty() {
            return selected;
        }

        public void setSelected(boolean selected) {
            this.selected.set(selected);
        }

        public SfaAuthority getExactSfaAuthority() {
            return sfaAuthority;
        }

        public SfaAuthority getUsedSfaAuthority() {
            assert authorityFinder != null;
            assert sfaAuthority != null;
            assert sfaAuthority.getUrn() != null;
            SfaAuthority res = authorityFinder.findByUrn(sfaAuthority.getUrn(), AuthorityFinder.Purpose.RECOVER);
            if (res == null)
                LOG.warn("Could not find authority for recover: \""+sfaAuthority.getUrn()+"\"");
            return res;
        }
    }
}
