package be.iminds.ilabt.jfed.experimenter_gui.editor;

import be.iminds.ilabt.jfed.experimenter_gui.editor.ribbon_tabs.ControlEditorRibbonTab;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ribbon_tabs.RspecEditorRibbonTab;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ribbon_tabs.TopologyEditorRibbonTab;
import be.iminds.ilabt.jfed.experimenter_gui.tabs.RibbonEnabled;
import be.iminds.ilabt.jfed.experimenter_gui.tabs.StatusEnabled;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonTab;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import com.cathive.fx.guice.GuiceFXMLLoader;
import javafx.beans.property.StringProperty;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.MenuItem;
import javafx.scene.control.Tab;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.glyphfont.Glyph;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * User: twalcari
 * Date: 11/7/13
 * Time: 4:58 PM
 */
public class ExperimentEditorTab extends Tab implements RibbonEnabled, StatusEnabled {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentEditorTab.class);


    private static List<RibbonTab> ribbonTabs;
    private final ExperimentEditor experimentEditor;

    private static final RspecEditorRibbonTab rspecEditorRibbonTab;
    private static final TopologyEditorRibbonTab topologyEditorRibbonTab;
    private static final ControlEditorRibbonTab controlEditorRibbonTab;

    private final GuiceFXMLLoader loader;

    static {
        ribbonTabs = new ArrayList<>();

        rspecEditorRibbonTab = new RspecEditorRibbonTab();
        topologyEditorRibbonTab = new TopologyEditorRibbonTab(rspecEditorRibbonTab);
        controlEditorRibbonTab = new ControlEditorRibbonTab(rspecEditorRibbonTab);

        ribbonTabs.add(topologyEditorRibbonTab);
        ribbonTabs.add(rspecEditorRibbonTab);
        ribbonTabs.add(controlEditorRibbonTab);
    }

    ExperimentEditorTab(RequestRspecSource requestRspecSource,
                               String name, File file, GuiceFXMLLoader loader) {
        this.loader = loader;

        getStyleClass().add("tab-experiment-design");

        //FIXME Temporary fix until https://bitbucket.org/controlsfx/controlsfx/pull-request/452/fix-edit-and-pencil_square_alt-fontawesome/diff is accepted
        setGraphic(new Glyph("FontAwesome", '\uf044')); // GlyphUtils.createGlyph(FontAwesome.Glyph.PENCIL_SQUARE_ALT));

        ContextMenu contextMenu = new ContextMenu();
        MenuItem changeNameMenuItem = new MenuItem("Change name", GlyphUtils.createGlyph(FontAwesome.Glyph.PENCIL));
        changeNameMenuItem.setOnAction(actionEvent -> showEditNameDialog());
        contextMenu.getItems().add(changeNameMenuItem);
        setContextMenu(contextMenu);

        //create the content by loading the experiment-editor-fxml

        GuiceFXMLLoader.Result result;
        ExperimentEditor experimentEditor = null;
        try {
            result = loader.load(ExperimentEditor.class.getResource("ExperimentEditor.fxml"));

            experimentEditor = result.getController();
            experimentEditor.setContent(requestRspecSource, name, file);
            setContent(result.getRoot());
            textProperty().bind(experimentEditor.nameProperty());
        } catch (IOException e) {
            LOG.error("Error loading Experiment Editor", e);
        }
        this.experimentEditor = experimentEditor;
    }

    private void showEditNameDialog() {
        JFDialogs.create()
                .owner(null)
                .message("Enter a new name:")
                .masthead("Change Experiment Definition name")
                .title("Change Experiment Definition name")
                .showTextInput(this.experimentEditor.getName()).filter(x -> !x.isEmpty())
                .ifPresent(experimentEditor::setName);
    }

    public ExperimentEditor getExperimentEditor() {
        return experimentEditor;
    }

    @Override
    public List<? extends RibbonTab> getRibbonTabs() {
        return ribbonTabs;
    }

    @Override
    public RibbonTab getInitialTabToActivate() {
        if (experimentEditor.isRawRspecEditorVisible())
            return rspecEditorRibbonTab;
        if (experimentEditor.isExperimentControllerVisible())
            return controlEditorRibbonTab;

        //use default: topology
        return topologyEditorRibbonTab;
    }

    @Override
    public StringProperty statusProperty() {
        return experimentEditor.statusProperty();
    }
}
