package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.NodeDescription;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.bonfire.BonfireNodePropertiesTab;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManager;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManagerFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.ComponentManagerInfo;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.model.EasyOcciModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.SliverType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.*;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Pane;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * User: twalcari
 * Date: 4/10/2014
 * Time: 11:55
 */
public class AdaptivePropertiesDialog extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(AdaptivePropertiesDialog.class);
    private static final String ADAPTIVE_PROPERTIES_DIALOG_FXML = "AdaptivePropertiesDialog.fxml";

    private final JFedConfiguration config;
    private final AuthorityList authorityList;

    private final FXRspecNode node;
    private final ModelRspecEditor modelRspecEditor;

    private final List<AdaptivePropertiesTab> availableTabs;
    private final CMIManager cmiManager;


    @FXML
    private VBox propertiesBox;

    @FXML
    private Pane diskImageBox;
    @FXML
    private Button addDiskImageButton;

    @FXML
    private TextField nameTextField;

    @FXML
    private Button saveButton;
    @FXML
    private Button cancelButton;
    @FXML
    private ComboBox<ComponentManagerInfo> componentManagerInfoComboBox;
    @FXML
    private Button testbedsMapButton;

    @FXML
    private TextField diskImageTextField;

    @FXML
    private TabPane tabPane;

    private final Map<ComponentManagerInfo, ComponentManagerPropertiesPane> cmiPropertiesPanes = new HashMap<>();
    private final ReadOnlyObjectProperty<ComponentManagerInfo> currentCmi;
    private final CMIManagerFactory cmiManagerFactory;
    private final MapDialogFactory mapDialogFactory;
    private final EasyOcciModel easyOcciModel;
    private final HighLevelController hlc;
    private final TasksFactory tasksFactory;
    private final TaskService taskService;


    AdaptivePropertiesDialog(ModelRspecEditor modelRspecEditor, final FXRspecNode node, JFedConfiguration config,
                             AuthorityList authorityList, CMIManagerFactory cmiManagerFactory, MapDialogFactory mapDialogFactory, EasyOcciModel easyOcciModel,
                             HighLevelController hlc, TasksFactory tasksFactory, TaskService taskService) {
        this.config = config;
        this.authorityList = authorityList;
        this.cmiManagerFactory = cmiManagerFactory;
        this.mapDialogFactory = mapDialogFactory;
        this.easyOcciModel = easyOcciModel;
        this.hlc = hlc;
        this.tasksFactory = tasksFactory;
        this.taskService = taskService;
        assert (modelRspecEditor != null);
        assert (node != null);

        this.modelRspecEditor = modelRspecEditor;
        this.node = node;

        this.cmiManager = cmiManagerFactory.getCMIManager(node);

        //initialize the Scene
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(ADAPTIVE_PROPERTIES_DIALOG_FXML));

        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }

        //populate tabPane
        availableTabs = new ArrayList<>();
        BonfireNodePropertiesTab bonfireNodePropertiesTab =
                new BonfireNodePropertiesTab(this, this.easyOcciModel, authorityList, hlc);
        if (node instanceof BonfireRspecNode)
            bonfireNodePropertiesTab.load((BonfireRspecNode) node);
        availableTabs.add(bonfireNodePropertiesTab);


        BootScriptsTab bootScriptsTab = new BootScriptsTab(this);
        bootScriptsTab.load(node);
        availableTabs.add(bootScriptsTab);

        ExoGeniConfigTab exoGeniConfigTab = new ExoGeniConfigTab(this);
        exoGeniConfigTab.load(node);
        availableTabs.add(exoGeniConfigTab);

        RoutableIPTab routableIPTab = new RoutableIPTab(this);
        routableIPTab.load(node);
        availableTabs.add(routableIPTab);

        XenTab xenTab = new XenTab(this);
        xenTab.load(node);
        availableTabs.add(xenTab);

        //add tab for each interface
        for (FXRspecInterface iface : node.getInterfaces()) {
            tabPane.getTabs().add(new InterfacePropertiesTab(modelRspecEditor, node, iface));
        }


        nameTextField.setTooltip(
                TooltipBuilder.create()
                        .text("Only alphanumerical, max. " + getMaximumNodeNameLength() + " characters")
                        .build()
        );

        componentManagerInfoComboBox.setItems(cmiManager.getComponentManagerInfos());
        componentManagerInfoComboBox.setConverter(StringConverters.CMI_STRING_CONVERTER);
        this.currentCmi = componentManagerInfoComboBox.getSelectionModel().selectedItemProperty();

        this.currentCmi.addListener(new ChangeListener<ComponentManagerInfo>() {
            @Override
            public void changed(ObservableValue<? extends ComponentManagerInfo> observableValue,
                                ComponentManagerInfo oldValue, ComponentManagerInfo newValue) {

                boolean activeNodesOnly = true;

                if (oldValue != null) {
                    final JFedConfiguration.TestbedDescription testbedDescription =
                            AdaptivePropertiesDialog.this.config.getTestbedDescription(oldValue.getAuthority().getUrn());

                    if (testbedDescription != null) {

                        List<JFedConfiguration.SliverType> sliverTypes = testbedDescription.getSliverTypes();
                        if (!sliverTypes.isEmpty() &&
                                sliverTypes.get(0).getDiskImageName() != null &&
                                sliverTypes.get(0).getDiskImageName().equals(diskImageTextField.getText())) {
                            diskImageTextField.setText("");
                        }
                    } else {
                        LOG.warn("No testbed description available for " + oldValue.getAuthority().getUrn());
                    }
                    activeNodesOnly = cmiPropertiesPanes.get(oldValue).isActiveNodesOnly();
                }

                if (newValue != null) {
                    if (!cmiPropertiesPanes.containsKey(newValue)) {
                        cmiPropertiesPanes.put(currentCmi.get(), new ComponentManagerPropertiesPane(newValue, node,
                                activeNodesOnly, cmiManager, config, mapDialogFactory));
                    }
                    propertiesBox.getChildren().add(cmiPropertiesPanes.get(newValue));


                    JFedConfiguration.TestbedDescription testbedDescription =
                            AdaptivePropertiesDialog.this.config.getTestbedDescription(newValue.getAuthority().getUrn());

                    if (testbedDescription != null) {
                        List<JFedConfiguration.SliverType> sliverTypes = testbedDescription.getSliverTypes();

                        //mobile12, zotacH1
                        if (!sliverTypes.isEmpty() &&
                                sliverTypes.get(0).getDiskImageName() != null &&
                                diskImageTextField.getText().isEmpty()) {
                            diskImageTextField.setText(sliverTypes.get(0).getDiskImageName());
                        }
                    } else {
                        LOG.warn("No testbed description available for " + newValue.getAuthority().getUrn());
                    }
                }

                if (oldValue != null) {
                    propertiesBox.getChildren().remove(cmiPropertiesPanes.get(oldValue));
                }

                if (newValue != null) {
                    //only show disk image button for instageni and emulab-testbeds
                    SfaAuthority authority = newValue.getAuthority();
                    boolean showDiskImageDialog = authority.isType("instageni") || authority.isType("emulab")
                            || AdaptivePropertiesDialog.this.config.getSupportsListDiskImages(authority.getUrn());
                    addDiskImageButton.setVisible(showDiskImageDialog);
                }

                determineTabsVisibility(node);
            }
        });

        //set initial values to these of the received node
        nameTextField.setText(node.getClientId());

        ComponentManagerInfo componentManagerInfo =
                node.getComponentManagerId() != null ?
                        cmiManager.getComponentManagerInfo(node.getComponentManagerId()) :
                        null;

        if (componentManagerInfo == null) {
            //no component manager: select nothing
            componentManagerInfoComboBox.getSelectionModel().clearSelection();
        } else {
            //select the current testbed in the list
            componentManagerInfoComboBox.getSelectionModel().select(componentManagerInfo);
        }

        if (!node.getSliverTypes().isEmpty()) {
            List<DiskImage> diskImages = node.getSliverTypes().get(0).getDiskImages();

            if (!diskImages.isEmpty())
                diskImageTextField.setText(diskImages.get(0).getName() != null ?
                        diskImages.get(0).getName() :
                        diskImages.get(0).getUrl());
        }

        diskImageBox.managedProperty().bind(diskImageBox.visibleProperty());
        addDiskImageButton.managedProperty().bind(addDiskImageButton.visibleProperty());

        determineTabsVisibility(node);

        //only show maps-button if running on a 64-bit JVM (see FEDIBBTDEV-1344)
        String arch = System.getProperty("os.arch");

        if (arch.equals("x86")) {
            testbedsMapButton.setVisible(false);
        }


    }

    @FXML
    private void showTestbedsMap() {
        ComponentManagerInfo selectedCmi =
                mapDialogFactory.showTestbedMapDialog(getScene().getWindow(), cmiManager);
        if (selectedCmi != null)
            componentManagerInfoComboBox.getSelectionModel().select(selectedCmi);
    }

    /**
     * This function shows and hides tabs as needed. It is called by evenListeners on the testbeds- and node-comboboxes.
     */
    private void determineTabsVisibility(FXRspecNode rspecNode) {
        int insertPosition = 1;
        for (AdaptivePropertiesTab apt : availableTabs) {
            if (apt.shouldBeVisible(rspecNode)) {
                LOG.debug("Tab {} should be made visible.", apt.getClass().getSimpleName());
                if (tabPane.getTabs().contains(apt)) {
                    insertPosition++;
                } else {
                    tabPane.getTabs().add(insertPosition, apt);
                }
            } else {
                LOG.debug("Tab {} should be made invisible.", apt.getClass().getSimpleName());
                if (tabPane.getTabs().contains(apt)) {
                    tabPane.getTabs().remove(apt);
                }
            }
        }
    }

    private Class<? extends FXRspecNode> determineRspecNodeClass() {
        if (componentManagerInfoComboBox.getSelectionModel().getSelectedItem() == null)
            return FXRspecNode.class;

        if (componentManagerInfoComboBox.getSelectionModel().getSelectedItem()
                .getAuthority().getUrn().equals(BonfireNodePropertiesTab.BONFIRE_URN)) {
            return BonfireRspecNode.class;
        } else {
            return FXRspecNode.class;
        }

    }

    @FXML
    @SuppressWarnings("unchecked")
    public void onSaveAction(ActionEvent actionEvent) {
        if (!isValidNodeName(nameTextField.getText())) {
            JFDialogs.create()
                    .owner(this)
                    .message("A node name may only contain alphanumerical characters, and is maximum " + getMaximumNodeNameLength() + " characters long.")
                    .masthead("The entered node name is invalid.")
                    .showError();
            return;
        }

        if (currentCmi.get() == null) {
            JFDialogs.create()
                    .owner(this)
                    .message("You must select an authority for this node.")
                    .masthead("No authority selected")
                    .title("No authority selected")
                    .showError();
            return;
        }


        if (!diskImageTextField.getText().isEmpty()) {
            //first remove any extra spaces that have slipped in
            diskImageTextField.setText(diskImageTextField.getText().trim().replace(" ", ""));

            String diskImageValue = diskImageTextField.getText();

            //now check if it is a valid URN or HTTP(S)-link

            if (diskImageValue.startsWith("urn")) {
                GeniUrn diskImageUrn = GeniUrn.parse(diskImageValue);

                if (diskImageUrn == null || !"image".equals(diskImageUrn.getEncodedResourceType())) {
                    JFDialogs.create()
                            .owner(this)
                            .message("The provided disk image URN is invalid. Please correct the entered value.")
                            .masthead("Invalid disk image URN")
                            .showError();

                    LOG.warn("User entered invalid disk image URN '{}'", diskImageValue);

                    return;
                }

            } else {
                //we expect a valid URL instead
                try {
                    new URL(diskImageValue);
                } catch (MalformedURLException e) {
                    JFDialogs.create()
                            .owner(this)
                            .message("The provided disk image URL is invalid. Please correct the entered value.")
                            .masthead("Invalid disk image URL")
                            .showError();


                    LOG.warn("User entered invalid disk image URL '{}'", diskImageValue);

                    return;
                }
            }
        }

        ComponentManagerPropertiesPane componentManagerPropertiesPane = cmiPropertiesPanes.get(currentCmi.get());

        if (!componentManagerPropertiesPane.validate())
            return;


        for (AdaptivePropertiesTab visibleTab : Iterables.filter(tabPane.getTabs(), AdaptivePropertiesTab.class)) {
            String validation = visibleTab.validateConfiguration();

            if (validation != null) {
                JFDialogs.create()
                        .owner(this)
                        .message(validation)
                        .masthead("Node configuration could not be saved")
                        .showError();
                return;
            }

        }


        for (PropertiesTab visibleTab : Iterables.filter(tabPane.getTabs(), PropertiesTab.class)) {
            String validation = visibleTab.validateConfiguration();

            if (validation != null) {
                JFDialogs.create()
                        .owner(this)
                        .message(validation)
                        .masthead("Node configuration could not be saved")
                        .showError();
                return;
            }

        }


        FXRspecNode currentNode = this.node;

        //determine if the implementation-class of the node has to be changed
        FXModelRspec model = modelRspecEditor.getModelRspec();
        Class<? extends FXRspecNode> newNodeClazz = determineRspecNodeClass();
        if (!currentNode.getClass().equals(newNodeClazz)) {
            //change the node implementation class by calling the copy constructor

            try {
                Constructor<? extends FXRspecNode> c =
                        newNodeClazz.getConstructor(FXModelRspec.class, FXRspecNode.class,
                                RspecNode.InterfaceCopyMethod.class);
                FXRspecNode newNode =
                        c.newInstance(currentNode.getRspec(), currentNode, RspecNode.InterfaceCopyMethod.STEAL_INTERFACE);

                model.getNodes().add(newNode);
                model.deleteNode(currentNode);

                currentNode = newNode;


            } catch (NoSuchMethodException e) {
                LOG.error("Could not find copy constructor for class " + newNodeClazz.getName());
                throw new RuntimeException("Could not find copy constructor for class " + newNodeClazz.getName(), e);
            } catch (InvocationTargetException | InstantiationException | IllegalAccessException e) {
                LOG.error("An error occurred when calling the copy constructor", e);
                throw new RuntimeException("An error occurred when calling the copy constructor", e);

            }
        }

        currentNode.setClientId(nameTextField.getText());
        ComponentManagerInfo cmi = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();

        boolean componentManagerChanged = !cmi.getAuthority().getUrn().equals(currentNode.getComponentManagerId());
        currentNode.setComponentManagerId(cmi.getAuthority().getUrn());

        //force SliverType to an existing type on the testbed
        //NOTE: this is a good default, that gets overwritten when the ExoGeni tab is present and is saved later in this step
        if (!cmi.hasSliverType(currentNode.getSliverTypeName()) && cmi.getDefaultSliverType() != null) {
            SliverType chosenNewType = cmi.getDefaultSliverType();
            if (chosenNewType != null) {

                if (!cmi.getSliverTypes().isEmpty() && !cmi.hasSliverType(chosenNewType.getName())) {
                    LOG.warn("Default sliver_type for " + cmi.getAuthority().getUrn() + " is \"" + chosenNewType.getName() + "\", but that sliver_type is not in ListResources. Picking an existing one instead.");
                    chosenNewType = null;
                } else {
                    LOG.warn("Default sliver_type for " + cmi.getAuthority().getUrn() + " is \"" + chosenNewType.getName() + "\", but that sliver_type is not in ListResources. However, no alternative is found in ListResources...");
                }
            }

            if (chosenNewType == null && !cmi.getSliverTypes().isEmpty())
                chosenNewType = cmi.getSliverTypes().get(0); //choose first (= quite randomly)

            if (chosenNewType != null)
                currentNode.setSliverTypeName(chosenNewType.getName());
        }

        currentNode.setComponentId(componentManagerPropertiesPane.getSelectedNode() != null &&
                componentManagerPropertiesPane.getSelectedNode().getComponentId() != null ?
                componentManagerPropertiesPane.getSelectedNode().getComponentId()
                : null);


        if (componentManagerChanged) {
            List<FXRspecLink> toRemove = new ArrayList<>();

            if (!cmi.isLinkSupport() && node.getInterfaces().size() > 0) {
                JFDialogs.create()
                        .owner(this)
                        .message(String.format("'%s' does not support dedicated links. Existing links to '%s' are being removed.",
                                cmi.getAuthority().getName(),
                                node.getClientId()))
                        .masthead("Illegal links detected")
                        .title("Illegal links detected")
                        .showWarning();

                toRemove.addAll(Lists.transform(node.getInterfaces(),
                        rspecInterface -> rspecInterface.isLinkBound() ? rspecInterface.getLink() : null
                ));
            } else {
                for (FXRspecInterface iface : currentNode.getInterfaces()) {
                    if (iface.isLinkBound() && iface.getLink().isStitched(authorityList.getAuthorityListModel())) {

                        if (JFDialogs.create()
                                .owner(this)
                                .message(
                                        "Link '" + iface.getLink().getClientId() + "' is now a stitched link.\n\n" +
                                                "Stitched links between testbeds are currently VERY experimental and not supported." +
                                                " Do you want to remove this link?")
                                .masthead("Experimental functionality detected.")
                                .title("Experimental functionality detected.")
                                .buttonTypes(ButtonType.YES, ButtonType.NO)
                                .showWarning().filter(r -> r == ButtonType.YES).isPresent())
                            toRemove.add(iface.getLink());
                        else {
                            modelRspecEditor.convertLinkToStitchedLink(iface.getLink());
                        }

                    } else {
                        if (iface.isLinkBound()) {
                            //the link used to be stitched, but isn't anymore
                            JFDialogs.create()
                                    .owner(this)
                                    .message("Link '" + iface.getLink().getClientId() + "' is now a normal link." +
                                            " Some settings are reverted to default.")
                                    .masthead("Link type change detected")
                                    .title("Link type change detected")
                                    .showInformation();

                            modelRspecEditor.convertLinkToNormalLink(iface.getLink());
                        }
                    }
                }
            }
            toRemove.forEach(model::deleteLink);
        }

        if (!currentNode.getSliverTypes().isEmpty()) {
            SliverType sliverType = currentNode.getSliverTypes().get(0);

            if (sliverType != null) {
                assert diskImageTextField != null;
                if (diskImageTextField.getText() != null) {
                    String diskImageString = diskImageTextField.getText().trim();

                    if (!diskImageString.isEmpty()) {
                        currentNode.setSliverTypeDiskImage(diskImageString);
                    } else {
                        currentNode.setSliverTypeDiskImage(null);
                    }
                } else {
                    currentNode.setSliverTypeDiskImage(null);
                }
            }
        } else {
            LOG.error("Could not find sliver-type on currentNode: cannot save disk image");
        }

        //check if we need to change the node exclusivity
        Boolean exclusive =
                config.getNodeExclusivity(currentNode.getComponentManagerId(), currentNode.getSliverTypeName());

        if (exclusive != null) {
            currentNode.setExclusive(exclusive);
        }

        final FXRspecNode resultNode = currentNode;

        tabPane.getTabs().stream()
                .filter(tab -> tab instanceof AdaptivePropertiesTab)
                .map(tab -> (AdaptivePropertiesTab) tab).forEach(tab -> {
            assert (tab.shouldBeVisible(resultNode)) :
                    tab.getClass().getName() + "indicates it shouldn't be visible at this time!";

            //we assume here that when the tab is visible, it is also able to save it's information
            // to the new FXRspecNode-class!
            // THIS IS POTENTIALLY UNSAFE!
            tab.save(resultNode);
        });


        tabPane.getTabs().stream()
                .filter(tab -> tab instanceof PropertiesTab)
                .map(tab -> (PropertiesTab) tab)
                .forEach(PropertiesTab::save);


        NodeDescription nodeDescription = config.detectJFedNodeType(currentNode);
        if (nodeDescription != null)
            currentNode.setNodeDescriptionId(nodeDescription.getId());


        Stage stage = (Stage) saveButton.getScene().getWindow();
        stage.close();
    }

    String getCurrentSliverTypeName() {

        //get the CMI for the authority that is currently visible
        ComponentManagerInfo cmi = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();

        if (node.getSliverTypeName() != null && cmi.hasSliverType(node.getSliverTypeName())) {
            return node.getSliverTypeName();
        }
        //fallback onto the default sliver-type defined in the CMI
        SliverType fallbackSliverType = cmi.getDefaultSliverType();

        if (fallbackSliverType != null && cmi.hasSliverType(fallbackSliverType.getName())) {
            return fallbackSliverType.getName();
        }
        LOG.warn("Default sliver_type for {} is '{}', but that sliver_type is not in ListResources. " +
                        "Picking an existing one instead.",
                cmi.getAuthority().getHrn(), fallbackSliverType);

        if (!cmi.getSliverTypes().isEmpty())
            return cmi.getSliverTypes().get(0).getName(); //choose first (= quite randomly)
        else
            return null;
    }

    private int getMaximumNodeNameLength() {
        return 10;
    }

    private boolean isValidNodeName(String text) {
        if (text.length() > getMaximumNodeNameLength())
            return false;

        for (char c : text.toCharArray()) {
            boolean validChar = (c >= 'a' && c <= 'z')
                    || (c >= 'A' && c <= 'Z')
                    || (c >= '0' && c <= '9');

            if (!validChar) {
                return false;
            }
        }

        return true;
    }

    @FXML
    public void onCancelAction(ActionEvent actionEvent) {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        stage.close();
    }

    @FXML
    private void onRequestDiskImageAction() {

        String newDiskImage =
                new SelectDiskImageDialog(this, componentManagerInfoComboBox.getValue(), cmiManager, hlc, tasksFactory, taskService).showDialog();

        if (newDiskImage != null) {
            diskImageTextField.setText(newDiskImage);
        }
    }

    /**
     * @return may return null if nothing is selected
     */
    public ComponentManagerInfo getSelectedComponentManagerInfo() {
        return currentCmi.get();
    }

    public FXRspecNode getSelectedComponentId() {
        return cmiPropertiesPanes.containsKey(currentCmi.get()) ?
                cmiPropertiesPanes.get(currentCmi.get()).getSelectedNode() : null;
    }


}
