package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.EditableExperimentCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.FXUtils;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.DelayRspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXLinkSetting;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.binding.Bindings;
import javafx.beans.property.*;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.control.cell.TextFieldTableCell;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import javafx.stage.Window;
import javafx.util.converter.NumberStringConverter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;

/**
 * This dialog is meant to act upon the FXLinkSetting-variables.
 * In practice however, these settings do not work. Because of that, this dialog is not in use.
 * <p/>
 * User: twalcari
 * Date: 11/28/13
 * Time: 5:03 PM
 */
public class DelayNodePropertiesDialog extends BorderPane {

    private static final Logger LOG = LoggerFactory.getLogger(DelayNodePropertiesDialog.class);
    private static final String DELAY_NODE_PROPERTIES = "DelayNodeProperties.fxml";

    private final DelayRspecNode node;
    private final FXModelRspec model;

    private final ListProperty<FXLinkSetting> impairmentList;
    private final LongProperty capacity = new SimpleLongProperty();
    private final IntegerProperty latency = new SimpleIntegerProperty();
    private final DoubleProperty packetLoss = new SimpleDoubleProperty();
    @FXML
    private Button saveButton;
    @FXML
    private Button cancelButton;
    @FXML
    private TextField nameTextField;
    @FXML
    private ToggleGroup impairmentToggleGroup;
    @FXML
    private RadioButton simpleLinkImpairmentRadioButton;
    @FXML
    private TextField capacityTextField;
    @FXML
    private TextField latencyTextField;
    @FXML
    private TextField packetLossTextField;
    @FXML
    private CheckBox capacityCheckBox;
    @FXML
    private CheckBox latencyCheckBox;
    @FXML
    private CheckBox packetLossCheckBox;

    @FXML
    private RadioButton advancedLinkImpairmentRadioButton;
    @FXML
    private TableView<FXLinkSetting> advancedImpairmentTableView;

    @FXML
    private TableColumn<FXLinkSetting, String> fromTableColumn;
    @FXML
    private TableColumn<FXLinkSetting, String> toTableColumn;
    @FXML
    private TableColumn<FXLinkSetting, Number> capacityTableColumn;
    @FXML
    private TableColumn<FXLinkSetting, Number> latencyTableColumn;
    @FXML
    private TableColumn<FXLinkSetting, Number> packetLossTableColumn;

    private DelayNodePropertiesDialog(FXModelRspec model, DelayRspecNode node) {
        this.model = model;
        this.node = node;

        impairmentList = node.getLinkSettings();

        //initialize the Scene
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(DELAY_NODE_PROPERTIES));

        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }

        nameTextField.setText(node.getClientId());

        //initialize simple values on first LinkSetting-object values

        FXLinkSetting ls = node.getLinkSettings().get(0);

        capacity.setValue(ls.getCapacity_Kbps());
        latency.setValue(ls.getLatency_ms());
        packetLoss.setValue(ls.getPacketLoss());

        if (hasAdvancedImpairmentSettings())
            impairmentToggleGroup.selectToggle(advancedLinkImpairmentRadioButton);


        //initialize simple values
        capacityTextField.textProperty().bindBidirectional(capacity, new NumberStringConverter());

        capacityTextField.focusedProperty().addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                if (!capacityTextField.isFocused()) {
                    if (capacity.get() < 0) {
                        JFDialogs.create()
                                .owner(advancedImpairmentTableView.getScene().getWindow())
                                .message("Capacity must be a non-negative value")
                                .masthead("Invalid Capacity Loss value")
                                .showError();
                        capacity.setValue(0);
                    }

                }
            }
        });
        capacityCheckBox.selectedProperty().bind(Bindings.notEqual(0, capacity));

        capacityCheckBox.selectedProperty().addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                LOG.debug("new checkbox value:" + capacityCheckBox.isSelected());
            }
        });

        latencyTextField.textProperty().bindBidirectional(latency, new NumberStringConverter());

        latencyTextField.focusedProperty().addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                if (!latencyTextField.isFocused()) {
                    if (latency.get() < 0) {
                        JFDialogs.create()
                                .owner(advancedImpairmentTableView.getScene().getWindow())
                                .message("Latency must be a non-negative value")
                                .masthead("Invalid Latency value")
                                .showError();
                        latency.setValue(0);
                    }

                }
            }
        });

        latencyCheckBox.selectedProperty().bind(Bindings.notEqual(0, latency));

        packetLossTextField.textProperty().bindBidirectional(packetLoss, new NumberStringConverter());

        packetLossTextField.focusedProperty().addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                if (!packetLossTextField.isFocused()) {
                    if (packetLoss.get() < 0 || packetLoss.get() > 1) {
                        JFDialogs.create()
                                .owner(advancedImpairmentTableView.getScene().getWindow())
                                .message("Packet loss must be between 0 and 1")
                                .masthead("Invalid Packet Loss value")
                                .showError();
                        packetLoss.setValue(0);
                    }

                }
            }
        });
        packetLossCheckBox.selectedProperty().bind(Bindings.notEqual(0, packetLoss));

        //populate advancedDelayRspecNode.LinkSettingTableView
        fromTableColumn.setCellValueFactory(p -> p.getValue().getFromIface().clientIdProperty());
        toTableColumn.setCellValueFactory(p -> p.getValue().getToIface().clientIdProperty());

        capacityTableColumn.setCellFactory(TextFieldTableCell.<FXLinkSetting, Number>forTableColumn(new NumberStringConverter()));

        capacityTableColumn.setOnEditCommit(e -> {
            if (e.getNewValue().intValue() < 0) {

                JFDialogs.create()
                        .owner(advancedImpairmentTableView.getScene().getWindow())
                        .message("Capacity must be a non-negative value")
                        .masthead("Invalid Capacity value")
                        .showError();

                return;
            }
            e.getRowValue().capacity_KbpsProperty().setValue(e.getNewValue());
        });

        capacityTableColumn.setCellValueFactory(p -> p.getValue().capacity_KbpsProperty());

        latencyTableColumn.setCellFactory(TextFieldTableCell.<FXLinkSetting, Number>forTableColumn(new NumberStringConverter()));

        latencyTableColumn.setOnEditCommit(e -> {
            if (e.getNewValue().intValue() < 0) {

                JFDialogs.create()
                        .owner(advancedImpairmentTableView.getScene().getWindow())
                        .message("Latency must be a non-negative value")
                        .masthead("Invalid Latency value")
                        .showError();
                return;
            }
            e.getRowValue().latency_msProperty().setValue(e.getNewValue());
        });

        latencyTableColumn.setCellValueFactory(p -> p.getValue().latency_msProperty());

        packetLossTableColumn.setCellFactory(TextFieldTableCell.<FXLinkSetting, Number>forTableColumn(new NumberStringConverter()));

        packetLossTableColumn.setOnEditCommit(e -> {
            if (e.getNewValue().doubleValue() < 0 || e.getNewValue().doubleValue() > 1) {
                JFDialogs.create()
                        .owner(advancedImpairmentTableView.getScene().getWindow())
                        .message("Packet loss must be between 0 and 1")
                        .masthead("Invalid Packet Loss value")
                        .showError();
                return;
            }

            e.getRowValue().setPacketLoss(e.getNewValue().doubleValue());
        });

        packetLossTableColumn.setCellValueFactory(p -> p.getValue().packetLossProperty());


        capacityTableColumn.setCellFactory(TextFieldTableCell.<FXLinkSetting, Number>forTableColumn(new NumberStringConverter()));
        packetLossTableColumn.setCellFactory(TextFieldTableCell.<FXLinkSetting, Number>forTableColumn(new NumberStringConverter()));
        latencyTableColumn.setCellFactory(TextFieldTableCell.<FXLinkSetting, Number>forTableColumn(new NumberStringConverter()));


        advancedImpairmentTableView.setItems(impairmentList);

        latencyTextField.disableProperty().bind(Bindings.not(simpleLinkImpairmentRadioButton.selectedProperty()));
        capacityTextField.disableProperty().bind(Bindings.not(simpleLinkImpairmentRadioButton.selectedProperty()));
        packetLossTextField.disableProperty().bind(Bindings.not(simpleLinkImpairmentRadioButton.selectedProperty()));

        advancedImpairmentTableView.disableProperty().bind(Bindings.not(advancedLinkImpairmentRadioButton.selectedProperty()));


        //add link between simple and advanced impairment fields
        capacity.addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                for (FXLinkSetting ls : DelayNodePropertiesDialog.this.node.getLinkSettings()) {
                    ls.setCapacity_Kbps(capacity.get());
                }
            }
        });

        packetLoss.addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                for (FXLinkSetting ls : DelayNodePropertiesDialog.this.node.getLinkSettings()) {
                    ls.setPacketLoss(packetLoss.get());
                }
            }
        });

        latency.addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                for (FXLinkSetting ls : DelayNodePropertiesDialog.this.node.getLinkSettings()) {
                    ls.setLatency_ms(latency.get());
                }
            }
        });


        //add warning when switching from advanced to simple

        impairmentToggleGroup.selectedToggleProperty().addListener(new ChangeListener<Toggle>() {
            @Override
            public void changed(ObservableValue<? extends Toggle> observableValue, Toggle oldValue, Toggle newValue) {
                if (newValue == simpleLinkImpairmentRadioButton && hasAdvancedImpairmentSettings()) {
                    if (JFDialogs.create()
                            .owner(this)
                            .message("Switching to simple impairment-mode will cause your advanced settings to reverse. Are you sure?")
                            .title("Switching to simple impairment mode").showConfirm().filter(r -> r == ButtonType.YES).isPresent()) {
                        resetAdvancedImpairmentSettings();
                    } else {
                        impairmentToggleGroup.selectToggle(advancedLinkImpairmentRadioButton);
                    }
                }
            }
        });

    }

    public static void showNodePropertiesDialog(Window parentWindow, EditableExperimentCanvas editableExperimentCanvas,
                                                DelayRspecNode link) {

        Scene scene = new Scene(new DelayNodePropertiesDialog(editableExperimentCanvas.getModelRspec(), link));

        Stage dialogStage = new Stage();
        dialogStage.setScene(scene);

        dialogStage.setTitle("Properties of " + link.getClientId());
        dialogStage.setResizable(false);
        FXUtils.centerToOwner(parentWindow, dialogStage);
        dialogStage.showAndWait();
    }

    private void resetAdvancedImpairmentSettings() {
        for (FXLinkSetting ls : node.getLinkSettings()) {
            ls.setCapacity_Kbps(capacity.get());
            ls.setLatency_ms(latency.get());
            ls.setPacketLoss(packetLoss.get());
        }

    }

    private boolean hasAdvancedImpairmentSettings() {
        for (FXLinkSetting ls : node.getLinkSettings()) {
            if (ls.getLatency_ms() != latency.get())
                return true;
            if (ls.getCapacity_Kbps() != capacity.get())
                return true;
            if (ls.getPacketLoss() != packetLoss.get())
                return true;
        }
        return false;
    }

    @FXML
    public void onSaveAction(ActionEvent actionEvent) {

        if (!isValidLinkName(nameTextField.getText())) {
            JFDialogs.create()
                    .owner(this)
                    .message("A link name may only contain alphanumerical characters, " +
                            "and is maximum " + getMaximumLinkNameLength() + " characters long.")
                    .masthead("The entered link name is invalid.")
                    .showError();
            return;
        }

        node.setClientId(nameTextField.getText());

        Stage stage = (Stage) saveButton.getScene().getWindow();
        stage.close();
    }

    private int getMaximumLinkNameLength() {
        return 25;
    }

    private boolean isValidLinkName(String text) {
        if (text.length() > getMaximumLinkNameLength())
            return false;

        for (char c : text.toCharArray()) {
            boolean validChar = (c >= 'a' && c <= 'z')
                    || (c >= 'A' && c <= 'Z')
                    || (c >= '0' && c <= '9');

            if (!validChar) {
                return false;
            }
        }

        return true;
    }

    @FXML
    public void onCancelAction(ActionEvent actionEvent) {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        stage.close();
    }
}

