package be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.NodeDescription;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;

import javax.annotation.Nonnull;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 8/28/2014
 * Time: 17:02
 */
public class FilteredCMIManager implements CMIManager {
    private final NodeDescription nodeDescription;
    private final CMIManager cmiManager;

    public FilteredCMIManager(@Nonnull NodeDescription nodeDescription, @Nonnull CMIManager cmiManager) {
        this.nodeDescription = nodeDescription;
        this.cmiManager = cmiManager;
    }

    @Override
    public ObservableList<ComponentManagerInfo> getComponentManagerInfos() {
        ObservableList<ComponentManagerInfo> filteredResult = FXCollections.observableArrayList();

        filteredResult.addAll(cmiManager.getComponentManagerInfos().stream()
                .filter(cmi -> nodeDescription.getEnabledTestbeds().containsKey(cmi.getAuthority().getUrn()))
                .collect(Collectors.toList()));
        return filteredResult;
    }

    @Override
    public ObservableList<FXRspecNode> filterComponentIds(ComponentManagerInfo cmi, ObservableList<FXRspecNode> nodesList) {
        final JFedConfiguration.EnabledTestbed enabledTestbed =
                nodeDescription.getEnabledTestbeds().get(cmi.getAuthority().getUrn());
        if (enabledTestbed == null || enabledTestbed.getEnabledComponentIds().isEmpty())
            return cmiManager.filterComponentIds(cmi, nodesList);

        else {
            return nodesList.filtered(input -> input.getComponentId() != null
                    && enabledTestbed.getEnabledComponentIds().contains(input.getComponentId()));
        }
    }

    @Override
    public ComponentManagerInfo getComponentManagerInfo(GeniUrn componentManagerId) {
        if (nodeDescription.getEnabledTestbeds().containsKey(componentManagerId))
            return cmiManager.getComponentManagerInfo(componentManagerId);
        else
            return null;
    }
}
