package be.iminds.ilabt.jfed.experimenter_gui.editor.ribbon_tabs;

import be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation.RspecCheckReason;
import be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation.RspecValidationEngine;
import be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation.RspecValidationResult;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonButton;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;

import java.io.IOException;

/**
 * User: twalcari
 * Date: 7/1/2014
 * Time: 11:10
 */
public class RspecEditorRibbonTab extends EditorRibbonTab {
    public static final String RSPEC_EDITOR_RIBBON_TAB_FXML = "RspecEditorRibbonTab.fxml";

    private final RspecValidationEngine validationEngine = new RspecValidationEngine();

    @FXML
    private RibbonButton formatCodeButton;
    @FXML
    private RibbonButton verifyCodeButton;
    @FXML
    private RibbonButton searchCodeButton;
    @FXML
    private RibbonButton replaceCodeButton;

    public RspecEditorRibbonTab() {

        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(RSPEC_EDITOR_RIBBON_TAB_FXML));

        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);
        fxmlLoader.setControllerFactory(aClass -> {
            if (aClass == GenericEditorRibbonComponentGroupController.class) {
                return new GenericEditorRibbonComponentGroupController(RspecEditorRibbonTab.this);
            }
            throw new RuntimeException("Unexpected class in ControllerFactory: " + aClass.getName());
        });


        try {
            fxmlLoader.load();
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }

        setOnSelectionChanged(event -> {
            if (isSelected() && !getActiveExperimentEditor().isRawRspecEditorVisible()) {
                getActiveExperimentEditor().switchToRawRspecEditor();
            }
        });

    }

    @FXML
    private void onFormatCodeAction() {
        assert formatCodeButton != null;
        if (!getActiveExperimentEditor().getRawRspecEditor().formatXmlCode())
            JFDialogs.create()
                    .owner(formatCodeButton)
                    .message("Formatting your Rspec failed. Please check if the XML is valid")
                    .title("The Rspec could not be formatted").showWarning();
    }

    @FXML
    private void onVerifyRspecAction() {
        RequestRspecSource source = getActiveExperimentEditor().getRequestRspecSource();
        RspecValidationResult verifyRspecResult =
                validationEngine.validateRSpec(RspecCheckReason.CHECK_RSPEC, source, verifyCodeButton.getScene().getWindow());


        assert verifyCodeButton != null;
        if (verifyRspecResult == null) {
            JFDialogs.create()
                    .owner(verifyCodeButton)
                    .title("Rspec Verify Result")
                    .message("No problems found in RSpec.").showInformation();
        } else {
            getActiveExperimentEditor().getRawRspecEditor().setContent(verifyRspecResult.getReplacementRequestRspecSource().getRspecXmlString());
        }
    }

    @FXML
    private void onSearchAction() {
        getActiveExperimentEditor().getRawRspecEditor().showSearch();
    }

    @FXML
    private void onReplaceAction() {
        getActiveExperimentEditor().getRawRspecEditor().showReplace();
    }


}
