package be.iminds.ilabt.jfed.experimenter_gui.editor.rspec_validation;

import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import javafx.stage.Window;
import org.controlsfx.control.ButtonBar;
import org.controlsfx.control.action.Action;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.dialog.Dialogs;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * User: twalcari
 * Date: 1/22/2015
 * Time: 17:01
 */
public class ExperimentCommandNamespaceValidator implements RspecValidator {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentCommandNamespaceValidator.class);

    @Override
    public RspecValidationResult validateRSpec(RspecCheckReason reason, RequestRspecSource requestRspecSource, Window dialogOwner) {
        if (requestRspecSource.getRspecXmlString().contains("experimentCommand xmlns=\"http://jfed.iminds.be/rspec/ext/jfed/1\"")) {
            String abortText = "Abort";
            String fixText = "Fix and continue";
            String ignoreText = "Keep as is and continue";
            boolean showIgnore = false;
            switch (reason) {
                case SWITCH_TO_VISUAL_EDITOR: {
                    abortText = "Keep as is but return to editor";
                    ignoreText = "Ignore";
                    showIgnore = true;
                    break;
                }
                case RUN_EXPERIMENT: {
                    abortText = "Abort and continue editing";
                    ignoreText = "Ignore and Run Experiment Anyway";
                    fixText = "Fix and Run Experiment";
                    showIgnore = true;
                    break;
                }
                case CHECK_RSPEC: {
                    abortText = "Keep as is";
                    fixText = "Fix";
                    showIgnore = false;
                    break;
                }
                case OPENED_FILE: {
                    abortText = "Keep as is";
                    fixText = "Fix";
                    showIgnore = false;
                    break;
                }
            }

            Action fix = new DialogAction(fixText, ButtonBar.ButtonType.YES, false, true, true);
            Action ignore = new DialogAction(ignoreText, ButtonBar.ButtonType.CANCEL_CLOSE, true, true, false);
            Action abort = new DialogAction(abortText, ButtonBar.ButtonType.CANCEL_CLOSE, false, true, false);

            Dialogs d = Dialogs.create().owner(dialogOwner)
                    .message("Problem in RSpec:\n" +
                                    "Some experimentCommands in the RSpec are from an older jFed version that had a bug causing them to be in the wrong namespace.\n\n" +
                                    "Do you want to fix this by automatically changing to the correct namespace?"
                    )
                    .title("Wrong experimentCommands namespace");
            if (showIgnore)
                d = d.actions(fix, ignore, abort);
            else
                d = d.actions(fix, abort);
            Action chosenAction = d.showError();

            if (chosenAction == fix) {
                LOG.info("Fixing experimentCommands namespace");
                //automatically fixed when going to FXModelRspec and back.
                return new RspecValidationResult((FXModelRspec) requestRspecSource.getModelRspec(ModelRspecType.FX));
            } else if (chosenAction == ignore) {
                return new RspecValidationResult(requestRspecSource, false /* == ignore */);

            } else if (chosenAction == abort) {
                return new RspecValidationResult(requestRspecSource, true/* == abort*/);
            } else {
                throw new RuntimeException("Unknown chosen action: " + chosenAction);
            }
        }

        return null;
    }
}
