package be.iminds.ilabt.jfed.experimenter_gui.preferences.subsections;

import be.iminds.ilabt.jfed.experimenter_gui.util.ui.TextFieldWithStatus;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PageantHelper;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuTTYPrivateKeyFile;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.layout.BorderPane;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class WindowsSshAuthenticationPane extends SshAuthenticationPane {
    //private static final FileChooser.ExtensionFilter PPK_FILE_EXTENSION =
    //        new FileChooser.ExtensionFilter("PuTTY Private Key (*.ppk)", "*.ppk");

    @FXML
    protected Label textKeyLabel;
    @FXML
    protected TextArea textKeyTextArea;
    @FXML
    private BorderPane root;

    @Override
    public Node getRoot() {
        return root;
    }

    private final GeniUserProvider geniUserProvider;
    private final PageantHelper pageantHelper;

    @Inject
    WindowsSshAuthenticationPane(UserLoginModelManager userLoginModelManager, JFedGuiPreferences jFedPreferences, PageantHelper pageantHelper) {
        super(userLoginModelManager, jFedPreferences);
        this.geniUserProvider = userLoginModelManager;
        this.pageantHelper = pageantHelper;
    }

    @FXML
    protected void initialize() {
        super.initialize();
        textKeyLabel.disableProperty().bind(useCustomKeyPairRadioButton.selectedProperty().not());
        textKeyTextArea.disableProperty().bind(useCustomKeyPairRadioButton.selectedProperty().not());
    }

    @Override
    protected void testPrivateKey() {
        super.testPrivateKey();
        if (fileTextField.getStatus() == TextFieldWithStatus.Status.OK) {
            textKeyTextArea.setText(getPublicKeyFromFile(new File(fileTextField.getText())));
        } else {
            textKeyTextArea.setText("");
        }
    }

    @Override
    protected boolean testPrivateKeyFile(File file) throws IOException {
        return file.exists() && file.isFile() && PuTTYPrivateKeyFile.isPuttyPrivateKey(file);
    }

    protected String getPublicKeyFromFile(File file) {
        String key = null;

        //try to extract public key from this file
        try {
            if (PuTTYPrivateKeyFile.isPuttyPrivateKey(file)) {
                PuTTYPrivateKeyFile ppk = PuTTYPrivateKeyFile.read(file);
                key = ppk.getPublicKeyOpenSshString();
            } else if (KeyUtil.isOpenSshRsaKey(file)) {
                key = IOUtils.fileToString(file).replace("\n", "");
            }
        } catch (IOException ex) {
            JFDialogs.create()
                    .owner(fileTextField)
                    .message("An error occured while processing the key file")
                    .masthead("Error")
                    .title("Error")
                    .showException(ex);

        }
        return key;
    }

    @Override
    public boolean check() {
        return super.check();
    }

    public boolean save() {
        //only continue to save if parent-save is also OK
        if (!super.save())
            return false;

        //Ok, finally, register keys with pageant
        if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT)) {
            List<SshKeyInfo> keys = new ArrayList<>();

            SshKeyInfo sshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();
            if (sshKeyInfo == null && geniUserProvider.isUserLoggedIn()) {
                GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
                sshKeyInfo = new UserSshKeyInfo(geniUser);
            }
            //PageantHelper.registerKey(sshKeyInfo);
            keys.add(sshKeyInfo);

            if (!jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH).equalsIgnoreCase("NEVER")) {
                SshKeyInfo proxySshKeyInfo = jFedPreferences.getOverriddenProxySshKeyInfo();
                if (proxySshKeyInfo == null && geniUserProvider.isUserLoggedIn()) {
                    GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
                    proxySshKeyInfo = new UserSshKeyInfo(geniUser);
                }
                //PageantHelper.registerKey(proxySshKeyInfo);
                keys.add(proxySshKeyInfo);
            }

            pageantHelper.registerKeys(keys);
        }

        return true;
    }
}
