package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.tabs.RibbonEnabled;
import be.iminds.ilabt.jfed.experimenter_gui.tabs.StatusEnabled;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonTab;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.application.Platform;
import javafx.beans.value.ObservableStringValue;
import javafx.collections.ObservableMap;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Tab;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.glyphfont.GlyphFont;
import org.controlsfx.glyphfont.GlyphFontRegistry;

import java.util.List;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 11/19/13
 * Time: 5:10 PM
 */
public class ExperimentViewControllerTab extends Tab implements StatusEnabled, RibbonEnabled {

    private static final GlyphFont fontAwesome = GlyphFontRegistry.font("FontAwesome");
    private final ObservableMap<GeniUrn, ExperimentViewController> sliceControllers;

    private final ExperimentViewController experimentViewController;
    private final SliceRibbonTabs sliceRibbonTabs;


    //instance initializer
    {
        setGraphic(fontAwesome.create(FontAwesome.Glyph.PLAY_CIRCLE));

    }

    private static boolean isActiveSliceState(SliceState sliceState) {
        switch (sliceState) {
            case FAILED:
            case EXPIRED:
            case RESERVATION_FAILED:
                return false;
            default:
                return true;
        }
    }


    public ExperimentViewControllerTab(final ExperimentViewController experimentViewController,
                                       ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers,
                                       SliceRibbonTabs sliceRibbonTabs) {
        this.experimentViewController = experimentViewController;
        this.sliceControllers = experimentViewControllers;
        this.sliceRibbonTabs = sliceRibbonTabs;

        setContent(experimentViewController);
        setText(experimentViewController.getExperiment().getName());

        TabStyle.updateTabStyle(this, experimentViewController.getExperiment().getSliceState());
        experimentViewController.getExperiment().sliceStateProperty().addListener(observable -> {
            Platform.runLater(() -> {
                TabStyle.updateTabStyle(this, experimentViewController.getExperiment().getSliceState());
            });
        });

        setOnCloseRequest(event -> {
            if (isActiveSliceState(getExperimentViewController().getExperiment().getSliceState())) {


                Optional<ButtonType> result = JFDialogs.create()
                        .message("There are currently resources in use." +
                                " Are you sure you want to close this tab?")
                        .masthead("Resources in use")
                        .title("Resources in use")
                        .buttonTypes(ButtonType.YES, ButtonType.NO, ButtonType.CANCEL)
                        .showWarning();


                if (result.isPresent() && result.get() != ButtonType.YES) {
                    event.consume();
                } else {
                    experimentViewController.getControllerUi().getController().unregister();
                    experimentViewControllers.remove(experimentViewController.getExperiment().getSlice().getUrn());
                }
            }
        });
    }

    public ExperimentViewController getExperimentViewController() {
        return experimentViewController;
    }

    @Override
    public ObservableStringValue statusProperty() {
        return experimentViewController.statusProperty();
    }

    @Override
    public List<? extends RibbonTab> getRibbonTabs() {
        return sliceRibbonTabs.getRibbonTabs();
    }

    @Override
    public RibbonTab getInitialTabToActivate() {
        if (experimentViewController.isControllerVisible())
            return sliceRibbonTabs.getControlViewerRibbonTab();
        else if (experimentViewController.isRawRspecVisible())
            return sliceRibbonTabs.getRspecViewerRibbonTab();
        else
            return sliceRibbonTabs.getTopologyViewerRibbonTab();
    }

    public static enum TabStyle {

        FUTURE_RESERVATION("tab-experiment-future-reservation"),
        RUNNING("tab-experiment-execution"),
        EXPIRED("tab-experiment-execution-expired"),
        FAILED("tab-experiment-execution-terminated");


        private final String styleClass;

        TabStyle(String styleClass) {
            this.styleClass = styleClass;
        }


        public static void updateTabStyle(Tab tab, SliceState newState) {
            for (TabStyle tabStyle : TabStyle.values()) {
                tab.getStyleClass().remove(tabStyle.styleClass);
            }

            TabStyle newTabStyle;
            switch (newState) {
                case FUTURE_RESERVATION:
                    newTabStyle = TabStyle.FUTURE_RESERVATION;
                    break;
                case FAILED:
                case RESERVATION_FAILED:
                    newTabStyle = TabStyle.FAILED;
                    break;
                case EXPIRED:
                    newTabStyle = TabStyle.EXPIRED;
                    break;
                default:
                    newTabStyle = TabStyle.RUNNING;
                    break;
            }

            tab.getStyleClass().add(newTabStyle.styleClass);
        }
    }
}
