package be.iminds.ilabt.jfed.experimenter_gui.slice.actions;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import javafx.stage.Window;

import javax.inject.Inject;
import javax.inject.Singleton;

/**
 * User: twalcari
 * Date: 2/4/2015
 * Time: 11:04
 */
@Singleton
public class ExperimentActionsFactory {
    private final AuthorityList authorityList;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final BrowserUtil browserUtil;
    private final TaskService taskService;
    private final JFedPreferences jFedPreferences;

    @Inject
    public ExperimentActionsFactory(AuthorityList authorityList, ExperimentTasksFactory experimentTasksFactory,
                                    BrowserUtil browserUtil, TaskService taskService, JFedPreferences jFedPreferences) {
        this.authorityList = authorityList;
        this.experimentTasksFactory = experimentTasksFactory;
        this.browserUtil = browserUtil;
        this.taskService = taskService;
        this.jFedPreferences = jFedPreferences;
    }


    public CreateDiskImageAction createCreateDiskImageAction(Window parentWindow, ExperimentController experimentController, FXRspecNode node) {
        return new CreateDiskImageAction(parentWindow, experimentController, node, authorityList, experimentTasksFactory);

    }

    public ReloadOSAction createReloadOSAction(Window parentWindow, ExperimentController experimentController, Sliver sliver) {
        return new ReloadOSAction(parentWindow, experimentController, sliver, experimentTasksFactory);
    }


    public OpenConsoleAction createOpenConsoleAction(Window parentWindow, ExperimentController experimentController, Sliver sliver) {
        return new OpenConsoleAction(parentWindow, experimentController, sliver, experimentTasksFactory, browserUtil);
    }

    public StopExperimentAction createStopExperimentAction(Window parentWindow, ExperimentController experimentController) {
        return new StopExperimentAction(parentWindow, experimentController, experimentTasksFactory);
    }

    public SaveManifestAction createSaveManifestAction(ExperimentViewController experimentViewController, boolean saveToCsv) {
        return new SaveManifestAction(experimentViewController, saveToCsv, taskService, jFedPreferences);
    }

    public ShareLanAction createShareLanAction(Window parentWindow, ExperimentController experimentController, FXRspecLink rspecLink) {
        return new ShareLanAction(parentWindow, experimentController, experimentTasksFactory, rspecLink);
    }
}
