package be.iminds.ilabt.jfed.experimenter_gui.slice.ribbon_tabs;

import be.iminds.ilabt.jfed.experimenter_gui.dialogs.DialogsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.actions.ExperimentActionsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon.RibbonButton;
import be.iminds.ilabt.jfed.experimenter_gui.util.DateTimeUtils;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.binding.Bindings;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.fxml.FXML;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;

/**
 * User: twalcari
 * Date: 7/1/2014
 * Time: 17:20
 */
public class GenericSliceRibbonComponentGroupController {

    private final SliceRibbonTab sliceRibbonTab;

    @FXML
    private RibbonButton updateExperimentButton;
    @FXML
    private RibbonButton renewSliceButton;
    @FXML
    private RibbonButton stopExperimentButton;

    private final ExperimentActionsFactory experimentActionsFactory;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final DialogsFactory dialogsFactory;
    private final BooleanProperty isDemo =
            new SimpleBooleanProperty(new JFedGuiPreferences().getBoolean(GuiPreferenceKey.PREF_IS_DEMO,false));

    public GenericSliceRibbonComponentGroupController(SliceRibbonTab sliceRibbonTab, ExperimentActionsFactory experimentActionsFactory, ExperimentTasksFactory experimentTasksFactory, DialogsFactory dialogsFactory) {
        this.sliceRibbonTab = sliceRibbonTab;
        this.experimentActionsFactory = experimentActionsFactory;
        this.experimentTasksFactory = experimentTasksFactory;
        this.dialogsFactory = dialogsFactory;

        sliceRibbonTab.activeExperimentViewControllerProperty().addListener((observableValue, oldValue, newValue) -> {
            if (oldValue != null) {
                updateExperimentButton.disableProperty().unbind();
                renewSliceButton.disableProperty().unbind();
                stopExperimentButton.disableProperty().unbind();
            }

            if (newValue != null) {
                updateExperimentButton.disableProperty().bind(newValue.getExperiment().sliceProperty().isNull());
                renewSliceButton.disableProperty().bind(
                        Bindings.or(Bindings.or(
                                        newValue.getExperiment().sliceProperty().isNull(),
                                        newValue.getExperiment().deletingProperty()),
                                isDemo)
                );
                stopExperimentButton.disableProperty().bind(
                        Bindings.or(
                                Bindings.or(
                                        newValue.getExperiment().sliceProperty().isNull(),
                                        newValue.getExperiment().deletingProperty()),
                                isDemo)
                );
            }


        });
    }

    @FXML
    public void onUpdateExperimentAction() {
        sliceRibbonTab.getActiveExperimentViewController().getExperimentController().submitExperimentTask(
                experimentTasksFactory.createUpdateExperimentTask(sliceRibbonTab.getActiveExperimentViewController().getExperiment())
        );
    }

    @FXML
    public void onStopExperimentAction() {
        experimentActionsFactory.
                createStopExperimentAction(stopExperimentButton.getScene().getWindow(),
                        sliceRibbonTab.getActiveExperimentViewController().getExperimentController())
                .run();
    }


    @FXML
    private void onRenewSliceAction() {

        Slice slice = sliceRibbonTab.getActiveExperimentViewController().getExperiment().getSlice();
        Instant sliceExpirationTime = slice.getExpirationDate();

        if (sliceExpirationTime.isBefore(Instant.now())) {
            JFDialogs.create()
                    .owner(renewSliceButton.getScene().getWindow())
                    .message(String.format("The slice %s has expired on %s. It can therefore not be renewed again.",
                            slice.getName(),
                            DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(LocalDateTime.ofInstant(sliceExpirationTime, ZoneId.systemDefault()))))
                    .masthead("Slice has expired")
                    .title("Slice has expired").showError();
        } else {
            dialogsFactory.showRenewSliceDialog(renewSliceButton.getScene().getWindow(), sliceRibbonTab.getActiveExperimentViewController().getExperimentController());
        }
    }


}
