package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Date;

/**
 *
 * User: twalcari
 * Date: 2/3/2015
 * Time: 07:34
 */
public abstract class InitializeSliversTaskGenerator extends ExperimentTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(InitializeSliversTaskGenerator.class);

    protected final HighLevelController hlc;

    InitializeSliversTaskGenerator(ExperimentController experimentController, HighLevelController hlc) {
        super(experimentController);
        this.hlc = hlc;
    }

    protected HighLevelController.RenewSliceTask createRenewSliceTask(){
        assert(experiment.getRequestedEndTime() != null);
        /*
        Note: when comparing times, take into account that the slice expiration date is set from the received date
         in RFC3339, which has no sub second precision.

         The experiment.getRequestedEndTime() is truncated to second precision. So this is not an issue (anymore).

         see also FEDIBBTDEV-738

         */
        if (experiment.getSlice().getExpirationDate() != null &&
                experiment.getSlice().getExpirationDate().isBefore(experiment.getRequestedEndTime())) {
            LOG.info("Slice expires at " + experiment.getSlice().getExpirationDate() +
                    " (= " + RFC3339Util.dateToRFC3339String(Date.from(experiment.getSlice().getExpirationDate()), true, true, true) +
                    " = " + experiment.getSlice().getExpirationDate().toEpochMilli() + ")" +
                    " which is before the requested sliver expiration at " + experiment.getRequestedEndTime() +
                    " (" + RFC3339Util.dateToRFC3339String(Date.from(experiment.getRequestedEndTime()), true, true, true) +
                    " = " + experiment.getRequestedEndTime().toEpochMilli() + "). -> will update slice expiration");
            return hlc.renewSlice(experiment.getSlice(), experiment.getRequestedEndTime(), false/*false = don't renew any slivers*/);
        } else {
            LOG.info("Slice expires at " + experiment.getSlice().getExpirationDate() +
                    " (= " + RFC3339Util.dateToRFC3339String(Date.from(experiment.getSlice().getExpirationDate()), true, true, true) +
                    " = " + experiment.getSlice().getExpirationDate().toEpochMilli() + ")" +
                    " which is NOT before the requested sliver expiration at " + experiment.getRequestedEndTime() +
                    " (" + RFC3339Util.dateToRFC3339String(Date.from(experiment.getRequestedEndTime()), true, true, true) +
                    " = " + experiment.getRequestedEndTime().toEpochMilli() + "). -> will NOT update slice expiration");
            return null;
        }
    }

}
