package be.iminds.ilabt.jfed.experimenter_gui.ui.ribbon;

import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Control;
import javafx.scene.control.ToggleButton;
import javafx.scene.input.KeyCombination;
import org.controlsfx.glyphfont.Glyph;

/**
 * User: twalcari
 * Date: 11/6/13
 * Time: 9:48 AM
 */
public class RibbonToggleButton extends ToggleButton implements Runnable {
    static final double SMALL_ICON_SIZE = 16;
    static final double BIG_ICON_SIZE = 24;
    private static final String BIG_RIBBON_BUTTON_STYLE = "ribbon-button";
    private static final String SMALL_RIBBON_BUTTON_STYLE = "ribbon-button-small";
    private boolean small = false;
    private KeyCombination accelerator = null;

    public RibbonToggleButton() {
        initialize();
    }

    public RibbonToggleButton(String s) {
        super(s);
        initialize();
    }

    public RibbonToggleButton(String s, Node node) {
        super(s, node);
        initialize();
    }

    /**
     * Helper method for accelerators
     */

    public void run() {
        this.fire();
    }

    private void initialize() {
        getStylesheets().add(getClass().getResource(RibbonBar.STYLESHEET_URL).toExternalForm());
        getStyleClass().add(BIG_RIBBON_BUTTON_STYLE);

        graphicProperty().addListener(observable -> updateStyle());

        sceneProperty().addListener((observableValue, oldValue, newValue) -> {

            if (oldValue != null) {
                unregisterAccelerator(oldValue);
            }
            if (newValue != null) {
                registerAccelerator();
            }

        });
        if (getScene() != null)
            registerAccelerator();
    }

    public boolean isSmall() {
        return small;
    }

    public void setSmall(boolean small) {
        this.small = small;
        updateStyle();
    }

    public KeyCombination getAccelerator() {
        return accelerator;
    }

    public void setAccelerator(KeyCombination accelerator) {
        if (this.getScene() != null && this.accelerator != null)
            unregisterAccelerator(getScene());

        this.accelerator = accelerator;

        if (this.getScene() != null)
            registerAccelerator();
    }

    protected void registerAccelerator() {
        assert (getScene() != null);

        if (accelerator != null) {
            if (getScene().getAccelerators().get(accelerator) != null)
                throw new RuntimeException("Accelerator " + accelerator + " is already defined in scene.");
            getScene().getAccelerators().put(accelerator, RibbonToggleButton.this);
        }
    }

    protected void unregisterAccelerator(Scene oldScene) {
        if (accelerator == null)
            return;
        assert (oldScene.getAccelerators().containsKey(accelerator));
        oldScene.getAccelerators().remove(accelerator);

    }

    protected void updateStyle() {
        if (small) {
            getStyleClass().remove(BIG_RIBBON_BUTTON_STYLE);
            getStyleClass().add(SMALL_RIBBON_BUTTON_STYLE);
        } else {
            getStyleClass().remove(SMALL_RIBBON_BUTTON_STYLE);
            getStyleClass().add(BIG_RIBBON_BUTTON_STYLE);
        }


        if (getGraphic() == null)
            return;

        if (getGraphic() instanceof Glyph) {
            Glyph glyph = (Glyph) getGraphic();
            glyph.setFontSize(small ? SMALL_ICON_SIZE : BIG_ICON_SIZE);
        } else if (getGraphic() instanceof Control) {
            if (small) {
                ((Control) getGraphic()).setPrefHeight(SMALL_ICON_SIZE);
                ((Control) getGraphic()).setPrefWidth(SMALL_ICON_SIZE);
            } else {
                ((Control) getGraphic()).setPrefHeight(BIG_ICON_SIZE);
                ((Control) getGraphic()).setPrefWidth(BIG_ICON_SIZE);
            }
        }
    }

}
