package be.iminds.ilabt.jfed.highlevel;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.SimpleGeniUser;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;

import javax.inject.Singleton;
import java.security.cert.X509Certificate;
import java.util.List;

/**
 * User: twalcari
 * Date: 11/27/2014
 * Time: 16:32
 */

public class MockGeniUserModule  extends AbstractModule{

    public static final String LOGGED_IN_USER_URN_STRING = "urn:publicid:IDN+wall2.ilabt.iminds.be+user+wvdemeer";

    public static final GeniUrn LOGGED_IN_USER_URN = GeniUrn.parse(LOGGED_IN_USER_URN_STRING);


    @Provides
    @Singleton
    public GeniUser provideGeniUser(AuthorityListModel authorityListModel) {
        SfaAuthority userAuthority = authorityListModel.getByUrnExact(
                "urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm");

        SimpleGeniUser loggedInUser = new SimpleGeniUser(userAuthority,
                LOGGED_IN_USER_URN,
                (List<X509Certificate>) null/*certificateChain*/,
                null/*privateKey*/,
                null/*certificateKeyFile*/,
                null/*privateKeyFile*/);

        return loggedInUser;
    }


    @Provides
    @Singleton
    public GeniUserProvider provideGeniUserProvider(GeniUser user) {
        return new GeniUserProvider() {
            @Override
            public GeniUser getLoggedInGeniUser() {
                return user;
            }

            @Override
            public boolean isUserLoggedIn() {
                return true;
            }
        };
    }

    @Override
    protected void configure() {

    }
}
