package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;

import javax.inject.Inject;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

/**
 * AutomaticAggregateManagerWrapper
 */
public class AutomaticAggregateManagerWrapper extends AggregateManagerWrapper {
    private static final String EMULAB_TYPE = "emulab";
    public static boolean preferAMv3forNonEmulab = true;  //TODO put this in a preferences system somehow
    public static boolean preferAMv3forEmulab = true;     //TODO put this in a preferences system somehow
    protected ServerType serverType;
    protected ServerType poa_serverType;
    protected AggregateManagerWrapper impl;
    protected AggregateManagerWrapper impl_amv2;
    protected AggregateManagerWrapper impl_amv3;
    protected AggregateManagerWrapper poa_impl; //seperate implementation which supports PerformOperationalAction
    protected ProtoGeniAMExtensionsWrapper protoGeniAMExtensionsWrapper;
    //seperate implementation for ProtoGeni Extensions


    protected AggregateManagerWrapper last_used_impl;

    private AutomaticAggregateManagerWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider,
                                             JFedPreferences jFedPreferences, SfaAuthority amAuthority) throws NoAmException {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);

        assert jFedPreferences != null;

        choose();
    }

    protected void choose() {
        //handle no authority given (occurs only when probe needs to check our methods).
        if (amAuthority == null)
            return;

        //first choose poa_impl
        choosePoaImpl();

        boolean preferAMv3;
        if (EMULAB_TYPE.equalsIgnoreCase(amAuthority.getType()))
            preferAMv3 = preferAMv3forEmulab;
        else
            preferAMv3 = preferAMv3forNonEmulab;

        boolean forceV2 = amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_V2);
        boolean forceV3 = amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_V3);

        if (amAuthority.getUrls().containsKey(new ServerType(ServerType.GeniServerRole.AM, 2))) {
            impl_amv2 = new AMv2Wrapper(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);
        }
        if (amAuthority.getUrls().containsKey(new ServerType(ServerType.GeniServerRole.AM, 3))) {
            impl_amv3 = new AMv3Wrapper(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);
        }

        if (impl_amv2 != null && (forceV2 || !preferAMv3 || impl_amv3 == null)) {
            serverType = new ServerType(ServerType.GeniServerRole.AM, 2);
            impl = impl_amv2;
            setupProtoGeniAMExtensions();
            return;
        }
        if (impl_amv3 != null && (forceV3 || preferAMv3 || impl_amv2 == null)) {
            serverType = new ServerType(ServerType.GeniServerRole.AM, 3);
            impl = impl_amv3;
            setupProtoGeniAMExtensions();
            return;
        }
        assert (impl == null && impl_amv2 == null && impl_amv3 == null) || impl != null || (impl_amv2 != null && impl_amv3 != null) : "impl="+impl+" impl_amv2="+impl_amv2+" impl_amv3="+impl_amv3;

        //OCCI doesn't provide a real AM, but we can wrap it (with limited functionality)...
        if (amAuthority.getUrls().containsKey(new ServerType(ServerType.GeniServerRole.OCCI, 1))) {
            serverType = new ServerType(ServerType.GeniServerRole.OCCI, 1);
            impl = new OcciToAggregateManagerWrapper(logger, geniUserProvider, connectionProvider, amAuthority, jFedPreferences);
            setupProtoGeniAMExtensions();
            return;
        }

        throw new NoAmException("There is no (supported) Aggregate Manager known for \"" + amAuthority.getName() + "\". amAuthority urn: " + amAuthority.getUrn() +
                "  types of known servers: " + amAuthority.getUrls().keySet());
    }

    protected void choosePoaImpl() {
        //handle no authority given (occurs only when probe needs to check our methods).
        if (amAuthority == null)
            return;

        if (amAuthority.getUrls().containsKey(new ServerType(ServerType.GeniServerRole.AM, 3))) {
            poa_serverType = new ServerType(ServerType.GeniServerRole.AM, 3);
            poa_impl = new AMv3Wrapper(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);
            return;
        }
    }

    protected void setupProtoGeniAMExtensions() {
        //handle no authority given (occurs only when probe needs to check our methods).
        if (amAuthority == null)
            return;

        if (amAuthority.getType() == null)
            return;

        if ((
                amAuthority.getType().equalsIgnoreCase("emulab") ||
                        amAuthority.getType().equalsIgnoreCase("protogeni") ||
                        amAuthority.getType().equalsIgnoreCase("instageni")
        )
                &&
                (
                        amAuthority.getUrls().containsKey(new ServerType(ServerType.GeniServerRole.AM, 2)) //only works correctly for AMv2 !!!! see FEDIBBTDEV-726
                )
                ) {
            protoGeniAMExtensionsWrapper =
                    new ProtoGeniAMExtensionsWrapper(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority, new ServerType(ServerType.GeniServerRole.AM, 2));
            return;
        }
    }

    protected AggregateManagerWrapper getImpl() {
        if (impl == null)
            choose();
        return impl;
    }

    @ApiMethod(order = 1, hint = "Call \"GetVersion\"")
    @Override
    public Hashtable getVersion() throws JFedException {
        last_used_impl = impl;
        return impl.getVersion();
    }

    @ApiMethod(order = 2, hint = "List all Resources at the Aggregate (optionally list only available)")
    @Override
    public String listResources(
            @ApiMethodParameter(name = "userCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> userCredentials,
            @ApiMethodParameter(name = "available", hint = "", parameterType = ApiMethodParameterType.BOOLEAN)
            boolean available) throws JFedException {
        last_used_impl = impl;
        return impl.listResources(userCredentials, available);
    }

    @ApiMethod(order = 2, hint = "List all Resources at the Aggregate (optionally list only available)")
    @Override
    public String listResources(
            @ApiMethodParameter(name = "userCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> userCredentials,
            @ApiMethodParameter(name = "available", hint = "", parameterType = ApiMethodParameterType.BOOLEAN)
            boolean available,
            @ApiMethodParameter(name = "type", hint = "type of advertisement RSpec", parameterType = ApiMethodParameterType.STRING)
            String type,
            @ApiMethodParameter(name = "version", hint = "version of advertisement RSpec", parameterType = ApiMethodParameterType.STRING)
            String version) throws JFedException {
        last_used_impl = impl;
        return impl.listResources(userCredentials, available, type, version);
    }


    @ApiMethod(order = 3, hint = "Create a Sliver reservation")
    @Override
    public String allocate(
            @ApiMethodParameter(name = "sliceUrn", hint = "", parameterType = ApiMethodParameterType.SLICE_URN)
            GeniUrn sliceUrn,
            @ApiMethodParameter(name = "sliceCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> sliceCredentials,
            @ApiMethodParameter(name = "rspec", hint = "", parameterType = ApiMethodParameterType.RSPEC_STRING)
            String rspec,
            @ApiMethodParameter(name = "startDate", hint = "", parameterType = ApiMethodParameterType.DATE, required = true)
            Date startDate,
            @ApiMethodParameter(name = "endDate", hint = "", parameterType = ApiMethodParameterType.DATE, required = true)
            Date endDate) throws JFedException {
        last_used_impl = impl;
        return impl.allocate(sliceUrn, sliceCredentials, rspec, startDate, endDate);
    }
    @ApiMethod(order = 3, hint = "Create a Sliver (includes allocating, provisioning and starting)")
    @Override
    public String createSliver(
            @ApiMethodParameter(name = "sliceUrn", hint = "", parameterType = ApiMethodParameterType.SLICE_URN)
            GeniUrn sliceUrn,
            @ApiMethodParameter(name = "sliceCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> sliceCredentials,
            @ApiMethodParameter(name = "rspec", hint = "", parameterType = ApiMethodParameterType.RSPEC_STRING)
            String rspec,
            @ApiMethodParameter(name = "userUrns", hint = "", parameterType = ApiMethodParameterType.LIST_OF_USERSPEC)
            List<UserSpec> userUrns,
            @ApiMethodParameter(name = "expirationDate", hint = "", parameterType = ApiMethodParameterType.DATE, required = false, guiDefaultOptional = false)
            Date expirationDate) throws JFedException, AnyVlanUnsupportedException {
        last_used_impl = impl;
        return impl.createSliver(sliceUrn, sliceCredentials, rspec, userUrns, expirationDate);
    }

    @Override
    public boolean hasSeperateAllocateAndProvision() {
        return impl.hasSeperateAllocateAndProvision();
    }

    @Override
    public String allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date expirationDate) throws JFedException, VlanUnavailableException, AnyVlanUnsupportedException {
        last_used_impl = impl;
        return impl.allocate(sliceUrn, sliceCredentials, rspec, expirationDate);
    }

    @Override
    public String provisionAndStart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> userSpecs, Date expirationDate) throws JFedException, VlanUnavailableException, AnyVlanUnsupportedException {
        last_used_impl = impl;
        return impl.provisionAndStart(sliceUrn, sliceCredentials, userSpecs, expirationDate);
    }

    @ApiMethod(order = 4, hint = "Delete a Sliver")
    @Override
    public void deleteSliver(
            @ApiMethodParameter(name = "sliceUrn", hint = "", parameterType = ApiMethodParameterType.SLICE_URN)
            GeniUrn sliceUrn,
            @ApiMethodParameter(name = "sliceCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> sliceCredentials) throws JFedException {
        if (impl_amv2 != null && amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_DELETE_V2)) {
            last_used_impl = impl_amv2;
            impl_amv2.deleteSliver(sliceUrn, sliceCredentials);
        }
        else {
            last_used_impl = impl;
            impl.deleteSliver(sliceUrn, sliceCredentials);
        }
    }

    @ApiMethod(order = 5, hint = "Request the status of a sliver")
    @Override
    public StatusDetails status(
            @ApiMethodParameter(name = "sliceUrn", hint = "", parameterType = ApiMethodParameterType.SLICE_URN)
            GeniUrn sliceUrn,
            @ApiMethodParameter(name = "sliceCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> sliceCredentials) throws JFedException {
        if (impl_amv2 != null && amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_STATUS_V2)) {
            last_used_impl = impl_amv2;
            return impl_amv2.status(sliceUrn, sliceCredentials);
        }
        last_used_impl = impl;
        return impl.status(sliceUrn, sliceCredentials);
    }

    @ApiMethod(order = 6, hint = "Request the manifest RSpec of a sliver")
    @Override
    public String describe(
            @ApiMethodParameter(name = "sliceUrn", hint = "", parameterType = ApiMethodParameterType.SLICE_URN)
            GeniUrn sliceUrn,
            @ApiMethodParameter(name = "sliceCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> sliceCredentials) throws JFedException {
        if (impl_amv2 != null && amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_DESCRIBE_V2)) {
            last_used_impl = impl_amv2;
            return impl_amv2.describe(sliceUrn, sliceCredentials);
        }
        last_used_impl = impl;
        return impl.describe(sliceUrn, sliceCredentials);
    }

    @ApiMethod(order = 7, hint = "Renew the expiration time of a sliver.")
    @Override
    public void renewSlivers(
            @ApiMethodParameter(name = "sliceUrn", hint = "", parameterType = ApiMethodParameterType.SLICE_URN, required = false)
            GeniUrn sliceUrn,
            @ApiMethodParameter(name = "sliverUrns", hint = "", parameterType = ApiMethodParameterType.LIST_OF_SLIVER_URN, required = false)
            List<GeniUrn> sliverUrns,
            @ApiMethodParameter(name = "sliceCredentials", hint = "", parameterType = ApiMethodParameterType.LIST_OF_CREDENTIAL)
            List<AnyCredential> sliceCredentials,
            @ApiMethodParameter(name = "newExpire", hint = "", parameterType = ApiMethodParameterType.DATE)
            Date newExpire) throws JFedException {
        last_used_impl = impl;
        impl.renewSlivers(sliceUrn, sliverUrns, sliceCredentials, newExpire);
    }

    //Also use last AM reply from impl
    @Override
    public AbstractGeniAggregateManager.AggregateManagerReply getLastReply() {
        if (last_used_impl == null) return null;
        return last_used_impl.getLastReply();
    }

    @Override
    public boolean isExirationdateSetByCreateSliver() {
        last_used_impl = impl;
        return impl.isExirationdateSetByCreateSliver();
    }


    @ApiMethod(order = 10, hint = "Ask if the Share LAN option is supported")
    @Override
    public boolean isShareLanSupported() {
        if (!amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_POA_SHARE_LAN))
            return false;
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            return poa_impl.isShareLanSupported();
        }
        return false;
    }

    @ApiMethod(order = 11, hint = "Ask if Restart is supported")
    @Override
    public boolean isRestartSupported() {
        if (!amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_POA_NODE_RESTART))
            return false;
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            return poa_impl.isRestartSupported(); //always true
        }
        last_used_impl = impl;
        return impl.isRestartSupported(); //always false
    }

    @Override
    public boolean isReloadOSSupported() {
        if (!amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_POA_NODE_RELOAD_OS))
            return false;
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            return poa_impl.isRestartSupported(); //always true
        }
        last_used_impl = impl;
        return impl.isRestartSupported(); //always false
    }

    @Override
    public boolean isOpenConsoleSupported() {
        if (!amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_POA_OPEN_CONSOLE))
            return false;
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            return poa_impl.isOpenConsoleSupported(); //always true
        }
        last_used_impl = impl;
        return impl.isOpenConsoleSupported(); //always false
    }

    @ApiMethod(order = 12, hint = "Share a LAN between slices")
    @Override
    public void shareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String newSharedLanName) throws JFedException {
        if (poa_impl != null && poa_impl.isShareLanSupported()) {
            last_used_impl = poa_impl;
            poa_impl.shareLan(sliceUrn, sliceCredentials, lanClientId, newSharedLanName);
        } else
            throw new RuntimeException("Aggregate Manager has no support for shared LAN");
    }

    @ApiMethod(order = 13, hint = "Unshare a shared LAN")
    @Override
    public void unshareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String sharedLanName) throws JFedException {
        if (poa_impl != null && poa_impl.isShareLanSupported()) {
            last_used_impl = poa_impl;
            poa_impl.unshareLan(sliceUrn, sliceCredentials, lanClientId, sharedLanName);
        } else
            throw new RuntimeException("Aggregate Manager has no support for shared LAN");
    }

    @ApiMethod(order = 14, hint = "Restart all resources in a Slice")
    @Override
    public void restart(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            poa_impl.restart(sliverUrn, sliceCredentials);
        } else
            throw new RuntimeException("Aggregate Manager has no support for PerformOperationAction, and thus not for restart.");
    }

    @Override
    public void reloadOS(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            poa_impl.reloadOS(sliverUrn, sliceCredentials);
        }
        else
            throw new RuntimeException("Aggregate Manager has no support for PerformOperationAction, and thus not for reload.");
    }

    @Override
    public URL openConsole(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            return poa_impl.openConsole(sliverUrn, sliceCredentials);
        }
        else
            throw new RuntimeException("Aggregate Manager has no support for PerformOperationAction, and thus not for open console.");
    }

    @ApiMethod(order = 15, hint = "Ask if SSH keys can be supported")
    @Override
    public boolean isEditSshKeysSupported() {
        if (!amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_UPDATE_SSH_KEYS))
            return false;
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            return poa_impl.isEditSshKeysSupported(); //always true
        }
        last_used_impl = impl;
        return impl.isEditSshKeysSupported(); //always false
    }

    @ApiMethod(order = 16, hint = "Add users and SSH keys to nodes within a slice")
    @Override
    public void editSshKeys(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> userSpecs) throws JFedException {
        if (poa_impl != null) {
            last_used_impl = poa_impl;
            poa_impl.editSshKeys(sliceUrn, sliceCredentials, userSpecs);
        } else
            throw new RuntimeException("Aggregate Manager has no support for PerformOperationAction, and thus not for editSshKeys");
    }

    @ApiMethod(order = 17, hint = "Ask if Create Image is supported")
    @Override
    public boolean areUserDiskImageFunctionsSupported() {
        if (!amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_EXT_USER_DISK_IMAGE_OPS))
            return false;
        return protoGeniAMExtensionsWrapper != null;
    }

    @ApiMethod(order = 18, hint = "Create a disk image from a node")
    @Override
    public ImageInfo createImage(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn componentUrn, String imageName) throws JFedException {
        if (protoGeniAMExtensionsWrapper == null)
            throw new RuntimeException("operation is not supported");

        return protoGeniAMExtensionsWrapper.createImage(sliceUrn, sliceCredentials, componentUrn, imageName);
    }

    @Override
    public List<ImageInfo> listImages(List<AnyCredential> userCredentials) throws JFedException {
        if (protoGeniAMExtensionsWrapper == null)
            throw new RuntimeException("operation is not supported");

        List<? extends ImageInfo> images = protoGeniAMExtensionsWrapper.listImages(userCredentials);
        return new ArrayList<ImageInfo>(images);
    }

    @Override
    public boolean deleteImage(List<AnyCredential> userCredentials, GeniUrn imageUrn) throws JFedException {
        if (protoGeniAMExtensionsWrapper == null)
            throw new RuntimeException("operation is not supported");

        boolean success = protoGeniAMExtensionsWrapper.deleteImage(userCredentials, imageUrn);

        if (!success)
            throw new JFedException("Did not successfully delete image");
        return success;
    }







    public static interface AutomaticAggregateManagerWrapperFactory {
        public AutomaticAggregateManagerWrapper create(SfaAuthority amAuthority) throws NoAmException;
        public AutomaticAggregateManagerWrapper create(Logger altLogger, SfaAuthority amAuthority) throws NoAmException;
    }

    public static class BasicAutomaticAggregateManagerWrapperFactory implements AutomaticAggregateManagerWrapperFactory {
        private final Logger logger;
        private final GeniUserProvider geniUserProvider;
        private final JFedConnectionProvider connectionProvider;
        private final JFedPreferences jFedPreferences;

        @Inject
        public BasicAutomaticAggregateManagerWrapperFactory(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
            this.logger = logger;
            this.geniUserProvider = geniUserProvider;
            this.connectionProvider = connectionProvider;
            this.jFedPreferences = jFedPreferences;
        }
        public BasicAutomaticAggregateManagerWrapperFactory(Logger logger, final GeniUser geniUser, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
            this(logger, new GeniUserProvider() {
                public GeniUser getLoggedInGeniUser() {
                    return geniUser;
                }

                public boolean isUserLoggedIn() {
                    return geniUser != null;
                }
            }, connectionProvider, jFedPreferences);
        }

        public AutomaticAggregateManagerWrapper create(SfaAuthority amAuthority) throws NoAmException {
            return this.create(logger, amAuthority);
        }
        public AutomaticAggregateManagerWrapper create(Logger altLogger, SfaAuthority amAuthority) throws NoAmException {
            return new AutomaticAggregateManagerWrapper(altLogger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);
        }
    }
}
