package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi2;
import be.iminds.ilabt.jfed.lowlevel.api.FederationMemberAuthorityApi2;
import be.iminds.ilabt.jfed.lowlevel.api.FederationSliceAuthorityApi2;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.*;
import org.slf4j.LoggerFactory;

import java.security.cert.X509Certificate;
import java.util.*;

/**
 * ProtoGeniSAUserAndSliceApiWrapper
 */
public class UniformFederationApi2UserAndSliceApiWrapper extends UserAndSliceApiWrapper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(UniformFederationApi2UserAndSliceApiWrapper.class);
    List<AnyCredential> userCredentials;

    public UniformFederationApi2UserAndSliceApiWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }

    private FederationMemberAuthorityApi2 ma(Logger logger) {
        return new FederationMemberAuthorityApi2(logger, jFedPreferences);
    }

    private FederationSliceAuthorityApi2 sa(Logger logger) {
        return new FederationSliceAuthorityApi2(logger, jFedPreferences);
    }

    private SfaConnection getSaConnection() throws JFedException {
        return getConnection(new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 2));
    }

    private SfaConnection getMaConnection() throws JFedException {
        return getConnection(new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 2));
    }

    public FederationSliceAuthorityApi2.GetVersionSAResult getGetVersionSAResult(Logger logger) throws JFedException {
        //TODO: buffer

        AbstractFederationApi2.FederationApiReply<FederationSliceAuthorityApi2.GetVersionSAResult> reply =
                sa(logger).getVersion(getSaConnection());

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("GetVersion call to SA not successful: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        return reply.getValue();
    }

    public List<AnyCredential> makeCredentialList(AnyCredential c) {
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        if (c != null)
            res.add(c);
        return res;
    }

    @Override
    public synchronized List<AnyCredential> getUserCredentials(Logger logger, GeniUrn user) throws JFedException {
        userCredentials = null;
        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> reply =
                ma(logger).getCredentials(
                        getMaConnection(),
                        new ArrayList<AnyCredential>(),
                        user,
                        null);

        assert reply != null;

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve user credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        assert reply.getValue() != null;
        assert !reply.getValue().isEmpty();

        userCredentials = reply.getValue();
        return new ArrayList<AnyCredential>(userCredentials);
    }

    @Override
    public boolean hasUserCredentials() {
        return hasUserCredentials(userCredentials);
    }

    public List<AnyCredential> getCachedUserCredentialsForAM() {
        if (userCredentials == null)
            return Collections.emptyList();
        else
            return new ArrayList<AnyCredential>(userCredentials);
    }

    @Override
    public List<AnyCredential> getSliceCredentials(Logger logger, GeniUrn sliceUrn) throws JFedException {
        assert userCredentials != null && !userCredentials.isEmpty();
        assert sliceUrn != null;

        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> reply =
                sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, null);

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve slice credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        return reply.getValue();
    }

    @Override
    public List<GeniUrn> getSlicesForUser(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();
        List<String> resStrings = new ArrayList<String>();

        //match is officically not supported but the API, but it works on the geni portal.
        Map<String, Object> extraOptions = new HashMap();
        Hashtable<String, Boolean> match = new Hashtable<String, Boolean>();
        match.put("SLICE_EXPIRED", false);
        extraOptions.put("match", match);

        AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupForMember(getSaConnection(), "SLICE", user, userCredentials, extraOptions);
        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        List<FederationSliceAuthorityApi2.UrnRoleTuple> urnRoleTuples = reply.getValue();
        for (FederationSliceAuthorityApi2.UrnRoleTuple urnRoleTuple : urnRoleTuples) {
            if (!urnRoleTuple.isExpired()) {
                res.add(urnRoleTuple.getUrn());
                resStrings.add(urnRoleTuple.getUrn().toString());
            }
        }

        //since it takes only 1 extra call, we'll slo get all other info about the slices (such as expire date)
        Map<String, Object> match2 = new HashMap<String, Object>();
        match2.put("SLICE_URN", new Vector(resStrings));
        sa(logger).lookup(getSaConnection(), "SLICE", userCredentials, match2, null/*filter*/, null/*extraOptions*/);

        return res;
        //get other info:
//        Map<String, String> match = new HashMap<String, String>();
//        match.put("MEMBER_URN", user.toString());
//        ma.lookupPublicMemberInfo(con, match, null, null);
//        ma.lookupPrivateMemberInfo(con, userCredentials, match, null, null);
    }

    public List<GeniUrn> getUserProjects(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();

        AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupForMember(getSaConnection(),
                        "PROJECT",
                        user,
                        userCredentials,
                        null);
        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        List<FederationSliceAuthorityApi2.UrnRoleTuple> urnRoleTuples = reply.getValue();
        LOG.debug("UniformFederationApi1UserAndSliceApiWrapper#getUserProjects urnRoleTuples.size()=" + urnRoleTuples.size());
        boolean hasExpired = false;
        for (FederationSliceAuthorityApi2.UrnRoleTuple urnRoleTuple : urnRoleTuples) {

            //first a special exception: server might include EXPIRED field in tuple. We can use that info
            if (urnRoleTuple.dictRaw != null && urnRoleTuple.dictRaw.containsKey("EXPIRED")) {
                hasExpired = true;
                if (TextUtil.objectToBoolean(urnRoleTuple.dictRaw.get("EXPIRED"))) {
                    LOG.debug("skipping expired project: " + urnRoleTuple.getUrn());
                    continue;
                }
            }

            res.add(urnRoleTuple.getUrn());
        }

        if (!hasExpired) {
            LOG.debug("doing second call to check for expired projects");

            Map<String, Object> match = new HashMap<String, Object>();
            match.put("EXPIRED", Boolean.FALSE);
            List<String> filter = new ArrayList<String>();
            filter.add("PROJECT_URN");
            AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.LookupResult> nonExpiredProj = null;
            try {
                nonExpiredProj =
                        sa(logger).lookup(getSaConnection(),
                                "PROJECT",
                                userCredentials,
                                match,
                                filter,
                                null);
            } catch (Throwable e) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).", e);
                nonExpiredProj = null;
                return res;
            }
            if (nonExpiredProj == null || nonExpiredProj.getGeniResponseCode() == null || !nonExpiredProj.getGeniResponseCode().isSuccess() || nonExpiredProj.getValue() == null) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).");
            } else {
                LOG.debug("Found " + nonExpiredProj.getValue().size() + " non expired projects on server.");
                List<GeniUrn> expiredUserProjects = new ArrayList<GeniUrn>();
                for (GeniUrn resUrn : res) {
                    if (nonExpiredProj.getValue().get(resUrn) == null)
                        expiredUserProjects.add(resUrn);
                }
                LOG.debug("found " + expiredUserProjects.size() + " expired projects in " + res.size() + " user projects: will remove them");
                res.removeAll(expiredUserProjects);
            }
        } else {
            LOG.debug("no second call needed to check for expired projects");
        }

        LOG.debug("UniformFederationApi1UserAndSliceApiWrapper#getUserProjects res.size()=" + res.size());
        return res;
    }

    public AbstractFederationApi2.LookupResult getUserIdentifyingInfo(Logger logger, GeniUrn user) throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        match.put("MEMBER_URN", getUser().getUserUrnString());

//        Vector<String> filter = new Vector<String>();

        AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.LookupResult> reply =
                ma(logger).lookup(getMaConnection(),
                        "MEMBER",
                        userCredentials,
                        match,
                        null/*filter*/,
                        null);


        if (reply == null)
            throw new JFedException("Problem looking up user info for " + getUser().getUserUrnString());

        if (reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            throw new JFedException("Problem looking up user info for " + getUser().getUserUrnString(), reply);

        return reply.getValue();
    }

    @Override
    public List<String> getAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException {
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                sa(logger).lookupSliverInfo(getSaConnection(),
                        sliceCredentials, sliceUrn, null/*match*/, null/*filter*/, null/*extraoptions*/);

        if (!reply.getGeniResponseCode().isSuccess()) return null;
        AbstractFederationApi.LookupResult lookupRes = reply.getValue();
        if (lookupRes == null) return null;

        List<String> res = new ArrayList<String>();
        List<String> keys = lookupRes.getKeys();
        for (String key : keys) {
            String aggManUrn = (String) lookupRes.get(key).get("SLIVER_INFO_AGGREGATE_URN");
            if (aggManUrn != null)
                res.add(aggManUrn);
        }

        return res;
    }
    @Override
    public boolean isRegisterAggregatesForSliceSupported() {
        return true;
    }
    @Override
    public void registerAggregatesForSlice(
            Logger logger,
            List<AnyCredential> sliceCredentials,
            GeniUrn sliceUrn,
            GeniUrn aggregateUrn,
            Collection<GeniUrn> sliverUrns,
            Date sliverExpirationDate) throws JFedException {
        LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will register: "+ sliverUrns.size());
        GeniUrn creatorUrn = getUser().getUserUrn();
        Date expiration = sliverExpirationDate;

        //fall back to slice expiration
        if (expiration == null)
            for (AnyCredential c : sliceCredentials)
                if (c.getExpiresDate() != null && (expiration == null || expiration.after(c.getExpiresDate())))
                    expiration = c.getExpiresDate();

        for (GeniUrn sliverUrn : sliverUrns) {
            LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is registering: "+ sliverUrn);
            AbstractFederationApi.FederationApiReply<Hashtable<String, Object>> reply =
                    sa(logger).createSliverInfo(getSaConnection(),
                            sliceCredentials,
                            sliceUrn.toString(),
                            sliverUrn.toString(),
                            aggregateUrn.toString(),
                            creatorUrn.toString(),
                            expiration,
                            null/*fields*/,
                            null/*extraoptions*/);
        }
    }
    @Override
    public void updateSliverExpirationDateInfoForSlice(
            Logger logger,
            List<AnyCredential> sliceCredentials,
            GeniUrn sliceUrn,
            Collection<GeniUrn> sliverUrns,
            Date newSliverExpirationDate) throws JFedException {
        LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will register: "+ sliverUrns.size());
        GeniUrn creatorUrn = getUser().getUserUrn();
        Date expiration = newSliverExpirationDate;

        //fall back to slice expiration
        if (expiration == null)
            for (AnyCredential c : sliceCredentials)
                if (c.getExpiresDate() != null && (expiration == null || expiration.after(c.getExpiresDate())))
                    expiration = c.getExpiresDate();

        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLIVER_INFO_EXPIRATION", RFC3339Util.dateToRFC3339String(expiration, true/*force zulu*/));

        for (GeniUrn sliverUrn : sliverUrns) {
            LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is registering: "+ sliverUrn);
            AbstractFederationApi.FederationApiReply<String> reply =
                    sa(logger).updateSliverInfo(
                            getSaConnection(),
                            sliceCredentials,
                            sliverUrn+"",
                            fields/*fields*/,
                            null/*extraoptions*/);
        }
    }
    @Override
    public void unregisterAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns) throws JFedException {
        LOG.debug("UniformFederationApi2UserAndSliceApiWrapper will unregister: "+ sliverUrns.size());
        GeniUrn creatorUrn = getUser().getUserUrn();
        Date expiration = null;
        for (AnyCredential c : sliceCredentials)
            if (c.getExpiresDate() != null && (expiration == null || expiration.after(c.getExpiresDate())))
                expiration = c.getExpiresDate();

        //first check which slivers are still registered
        List<String> registeredSliverUrns = null;
        if (sliceUrn != null)
            try {
                AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                        sa(logger).lookupSliverInfo(getSaConnection(),
                                sliceCredentials,
                                sliceUrn,
                                null /*match*/,
                                null/*filter*/,
                                null/*extraoptions*/);
                AbstractFederationApi.LookupResult lres = reply.getValue();
                /* example reply:
                {
                "output": "",
                "code": 0,
                "value": {
                              "urn:publicid:IDN+wall2.ilabt.iminds.be+sliver+43869": {
                                     "SLIVER_INFO_AGGREGATE_URN": "urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm",
                                     "SLIVER_INFO_SLICE_URN": "urn:publicid:IDN+ch.geni.net:fed4firetest3+slice+tstW",
                                     "SLIVER_INFO_EXPIRATION": "2015-01-12T14:10:30Z",
                                     "SLIVER_INFO_CREATOR_URN": "urn:publicid:IDN+ch.geni.net+user+wimvande",
                                     "SLIVER_INFO_CREATION": "",
                                     "SLIVER_INFO_URN": "urn:publicid:IDN+wall2.ilabt.iminds.be+sliver+43869"
                                     }
                              }
                } */
                if (lres != null) {
                    registeredSliverUrns = new ArrayList<String>(lres.getKeys());
                }
            } catch (Throwable t) {
                LOG.warn("Exception checking registered slivers before deleting them", t);
                /*ignore*/
                registeredSliverUrns = null;
            }

        for (GeniUrn sliverUrn : sliverUrns) {
            if (registeredSliverUrns == null || !registeredSliverUrns.contains(sliverUrn.toString())) {
                LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is unregistering: " + sliverUrn);
                AbstractFederationApi.FederationApiReply<Boolean> reply =
                        sa(logger).deleteSliverInfo(getSaConnection(),
                                sliceCredentials,
                                sliverUrn.toString(),
                                null/*extraoptions*/);
            } else {
                LOG.debug("UniformFederationApi2UserAndSliceApiWrapper is skipping delete of SLIVER_INFO because sliver is not registered anymore.");
            }
        }
    }

    @Override
    public void shareSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn userUrn) throws JFedException {
        List<FederationSliceAuthorityApi2.UrnRoleTuple> membersToAdd =
                new ArrayList<FederationSliceAuthorityApi2.UrnRoleTuple>();
        membersToAdd.add(new FederationSliceAuthorityApi2.UrnRoleTuple("SLICE"/*typeName*/, userUrn, "MEMBER"));

        AbstractFederationApi.FederationApiReply<String> reply =
                sa(logger).modifyMembership(getSaConnection(),
                        "SLICE"/*type*/,
                        sliceUrn,
                        membersToAdd,
                        new ArrayList<GeniUrn>() /*membersToRemove*/,
                        new ArrayList<FederationSliceAuthorityApi2.UrnRoleTuple>() /*membersToChange*/,
                        sliceCredentials,
                        null /*extraOptions*/
                );

        if (reply == null || reply.getGeniResponseCode() == null)
            throw new JFedException("Sharing slice with "+userUrn+" did not succeed: ", reply);

        if (!reply.getGeniResponseCode().isSuccess()) {
            if (reply.getOutput().contains("DUPLICATE")) {
                LOG.warn("Reply to shared slice was \""+reply.getOutput()+"\". Assuming this means it is successfully shared already.");
                return;
            }
            throw new JFedException("Sharing slice with "+userUrn+" did not succeed: ", reply);
        }
    }

    @Override
    public List<GeniUrn> getUsersForSubAuthority(Logger logger, String projectName) throws JFedException {
        GeniUrn projectUrn = GeniUrn.createGeniUrnFromEncodedParts(getUserAuthority().getNameForUrn(), "project", projectName);
        AbstractFederationApi.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupMembers(getSaConnection(), "PROJECT", projectUrn, userCredentials, null/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            if (reply == null)
                throw new JFedException("Error getting users in project "+projectUrn);
            else {
                throw new JFedException("Error getting users in project "+projectUrn, reply);
            }

        List<GeniUrn> res = new ArrayList<GeniUrn>();
        for (FederationSliceAuthorityApi2.UrnRoleTuple tuple : reply.getValue()) {
            res.add(tuple.getUrn());
        }

        return res;
    }

    @Override
    public List<GeniUrn> getUsersForSlice(Logger logger, GeniUrn sliceUrn) throws JFedException {
        AbstractFederationApi.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                sa(logger).lookupMembers(getSaConnection(), "SLICE", sliceUrn, userCredentials, null/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            if (reply == null)
                throw new JFedException("Error getting users in "+sliceUrn);
            else {
                throw new JFedException("Error getting users in "+sliceUrn, reply);
            }

        List<GeniUrn> res = new ArrayList<GeniUrn>();
        for (FederationSliceAuthorityApi2.UrnRoleTuple tuple : reply.getValue()) {
            res.add(tuple.getUrn());
        }

        return res;
    }

    @Override
    public void addUsersToSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<GeniUrn> userUrns, String role) throws JFedException {
        List<FederationSliceAuthorityApi2.UrnRoleTuple> membersToAdd = new ArrayList<FederationSliceAuthorityApi2.UrnRoleTuple>();
        for (GeniUrn userUrn : userUrns)
            membersToAdd.add(new FederationSliceAuthorityApi2.UrnRoleTuple("SLICE", userUrn, role));

        AbstractFederationApi.FederationApiReply<String> reply =
                sa(logger).modifyMembership(getSaConnection(),
                        "SLICE",
                        sliceUrn,
                        membersToAdd,
                        null /*membersToRemove*/,
                        null /*membersToChange*/,
                        sliceCredentials,
                        null/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess())
            if (reply == null)
                throw new JFedException("Error adding user to "+sliceUrn);
            else {
                throw new JFedException("Error adding user to "+sliceUrn, reply);
            }
    }

    @Override
    public SliceInfo createSlice(Logger logger, String sliceName, Date expirationDate, String subAuthName) throws JFedException {
        LOG.debug("createSlice called");
        //2 steps: create slice, and then get its credential(s)

        FederationSliceAuthorityApi2.GetVersionSAResult gv = getGetVersionSAResult(logger);
        assert gv != null;

        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLICE_DESCRIPTION", "jFed Experimenter GUI slice '" + sliceName + "' for user " + getUser().getUserUrnString());

        if (gv.getServices().contains("PROJECT")) {
            if (subAuthName == null) {
                List<GeniUrn> userProjects = getUserProjects(logger, getUser().getUserUrn());
                if (userProjects.isEmpty())
                    throw new JFedException("Slice creation requires a project, but user is not a member of any project.");

                if (userProjects.size() <= 0) {
                    LOG.error("User " + getUser().getUserUrnString() + " is member of no projects! Will try to create slice without project, but that will most likely fail.");
                    return null;
                } else {
                    GeniUrn chosenProject = null;
                    if (userProjects.size() > 1) {
                        if (ExtraInfoCallback.getFederationApiProjectSelectionCallback() != null) {
                            LOG.debug("User is member of " + userProjects.size() + " projects. Using callback to let user choose one.");
                            chosenProject = ExtraInfoCallback.getFederationApiProjectSelection(userProjects);
                        } else {
                            LOG.warn("User " + getUser().getUserUrnString() + " is member of multiple projects: " + userProjects + ". " +
                                    "No callback available to let user choose project: Choosing first project as project for this slice.");
                            chosenProject = userProjects.get(0);
                        }
                    } else
                        chosenProject = userProjects.get(0);
                    fields.put("SLICE_PROJECT_URN", chosenProject.getValue());
                }
            } else {
                GeniUrn project = GeniUrn.createGeniUrnFromEncodedParts(getUser().getUserUrn().getEncodedTopLevelAuthority(), "project", subAuthName);
                fields.put("SLICE_PROJECT_URN", project.getValue());
            }
        }

        if (gv.getFieldsForObject("PROJECT").containsKey("_GENI_SLICE_EMAIL")) {
            AbstractFederationApi2.LookupResult userIdentifyingInfo =
                    getUserIdentifyingInfo(logger, getUser().getUserUrn());

            if (userIdentifyingInfo != null && userIdentifyingInfo.get(getUser().getUserUrn()).containsKey("MEMBER_EMAIL")) {
                String userEmail = (String) userIdentifyingInfo.get(getUser().getUserUrn()).get("MEMBER_EMAIL");
                fields.put("_GENI_SLICE_EMAIL", userEmail);
            } else {
                throw new JFedException("Slice creation requires a user email, but the user email is not known.");
            }
        }

//      TODO  addRequiredSupplementaryFieldsToCreate(fields, "SLICE");

        if (expirationDate != null) {
            fields.put("SLICE_EXPIRATION", RFC3339Util.dateToRFC3339String(expirationDate, true, true, true));
        }

        LOG.debug("createSlice is calling server with sliceName=" + sliceName + " and fields: " + fields);

        AbstractFederationApi2.FederationApiReply<Hashtable<String, Object>> reply =
                sa(logger).createSlice(
                        getSaConnection(),
                        userCredentials,
                        sliceName/* automatically added to "SLICE_NAME" field*/,
                        fields,
                        null/*extraOptions*/);

        LOG.debug("createSlice got answer from server");

        if (reply != null && reply.getGeniResponseCode() != null && reply.getGeniResponseCode().isSuccess() && reply.getValue() != null) {
            Hashtable<String, Object> sliceInfo = reply.getValue();

//            if (sliceInfo.containsKey("SLICE_CREDENTIAL"?)) {
            //        //TODO slice credential is normally in returned fields? Otherwise, we can retrieve it with a call.
//            }

            if (sliceInfo.containsKey("SLICE_URN")) {
                String sliceUrnStr = sliceInfo.get("SLICE_URN").toString();
                GeniUrn sliceUrn = GeniUrn.parse(sliceUrnStr);
                if (sliceUrn != null) {
                    AbstractFederationApi2.FederationApiReply<List<AnyCredential>> credReply =
                            sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, null);

                    if (!credReply.getGeniResponseCode().isSuccess())
                        throw new JFedException("Failed to get slice credentials", credReply);

                    return new SliceInfo(sliceName, sliceUrn, credReply.getValue());
                }
                throw new JFedException("Error creating slice, invalid slice urn: " + sliceUrnStr, reply);
            }
            throw new JFedException("Error creating slice, no SLICE_URN returned", reply);
        } else {
            if (reply == null)
                throw new JFedException("Error creating slice");
            else {
                throw new JFedException("Error creating slice", reply);
            }
        }
    }

    @Override
    public List<AnyCredential> renewSlice(Logger logger, List<AnyCredential> sliceCredentials, Date newExpirationDate) throws JFedException {
        GeniUrn sliceUrn = null;
        for (AnyCredential cred : sliceCredentials)
            if (cred instanceof SfaCredential) {
                SfaCredential sfaCredential = (SfaCredential) cred;
                GeniUrn targetUrn = GeniUrn.parse(sfaCredential.getTargetUrn());
                if (targetUrn != null && targetUrn.getEncodedResourceType().equals("slice")) {
                    sliceUrn = targetUrn;
                    break;
                }
            }
        if (sliceUrn == null)
            throw new JFedException("Could not find slice URN in provided slice credential list.");
        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLICE_EXPIRATION", RFC3339Util.dateToRFC3339String(newExpirationDate));
        AbstractFederationApi.FederationApiReply<String> updateReply = sa(logger).
                updateSlice(getSaConnection(), sliceCredentials, sliceUrn, fields, null/*extraOptions*/);

        if (!updateReply.getGeniResponseCode().isSuccess())
            throw new JFedException("Failed to update SLICE_EXPIRATION", updateReply);

        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> credReply =
                sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, null);

        if (!credReply.getGeniResponseCode().isSuccess())
            throw new JFedException("Failed to get updated slice credentials", credReply);

        return  credReply.getValue();
    }

    @Override
    public List<String> getSshKeysForUser(Logger logger, GeniUrn userUrn) throws JFedException {
        Map<String, String> match = new HashMap<String, String>();
        match.put("KEY_MEMBER", userUrn.toString());
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                ma(logger).lookupKeyForMember(getMaConnection(), userCredentials, userUrn, null/*match*/, null/*filter*/, null);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Failed to get SSH keys", reply);

        AbstractFederationApi.LookupResult lRes = reply.getValue();
        List<String> res = new ArrayList<String>();
        for (Object o : lRes.getRaw().values()) {
            Hashtable<String, Object> ht = (Hashtable<String, Object>) o;
            Object publicKey = ht.get("KEY_PUBLIC");
            if (publicKey != null)
                res.add(publicKey+"");
        }

        Map<String, String> matchMember = new HashMap<String, String>();
        matchMember.put("MEMBER_URN", userUrn.toString());
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply2 =
                ma(logger).lookupMember(getMaConnection(), userCredentials, matchMember/*match*/, null/*filter*/, null);
        try {
            if (!reply.getGeniResponseCode().isSuccess())
                throw new JFedException("Failed to get SSH keys from user certificate because lookup MEMBER "+userUrn+" call failed",
                        reply);
            //add certificate ssh key
            AbstractFederationApi.LookupResult lookupResult = reply2.getValue();
            String key = "_GENI_MEMBER_SSL_CERTIFICATE";
            if (lookupResult == null || lookupResult.size() != 1 || lookupResult.get(0).get(key) == null)
                throw new JFedException("Failed to get SSH keys from user certificate because lookup MEMBER "+userUrn+" call does not contain "+key,
                        reply);
            String x509pemCerts = (String) lookupResult.get(0).get(key);
            List<X509Certificate> certs = KeyUtil.pemToX509CertificateChain(x509pemCerts);
            for (X509Certificate cert : certs) {
                //a bit of a hack, to find the user certificate in the chain
                Collection<List<?>> altNames = cert.getSubjectAlternativeNames();
                for (List<?> altName : altNames) {
                    Integer nameType = (Integer) altName.get(0);
                    String altKey = altName.get(1).toString();

                    if (altKey.contains("+user+")) {
                        res.add(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(cert.getPublicKey()));
                        break;
                    }
                }
            }
        } catch (Exception e) {
            LOG.warn("Failed to get SSH keys from user certificate", e);
        }

        return res;
    }


    /* sub authority support */
    @Override
    public SubAuthoritySupport getSubAuthoritySupport(Logger logger) throws JFedException {
        FederationSliceAuthorityApi2.GetVersionSAResult getSAVersionRes = getGetVersionSAResult(logger);
        if (!getSAVersionRes.getServices().contains("PROJECT"))
            return SubAuthoritySupport.SUB_AUTHORITY_FORBIDDEN;
        else
            return SubAuthoritySupport.SUB_AUTHORITY_MANDATORY;
    }

    @Override
    public List<String> getSubAuthorityNames(Logger logger, GeniUrn user) throws JFedException {
        List<String> res = new ArrayList<String>();

        List<GeniUrn> projectUrns = getUserProjects(logger, user);
        for (GeniUrn projectUrn : projectUrns) {
            res.add(projectUrn.getEncodedResourceName());
        }

        LOG.debug("returning subauthorities: " + res);

        return res;
    }
}
