package be.iminds.ilabt.jfed.util;


import java.util.Hashtable;
import java.util.Vector;

/**
 * This helper class "stringifies" XMLRPC output (a structure made out of: Vector, Hashtable, String, Integer, Double, ...)
 */
public class XmlRpcPrintUtil {
    private static String prependToEachNextLine(String in, String texttoprepend) {
        StringBuilder res = new StringBuilder(in.length() + 100);
        for (int i = 0; i < in.length(); i++) {
            char c = in.charAt(i);
            if (c == '\n' && i != in.length() - 1) {
                res.append(c).append(texttoprepend);
            } else
                res.append(c);
        }
        return res.toString();
    }

    public static String xmlRpcObjectToString(Object o) {
        return printXmlRpcResultObject(o, "");
    }

    /**
     * uses a JSON syntax   (not JSON like, but JSON compatible syntax, including character escapes)
     */
    private static String printXmlRpcResultObject(Object o, String texttoprepend) {
        if (o == null) return "";
        StringBuilder res = new StringBuilder(100);
        boolean knownClass = false;
        if (o.getClass() == Hashtable.class) {
            Hashtable h = (Hashtable) o;
            if (h.size() == 0)
                res.append("{ }");
            else {
                res.append("{\n");
                int j = 0;
                for (Object key : h.keySet()) {
                    Object val = h.get(key);
                    String preptext = texttoprepend;
                    int shiftLen = 6 + key.toString().length() + 3;
                    if (shiftLen > 6 + 6 + 3) shiftLen = 4 + 3;
                    for (int i = 0; i < shiftLen; i++)
                        preptext += " ";
                    if (key instanceof String)
                        res.append(texttoprepend).append("\"").append(key).append("\": ").append(printXmlRpcResultObject(val, preptext));
                    else
                        res.append(texttoprepend).append("").append(key).append(": ").append(printXmlRpcResultObject(val, preptext));
                    if (j++ < h.keySet().size() - 1)
                        res.append(",");
                    res.append("\n");
                }
                res.append(texttoprepend).append("}");
            }
            knownClass = true;
        }
        if (o.getClass() == Vector.class) {
            Vector v = (Vector) o;
            if (v.size() == 0)
                res.append("[ ]");
            else {
                res.append("[\n");
                int i = 0;
                for (Object e : v) {
                    res.append(texttoprepend).append(printXmlRpcResultObject(e, texttoprepend + "   "));
                    if (i++ < v.size() - 1)
                        res.append(",");
                    res.append("\n");
                }
                res.append(texttoprepend).append("]");
            }
            knownClass = true;
        }
        if (o.getClass() == String.class) {
            String str = (String) o;
            StringBuilder s = new StringBuilder();

            //escape what needs to be escaped or is nice to escape:
            //from JSON RFC:
            //    characters that must be escaped:
            //   quotation mark, reverse solidus, and the control characters (U+0000
            //   through U+001F).
            //
            //       char = unescaped /
            //              escape (
            //                  %x22 /          ; "    quotation mark  U+0022
            //                  %x5C /          ; \    reverse solidus U+005C
            //                  %x2F /          ; /    solidus         U+002F
            //                  %x62 /          ; b    backspace       U+0008
            //                  %x66 /          ; f    form feed       U+000C
            //                  %x6E /          ; n    line feed       U+000A
            //                  %x72 /          ; r    carriage return U+000D
            //                  %x74 /          ; t    tab             U+0009
            //                  %x75 4HEXDIG )  ; uXXXX                U+XXXX
            for (int i = 0 ; i < str.length(); i++) {
                char c = str.charAt(i);
                switch (c) {
                    case '"' : { s.append("\\\""); break;  }
                    case '\\' : { s.append("\\\\"); break; }
                    case '\b' : { s.append("\\b"); break;  }
                    case '\f' : { s.append("\\f"); break;  }
                    case '\n' : { s.append("\\n"); break;  } //must be escaped: is \u000A
                    case '\r' : { s.append("\\r"); break;  }
                    case '\t' : { s.append("\\t"); break;  }
                    default: {
                        long l = c | 0x10000;
                        if (l >= 0 && l <= 0x001F)
                            s.append("\\u").append(Integer.toHexString(c | 0x10000).substring(1));
                        else
                            s.append(c);
                        break;
                    }
                }
            }

            if (s.length() > 10 * 1024 * 1024) //limit to 10MB
                res.append("\"").append(prependToEachNextLine("=== WARNING: String to long (" + s.length() + " chars) to visualize here! ===", texttoprepend)).append("\"");
            else
                res.append("\"").append(prependToEachNextLine(s.toString(), texttoprepend)).append("\"");
            knownClass = true;
        }
        //useless, as a string is added:
//        if (o.getClass() == AnyCredential.class) {
//            AnyCredential c = (AnyCredential) o;
//            res += "AnyCredential named \""+c.getName()+"\" size="+c.getCredentialXml().length();
//            knownClass = true;
//        }
        if (!knownClass)
            res.append(prependToEachNextLine(o.toString(), texttoprepend));
        return res.toString();
    }
}
