package be.iminds.ilabt.jfed;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.TestCertHelper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnectionPool;
import com.google.inject.Provides;

import javax.inject.Singleton;
import java.util.HashMap;
import java.util.Map;

import static org.mockito.Mockito.mock;

/**
 * Fake automatic wrappers for testing
 *
 * This replaces the automatic wrappers by mocked objects.
 * This never goes low level, replies to requests are mocked at a high level.
 */
public class MockAutomaticWrappersTestCoreModule extends MockHttpRepliesTestCoreModule {
    public MockAutomaticWrappersTestCoreModule() {
        this(new TestCertHelper());
    }
    public MockAutomaticWrappersTestCoreModule(TestCertHelper testCertHelper) {
        super(testCertHelper);
    }

    @Override
    protected void configure() {
        bind(JFedConnectionProvider.class).to(SfaConnectionPool.class).in(Singleton.class);
//        bind(ConnectionFactory.class).to(BasicConnectionFactory.class).in(Singleton.class);
    }

    @Provides
    @Singleton
    public AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory provideAutomaticAggregateManagerWrapperFactory() {
        return new AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory() {
            Map<String, AutomaticAggregateManagerWrapper> mocks = new HashMap<String, AutomaticAggregateManagerWrapper>();

            @Override
            public AutomaticAggregateManagerWrapper create(SfaAuthority amAuthority) throws AggregateManagerWrapper.NoAmException {
                return create(null, amAuthority);
            }

            @Override
            public AutomaticAggregateManagerWrapper create(Logger altLogger, SfaAuthority amAuthority) throws AggregateManagerWrapper.NoAmException {
                assert amAuthority != null;
                String urn = amAuthority.getUrnString();
                AutomaticAggregateManagerWrapper res = mocks.get(urn);
                if (res == null) {
                    res = mock(AutomaticAggregateManagerWrapper.class);
                    mocks.put(urn, res);
                }
                return res;
            }
        };
    }

    @Provides
    @Singleton
    public AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory provideAutomaticUserAndSliceApiWrapperFactory() {
        return new AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory() {
            AutomaticUserAndSliceApiWrapper res = mock(AutomaticUserAndSliceApiWrapper.class);
            @Override
            public AutomaticUserAndSliceApiWrapper create()  {
                return res;
            }

            @Override
            public AutomaticUserAndSliceApiWrapper create(Logger altLogger) {
                return res;
            }

            @Override
            public AutomaticUserAndSliceApiWrapper create(GeniUser altUser) {
                return res;
            }

            @Override
            public AutomaticUserAndSliceApiWrapper create(Logger altLogger, GeniUser altUser) {
                return res;
            }
        };
    }
}
