package be.iminds.ilabt.jfed.util;

import org.testng.Assert;
import org.testng.annotations.Test;

import java.util.Deque;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Vector;

/**
 * JsonHelperTest
 */
public class JsonHelperTest {
    @Test
    public void testInt() {
        Object res = testIsJSONReconstructable("15");
        Assert.assertEquals(res, new Integer(15));
    }
    @Test
    public void testString() {
        Object res = testIsJSONReconstructable("\"teststring\"");
        Assert.assertEquals(res, "teststring");
    }
    @Test
    public void testArray() {
        Object res = testIsJSONReconstructable("[ \"test\", 15, 12.6 ]");
        Vector v = (Vector) res;
        Assert.assertEquals(v.size(), 3);
        Assert.assertEquals(v.get(0), "test");
        Assert.assertEquals(v.get(1), new Integer(15));
        Assert.assertEquals(v.get(2), new Double(12.6));
    }
    @Test
    public void testHt() {
        Object res = testIsJSONReconstructable("{ \"a\" : \"test\", \"b\" : 15, \"c\" : 12.6 }");
        Hashtable ht = (Hashtable) res;
        Assert.assertEquals(ht.size(), 3);
        Assert.assertEquals(ht.get("a"), "test");
        Assert.assertEquals(ht.get("b"), new Integer(15));
        Assert.assertEquals(ht.get("c"), new Double(12.6));
    }
    @Test
    public void testNull() {
        Object res = testIsJSONReconstructable("");
        Assert.assertNull(res);
    }
    @Test
    public void testEmptyString() {
        String c = "";
        Object res = testIsJSONReconstructable("\""+c+"\"");
        Assert.assertEquals(res, c);
    }
    @Test
    public void testComplexString1() {
        String c = "\nThis\nis\na\ntest\n\n";
        System.out.println("testComplexString1 origString="+c);
        String jsonString = JSonHelper.jsonObjectToString(c);
        System.out.println("testComplexString1 jsonString="+jsonString);
        Object res = testIsJSONReconstructable(jsonString);
        Assert.assertEquals(res, c);
    }
    @Test
    public void testComplexString2() {
        String c = "This\nis\"\nalso ' a test ] [ } < { with funny : \\ chars newline=\"\\n\"";
        System.out.println("testComplexString2 origString="+c);
        String jsonString = JSonHelper.jsonObjectToString(c);
        System.out.println("testComplexString2 jsonString="+jsonString);
        Object res = testIsJSONReconstructable(jsonString);
        Assert.assertEquals(res, c);
    }
    @Test
    public void testComplex() {
        String c = "[ \"test\", [\"test\", [1, 2], \"blah\" , 4.0 ], { \"t1\" : [3, 4], \"t2\" : [\"a\", { \"o\" : 5 } ], \"t3\" : { \"3\" : 4, \"p\" : \"x\" } } ]";
        Object res = testIsJSONReconstructable(c);
    }


    private class ComparePair {
        public Object o1;
        public Object o2;
        private ComparePair(Object o1, Object o2) {
            this.o1 = o1;
            this.o2 = o2;
        }
    }
    /**
     * changes the jsonString into an object, and back again, and checks if it is the same.
     * @param jsonString
     * @return json object created (can be used for further testing)
     */
    public Object testIsJSONReconstructable(String jsonString) {
        Object obj1 = JSonHelper.jsonStringToXmlRpcLikeObject(jsonString);
        String reconstructed = JSonHelper.jsonObjectToString(obj1);
        Object obj2 = JSonHelper.jsonStringToXmlRpcLikeObject(reconstructed);

        if (obj1 == null || obj2 == null) {
            Assert.assertNull(obj1,
                    "obj1="+(obj1 == null ? "null" : ""+obj1.getClass().getName()+"("+obj1+")")+
                    " obj2="+(obj2 == null ? "null" : ""+obj2.getClass().getName()+"("+obj2+")")+
                    " reconstructed="+reconstructed+
                    " orig="+jsonString);
            Assert.assertNull(obj2,
                    "obj1="+(obj1 == null ? "null" : ""+obj1.getClass().getName()+"("+obj1+")"+
                    " obj2="+(obj2 == null ? "null" : ""+obj2.getClass().getName()+"("+obj2+")"))+
                    " reconstructed="+reconstructed+
                    " orig="+jsonString);
        } else {
            Deque<ComparePair> toCompare = new LinkedList<ComparePair>();
            toCompare.add(new ComparePair(obj1, obj2));
            while (!toCompare.isEmpty()) {
                ComparePair comparePair = toCompare.pollFirst();
                Object o1 = comparePair.o1;
                Object o2 = comparePair.o2;

                boolean handled = false;
                if (o1 instanceof String) {
                    handled = true;
                    Assert.assertTrue(o2 instanceof String);

                    String s1 = (String) o1;
                    String s2 = (String) o2;
                }
                if (o1 instanceof Vector) {
                    handled = true;
                    Assert.assertTrue(o2 instanceof Vector);

                    Vector v1 = (Vector) o1;
                    Vector v2 = (Vector) o2;
                    Assert.assertEquals(v1.size(), v2.size());
                    for (int i = 0; i < v1.size(); i++)
                        toCompare.add(new ComparePair(v1.get(i), v2.get(i)));
                }
                if (o1 instanceof Hashtable) {
                    handled = true;
                    Assert.assertTrue(o2 instanceof Hashtable);

                    Hashtable ht1 = (Hashtable) o1;
                    Hashtable ht2 = (Hashtable) o2;
                    Assert.assertEquals(ht1.size(), ht2.size());
                    for (Object key : ht1.keySet()) {
                        Object value1 = ht1.get(key);
                        Object value2 = ht2.get(key);
                        Assert.assertNotNull(value1);
                        Assert.assertNotNull(value2);
                        toCompare.add(new ComparePair(value1, value2));
                    }
                }
                if (!handled) {
                    if (o1 != null && o2 != null)
                        Assert.assertEquals(o2, o1);
                }
            }
        }

        return obj1;
    }
}
